@@ -49,7 +49,7 @@
 #include "fcitx-utils/utils.h"
 #include "errorhandler.h"
 
-FcitxInstance* instance;
+FcitxInstance* instance = NULL;
 
 static void WaitForEnd(sem_t *sem, int count)
 {
@@ -61,15 +61,6 @@ static void WaitForEnd(sem_t *sem, int count)
 
 int main(int argc, char* argv[])
 {
-#ifdef M_TRIM_THRESHOLD
-#ifdef HAVE_UNISTD_H
-    int pagesize = sysconf(_SC_PAGESIZE);
-#else
-    int pagesize = 4 * 1024;
-#endif // HAVE_UNISTD_H
-    mallopt(M_TRIM_THRESHOLD, 5 * pagesize);
-#endif // M_TRIM_THRESHOLD
-
     char* localedir = fcitx_utils_get_fcitx_path("localedir");
     setlocale(LC_ALL, "");
     bindtextdomain("fcitx", localedir);
