@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2010~2011 by CSSlayer                                   *
+ *   Copyright (C) 2010~2012 by CSSlayer                                   *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -35,7 +35,7 @@
 #include "fcitximcontext.h"
 #include "fcitx-config/fcitx-config.h"
 #include "client.h"
-#include <fcitx-utils/log.h>
+#include "fcitx-utils/log.h"
 #include <dbus/dbus-glib.h>
 
 #define LOG_LEVEL DEBUG
@@ -90,6 +90,11 @@ static void     fcitx_im_context_set_cursor_location(GtkIMContext          *cont
         GdkRectangle             *area);
 static void     fcitx_im_context_set_use_preedit(GtkIMContext          *context,
         gboolean               use_preedit);
+static void
+fcitx_im_context_set_surrounding (GtkIMContext *context,
+                                 const gchar *text,
+                                 gint len,
+                                 gint cursor_index);
 static void     fcitx_im_context_get_preedit_string(GtkIMContext          *context,
         gchar                **str,
         PangoAttrList        **attrs,
@@ -130,6 +135,11 @@ _fcitx_im_context_forward_key_cb(DBusGProxy* proxy, guint keyval, guint state, g
 static void
 _fcitx_im_context_update_preedit_cb(DBusGProxy* proxy, char* str, int cursor_pos, void* user_data);
 static void
+_fcitx_im_context_delete_surrounding_text_cb (DBusGProxy* proxy,
+                                          gint offset_from_cursor,
+                                          guint nchars,
+                                          void* user_data);
+static void
 _fcitx_im_context_connect_cb(FcitxIMClient* client, void* user_data);
 static void
 _fcitx_im_context_destroy_cb(FcitxIMClient* client, void* user_data);
@@ -149,6 +159,9 @@ _create_gdk_event(FcitxIMContext *fcitxcontext,
 static gboolean
 _key_is_modifier(guint keyval);
 
+static void
+_request_surrounding_text (FcitxIMContext *context);
+
 static gint
 _key_snooper_cb (GtkWidget   *widget,
                  GdkEventKey *event,
@@ -200,20 +213,6 @@ static guint16 cedilla_compose_seqs[] = {
 #endif
 };
 
-static
-boolean FcitxIsHotKey(FcitxKeySym sym, int state, FcitxHotkey * hotkey);
-
-static
-boolean FcitxIsHotKey(FcitxKeySym sym, int state, FcitxHotkey * hotkey)
-{
-    state &= FcitxKeyState_Ctrl_Alt_Shift;
-    if (hotkey[0].sym && sym == hotkey[0].sym && (hotkey[0].state == state))
-        return true;
-    if (hotkey[1].sym && sym == hotkey[1].sym && (hotkey[1].state == state))
-        return true;
-    return false;
-}
-
 void
 fcitx_im_context_register_type(GTypeModule *type_module)
 {
@@ -281,6 +280,7 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
     im_context_class->focus_out = fcitx_im_context_focus_out;
     im_context_class->set_cursor_location = fcitx_im_context_set_cursor_location;
     im_context_class->set_use_preedit = fcitx_im_context_set_use_preedit;
+    im_context_class->set_surrounding = fcitx_im_context_set_surrounding;
     gobject_class->finalize = fcitx_im_context_finalize;
 
     _signal_commit_id =
@@ -329,7 +329,7 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
         }
         g_strfreev (apps);
     }
-    
+
     /* make ibus fix benefits us */
     _use_sync_mode = _get_boolean_env("IBUS_ENABLE_SYNC_MODE", FALSE)
                      || _get_boolean_env("FCITX_ENABLE_SYNC_MODE", FALSE);
@@ -352,6 +352,7 @@ fcitx_im_context_init(FcitxIMContext *context)
     context->cursor_pos = 0;
     context->preedit_string = NULL;
     context->attrlist = NULL;
+    context->capacity = CAPACITY_SURROUNDING_TEXT;
 
     context->slave = gtk_im_context_simple_new();
     gtk_im_context_simple_add_table(GTK_IM_CONTEXT_SIMPLE(context->slave),
@@ -410,7 +411,7 @@ fcitx_im_context_finalize(GObject *obj)
     if (context->attrlist)
         pango_attr_list_unref(context->attrlist);
     context->attrlist = NULL;
-    
+
     if (_key_snooper_id != 0) {
         gtk_key_snooper_remove (_key_snooper_id);
         _key_snooper_id = 0;
@@ -455,9 +456,9 @@ fcitx_im_context_filter_keypress(GtkIMContext *context,
 {
     FcitxLog(LOG_LEVEL, "fcitx_im_context_filter_keypress");
     FcitxIMContext *fcitxcontext = FCITX_IM_CONTEXT(context);
-    
+
     /* check this first, since we use key snooper, most key will be handled. */
-    if (IsFcitxIMClientValid(fcitxcontext->client)) {
+    if (IsFcitxIMClientValid(fcitxcontext->client) ) {
         /* XXX it is a workaround for some applications do not set client window. */
         if (fcitxcontext->client_window == NULL && event->window != NULL) {
             gtk_im_context_set_client_window((GtkIMContext *)fcitxcontext, event->window);
@@ -477,12 +478,8 @@ fcitx_im_context_filter_keypress(GtkIMContext *context,
     if (G_UNLIKELY(event->state & FcitxKeyState_IgnoredMask))
         return gtk_im_context_filter_keypress(fcitxcontext->slave, event);
 
-    if (IsFcitxIMClientValid(fcitxcontext->client)) {
-
-        if (!IsFcitxIMClientEnabled(fcitxcontext->client)) {
-            if (!FcitxIsHotKey(event->keyval, event->state, FcitxIMClientGetTriggerKey(fcitxcontext->client)))
-                return gtk_im_context_filter_keypress(fcitxcontext->slave, event);
-        }
+    if (IsFcitxIMClientValid(fcitxcontext->client) && fcitxcontext->has_focus) {
+        _request_surrounding_text (fcitxcontext);
 
         fcitxcontext->time = event->time;
 
@@ -605,22 +602,22 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
         g_free(context->preedit_string);
         context->preedit_string = NULL;
     }
-    
+
     if (context->attrlist != NULL) {
         pango_attr_list_unref(context->attrlist);
     }
-    
+
     context->attrlist = pango_attr_list_new();
-    
+
     GString* gstr = g_string_new(NULL);
-    
+
     int i = 0;
     for (i = 0; i < array->len; i++) {
         size_t bytelen = strlen(gstr->str);
         GValueArray* preedit = g_ptr_array_index(array, i);
         const gchar* s = g_value_get_string(g_value_array_get_nth(preedit, 0));
         gint type = g_value_get_int(g_value_array_get_nth(preedit, 1));
-        
+
         PangoAttribute *pango_attr = NULL;
         if ((type & MSG_NOUNDERLINE) == 0) {
             pango_attr = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
@@ -628,12 +625,12 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
             pango_attr->end_index = bytelen + strlen(s);
             pango_attr_list_insert(context->attrlist, pango_attr);
         }
-        
+
         if (type & MSG_HIGHLIGHT) {
             gboolean hasColor;
             GdkColor fg;
             GdkColor bg;
-            
+
             if (context->client_window) {
                 GtkWidget *widget;
                 gdk_window_get_user_data (context->client_window,
@@ -645,7 +642,7 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
                     bg = style->bg[GTK_STATE_SELECTED];
                 }
             }
-            
+
             if (!hasColor) {
                 fg.red = 0xffff;
                 fg.green = 0xffff;
@@ -654,7 +651,7 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
                 bg.green = 0xacff;
                 bg.blue = 0xe8ff;
             }
-            
+
             pango_attr = pango_attr_foreground_new(fg.red, fg.green, fg.blue);
             pango_attr->start_index = bytelen;
             pango_attr->end_index = bytelen + strlen(s);
@@ -666,9 +663,9 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
         }
         gstr = g_string_append(gstr, s);
     }
-    
+
     gchar* str = g_string_free(gstr, FALSE);
-    
+
     context->preedit_string = g_strdup(str);
     char* tempstr = g_strndup(str, cursor_pos);
     context->cursor_pos =  fcitx_utf8_strlen(tempstr);
@@ -711,11 +708,11 @@ fcitx_im_context_focus_in(GtkIMContext *context)
 
     if (fcitxcontext->has_focus)
         return;
-   
+
     _fcitx_im_context_set_capacity(fcitxcontext, FALSE);
 
     fcitxcontext->has_focus = true;
-    
+
     if (_focus_im_context != NULL) {
         g_assert (_focus_im_context != context);
         gtk_im_context_focus_out (_focus_im_context);
@@ -735,6 +732,8 @@ fcitx_im_context_focus_in(GtkIMContext *context)
                     g_object_ref(fcitxcontext),
                     (GDestroyNotify) g_object_unref);
 
+    _request_surrounding_text (fcitxcontext);
+
     g_object_add_weak_pointer ((GObject *) context,
                                (gpointer *) &_focus_im_context);
     _focus_im_context = context;
@@ -751,7 +750,7 @@ fcitx_im_context_focus_out(GtkIMContext *context)
     if (!fcitxcontext->has_focus) {
         return;
     }
-    
+
     g_assert (context == _focus_im_context);
     g_object_remove_weak_pointer ((GObject *) context,
                                   (gpointer *) &_focus_im_context);
@@ -763,13 +762,13 @@ fcitx_im_context_focus_out(GtkIMContext *context)
         FcitxIMClientFocusOut(fcitxcontext->client);
     }
 
-
-    if (fcitxcontext->preedit_string != NULL)
-        g_free(fcitxcontext->preedit_string);
-    fcitxcontext->preedit_string = NULL;
     fcitxcontext->cursor_pos = 0;
-    g_signal_emit(fcitxcontext, _signal_preedit_changed_id, 0);
-    g_signal_emit(fcitxcontext, _signal_preedit_end_id, 0);
+    if (fcitxcontext->preedit_string != NULL) {
+        g_free(fcitxcontext->preedit_string);
+        fcitxcontext->preedit_string = NULL;
+        g_signal_emit(fcitxcontext, _signal_preedit_changed_id, 0);
+        g_signal_emit(fcitxcontext, _signal_preedit_end_id, 0);
+    }
 
     gtk_im_context_focus_out(fcitxcontext->slave);
 
@@ -836,7 +835,7 @@ _set_cursor_location_internal(FcitxIMContext *fcitxcontext)
     }
 #endif
 
-    FcitxIMClientSetCursorLocation(fcitxcontext->client, area.x, area.y + area.height);
+    FcitxIMClientSetCursorRect(fcitxcontext->client, area.x, area.y, area.width, area.height);
     return FALSE;
 }
 
@@ -854,14 +853,114 @@ fcitx_im_context_set_use_preedit(GtkIMContext *context,
     gtk_im_context_set_use_preedit(fcitxcontext->slave, use_preedit);
 }
 
+
+static guint
+get_selection_anchor_point (FcitxIMContext *fcitxcontext,
+                            guint cursor_pos,
+                            guint surrounding_text_len)
+{
+    GtkWidget *widget;
+    if (fcitxcontext->client_window == NULL) {
+        return cursor_pos;
+    }
+    gdk_window_get_user_data (fcitxcontext->client_window, (gpointer *)&widget);
+
+    if (!GTK_IS_TEXT_VIEW (widget)){
+        return cursor_pos;
+    }
+
+    GtkTextView *text_view = GTK_TEXT_VIEW (widget);
+    GtkTextBuffer *buffer = gtk_text_view_get_buffer (text_view);
+
+    if (!gtk_text_buffer_get_has_selection (buffer)) {
+        return cursor_pos;
+    }
+
+    GtkTextIter start_iter, end_iter, cursor_iter;
+    if (!gtk_text_buffer_get_selection_bounds (buffer, &start_iter, &end_iter)) {
+        return cursor_pos;
+    }
+
+    gtk_text_buffer_get_iter_at_mark (buffer,
+                                      &cursor_iter,
+                                      gtk_text_buffer_get_insert (buffer));
+
+    guint start_index = gtk_text_iter_get_offset (&start_iter);
+    guint end_index = gtk_text_iter_get_offset (&end_iter);
+    guint cursor_index = gtk_text_iter_get_offset (&cursor_iter);
+
+    guint anchor;
+
+    if (start_index == cursor_index) {
+        anchor = end_index;
+    } else if (end_index == cursor_index) {
+        anchor = start_index;
+    } else {
+        return cursor_pos;
+    }
+
+    // Change absolute index to relative position.
+    guint relative_origin = cursor_index - cursor_pos;
+
+    if (anchor < relative_origin) {
+        return cursor_pos;
+    }
+    anchor -= relative_origin;
+
+    if (anchor > surrounding_text_len) {
+        return cursor_pos;
+    }
+
+    return anchor;
+}
+
+
+static void
+fcitx_im_context_set_surrounding (GtkIMContext *context,
+                                 const gchar *text,
+                                 gint len,
+                                 gint cursor_index)
+{
+    g_return_if_fail (context != NULL);
+    g_return_if_fail (FCITX_IS_IM_CONTEXT (context));
+    g_return_if_fail (text != NULL);
+    g_return_if_fail (strlen (text) >= len);
+    g_return_if_fail (0 <= cursor_index && cursor_index <= len);
+
+    FcitxIMContext *fcitxcontext = FCITX_IM_CONTEXT (context);
+
+    if (IsFcitxIMClientValid(fcitxcontext->client) && !(fcitxcontext->capacity & CAPACITY_PASSWORD)) {
+        guint cursor_pos;
+        guint utf8_len;
+        gchar *p;
+
+        p = g_strndup (text, len);
+        cursor_pos = g_utf8_strlen (p, cursor_index);
+        utf8_len = g_utf8_strlen(p, len);
+
+        guint anchor_pos = get_selection_anchor_point (fcitxcontext,
+                                                       cursor_pos,
+                                                       utf8_len);
+        FcitxIMClientSetSurroundingText (fcitxcontext->client,
+                                         p,
+                                         cursor_pos,
+                                         anchor_pos);
+        g_free (p);
+    }
+    gtk_im_context_set_surrounding (fcitxcontext->slave,
+                                    text,
+                                    len,
+                                    cursor_index);
+}
+
 void
 _fcitx_im_context_set_capacity(FcitxIMContext* fcitxcontext, gboolean force)
 {
     if (IsFcitxIMClientValid(fcitxcontext->client)) {
-        FcitxCapacityFlags flags = CAPACITY_NONE;
+        FcitxCapacityFlags flags = fcitxcontext->capacity & ~(CAPACITY_PREEDIT | CAPACITY_FORMATTED_PREEDIT | CAPACITY_PASSWORD);
         if (fcitxcontext->use_preedit)
             flags |= CAPACITY_PREEDIT | CAPACITY_FORMATTED_PREEDIT;
-        
+
         if (fcitxcontext->client_window != NULL) {
             GtkWidget *widget;
             gdk_window_get_user_data (fcitxcontext->client_window,
@@ -871,7 +970,7 @@ _fcitx_im_context_set_capacity(FcitxIMContext* fcitxcontext, gboolean force)
                 flags |= CAPACITY_PASSWORD;
             }
         }
-        
+
         gboolean update = FALSE;
         if (G_UNLIKELY(fcitxcontext->capacity != flags)) {
             fcitxcontext->capacity = flags;
@@ -905,7 +1004,7 @@ fcitx_im_context_get_preedit_string(GtkIMContext   *context,
     FcitxLog(LOG_LEVEL, "fcitx_im_context_get_preedit_string");
     FcitxIMContext *fcitxcontext = FCITX_IM_CONTEXT(context);
 
-    if (IsFcitxIMClientValid(fcitxcontext->client) && IsFcitxIMClientEnabled(fcitxcontext->client)) {
+    if (IsFcitxIMClientValid(fcitxcontext->client)) {
         if (str) {
             if (fcitxcontext->preedit_string)
                 *str = strdup(fcitxcontext->preedit_string);
@@ -948,7 +1047,7 @@ static void
 _slave_preedit_changed_cb(GtkIMContext *slave,
                           FcitxIMContext *context)
 {
-    if (context->client && IsFcitxIMClientEnabled(context->client)) {
+    if (context->client) {
         return;
     }
 
@@ -958,7 +1057,7 @@ static void
 _slave_preedit_start_cb(GtkIMContext *slave,
                         FcitxIMContext *context)
 {
-    if (context->client && IsFcitxIMClientEnabled(context->client)) {
+    if (context->client) {
         return;
     }
 
@@ -969,7 +1068,7 @@ static void
 _slave_preedit_end_cb(GtkIMContext *slave,
                       FcitxIMContext *context)
 {
-    if (context->client && IsFcitxIMClientEnabled(context->client)) {
+    if (context->client) {
         return;
     }
     g_signal_emit(context, _signal_preedit_end_id, 0);
@@ -981,7 +1080,7 @@ _slave_retrieve_surrounding_cb(GtkIMContext *slave,
 {
     gboolean return_value;
 
-    if (context->client && IsFcitxIMClientEnabled(context->client)) {
+    if (context->client) {
         return FALSE;
     }
     g_signal_emit(context, _signal_retrieve_surrounding_id, 0,
@@ -997,7 +1096,7 @@ _slave_delete_surrounding_cb(GtkIMContext *slave,
 {
     gboolean return_value;
 
-    if (context->client && IsFcitxIMClientEnabled(context->client)) {
+    if (context->client) {
         return FALSE;
     }
     g_signal_emit(context, _signal_delete_surrounding_id, 0, offset_from_cursor, nchars, &return_value);
@@ -1043,6 +1142,17 @@ void _fcitx_im_context_forward_key_cb(DBusGProxy* proxy, guint keyval, guint sta
     gdk_event_free((GdkEvent *)event);
 }
 
+static void
+_fcitx_im_context_delete_surrounding_text_cb (DBusGProxy* proxy,
+                                          gint offset_from_cursor,
+                                          guint nchars,
+                                          void* user_data)
+{
+    FcitxIMContext* context =  FCITX_IM_CONTEXT(user_data);
+    gboolean return_value;
+    g_signal_emit (context, _signal_delete_surrounding_id, 0, offset_from_cursor, nchars, &return_value);
+}
+
 /* Copy from gdk */
 static GdkEventKey *
 _create_gdk_event(FcitxIMContext *fcitxcontext,
@@ -1243,6 +1353,7 @@ void _fcitx_im_context_connect_cb(FcitxIMClient* client, void* user_data)
                                    G_CALLBACK(_fcitx_im_context_forward_key_cb),
                                    G_CALLBACK(_fcitx_im_context_update_preedit_cb),
                                    G_CALLBACK(_fcitx_im_context_update_formatted_preedit_cb),
+                                   G_CALLBACK(_fcitx_im_context_delete_surrounding_text_cb),
                                    context,
                                    NULL);
         _fcitx_im_context_set_capacity(context, TRUE);
@@ -1255,6 +1366,28 @@ void _fcitx_im_context_destroy_cb(FcitxIMClient* client, void* user_data)
     FcitxIMClientSetEnabled(client, false);
 }
 
+static void
+_request_surrounding_text (FcitxIMContext *context)
+{
+    if (context &&
+        IsFcitxIMClientValid(context->client)) {
+        gboolean return_value;
+        FcitxLog(LOG_LEVEL, "requesting surrounding text");
+        g_signal_emit (context, _signal_retrieve_surrounding_id, 0,
+                       &return_value);
+        if (return_value) {
+            context->capacity |= CAPACITY_SURROUNDING_TEXT;
+            _fcitx_im_context_set_capacity (context,
+                                            FALSE);
+        }
+        else {
+            context->capacity &= ~CAPACITY_SURROUNDING_TEXT;
+            _fcitx_im_context_set_capacity (context,
+                                            FALSE);
+        }
+    }
+}
+
 
 static gint
 _key_snooper_cb (GtkWidget   *widget,
@@ -1278,14 +1411,12 @@ _key_snooper_cb (GtkWidget   *widget,
         if (!IsFcitxIMClientValid(fcitxcontext->client)) {
             break;
         }
-        if (!IsFcitxIMClientEnabled(fcitxcontext->client)) {
-            if (!FcitxIsHotKey(event->keyval, event->state, FcitxIMClientGetTriggerKey(fcitxcontext->client)))
-                break;
-        }
+
+        _request_surrounding_text (fcitxcontext);
         fcitxcontext->time = event->time;
 
         if (_use_sync_mode) {
-            
+
             int ret = FcitxIMClientProcessKeySync(fcitxcontext->client,
                                                     event->keyval,
                                                     event->hardware_keycode,
@@ -1313,7 +1444,7 @@ _key_snooper_cb (GtkWidget   *widget,
             retval = TRUE;
         }
     } while(0);
-    
+
     if (!retval) {
         event->state |= FcitxKeyState_IgnoredMask;
         return FALSE;
