@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2010~2011 by CSSlayer                                   *
+ *   Copyright (C) 2010~2012 by CSSlayer                                   *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -37,6 +37,12 @@ typedef struct _FcitxIPCIC {
     int id;
     char path[32];
     char* appname;
+    int width;
+    int height;
+    pid_t pid;
+    char* surroundingText;
+    unsigned int anchor;
+    unsigned int cursor;
 } FcitxIPCIC;
 
 typedef struct _FcitxIPCFrontend {
@@ -69,20 +75,24 @@ static void IPCCloseIM(void* arg, FcitxInputContext* ic);
 static void IPCCommitString(void* arg, FcitxInputContext* ic, const char* str);
 static void IPCForwardKey(void* arg, FcitxInputContext* ic, FcitxKeyEventType event, FcitxKeySym sym, unsigned int state);
 static void IPCSetWindowOffset(void* arg, FcitxInputContext* ic, int x, int y);
-static void IPCGetWindowPosition(void* arg, FcitxInputContext* ic, int* x, int* y);
+static void IPCGetWindowRect(void* arg, FcitxInputContext* ic, int* x, int* y, int* w, int* h);
 static void IPCUpdatePreedit(void* arg, FcitxInputContext* ic);
+static void IPCDeleteSurroundingText(void* arg, FcitxInputContext* ic, int offset, unsigned int size);
+static boolean IPCGetSurroundingText(void* arg, FcitxInputContext* ic, char** str, unsigned int* cursor, unsigned int* anchor);
 static void IPCUpdateClientSideUI(void* arg, FcitxInputContext* ic);
 static DBusHandlerResult IPCDBusEventHandler(DBusConnection *connection, DBusMessage *message, void *user_data);
 static DBusHandlerResult IPCICDBusEventHandler(DBusConnection *connection, DBusMessage *msg, void *user_data);
 static void IPCICFocusIn(FcitxIPCFrontend* ipc, FcitxInputContext* ic);
 static void IPCICFocusOut(FcitxIPCFrontend* ipc, FcitxInputContext* ic);
 static void IPCICReset(FcitxIPCFrontend* ipc, FcitxInputContext* ic);
-static void IPCICSetCursorLocation(FcitxIPCFrontend* ipc, FcitxInputContext* ic, int x, int y);
+static void IPCICCommitPreedit(FcitxIPCFrontend* ipc, FcitxInputContext* ic);
+static void IPCICSetCursorRect(FcitxIPCFrontend* ipc, FcitxInputContext* ic, int x, int y, int w, int h);
 static int IPCProcessKey(FcitxIPCFrontend* ipc, FcitxInputContext* callic, const uint32_t originsym, const uint32_t keycode, const uint32_t originstate, uint32_t t, FcitxKeyEventType type);
 static boolean IPCCheckICFromSameApplication(void* arg, FcitxInputContext* icToCheck, FcitxInputContext* ic);
 static void IPCGetPropertyIMList(void* arg, DBusMessageIter* iter);
 static void IPCSetPropertyIMList(void* arg, DBusMessageIter* iter);
 static void IPCUpdateIMList(void* arg);
+static pid_t IPCGetPid(void* arg, FcitxInputContext* ic);
 
 static void FcitxDBusPropertyGet(FcitxIPCFrontend* ipc, DBusMessage* message);
 static void FcitxDBusPropertySet(FcitxIPCFrontend* ipc, DBusMessage* message);
@@ -132,7 +142,17 @@ const char * im_introspection_xml =
     "      <arg name=\"state2\" direction=\"out\" type=\"u\"/>\n"
     "    </method>\n"
     "    <method name=\"CreateICv2\">\n"
-    "      <arg name=\"appname\" direction=\"int\" type=\"s\"/>\n"
+    "      <arg name=\"appname\" direction=\"in\" type=\"s\"/>\n"
+    "      <arg name=\"icid\" direction=\"out\" type=\"i\"/>\n"
+    "      <arg name=\"enable\" direction=\"out\" type=\"b\"/>\n"
+    "      <arg name=\"keyval1\" direction=\"out\" type=\"u\"/>\n"
+    "      <arg name=\"state1\" direction=\"out\" type=\"u\"/>\n"
+    "      <arg name=\"keyval2\" direction=\"out\" type=\"u\"/>\n"
+    "      <arg name=\"state2\" direction=\"out\" type=\"u\"/>\n"
+    "    </method>\n"
+    "    <method name=\"CreateICv3\">\n"
+    "      <arg name=\"appname\" direction=\"in\" type=\"s\"/>\n"
+    "      <arg name=\"pid\" direction=\"in\" type=\"i\"/>\n"
     "      <arg name=\"icid\" direction=\"out\" type=\"i\"/>\n"
     "      <arg name=\"enable\" direction=\"out\" type=\"b\"/>\n"
     "      <arg name=\"keyval1\" direction=\"out\" type=\"u\"/>\n"
@@ -142,6 +162,25 @@ const char * im_introspection_xml =
     "    </method>\n"
     "    <method name=\"Exit\">\n"
     "    </method>\n"
+    "    <method name=\"GetCurrentIM\">\n"
+    "      <arg name=\"im\" direction=\"out\" type=\"s\"/>\n"
+    "    </method>\n"
+    "    <method name=\"SetCurrentIM\">\n"
+    "      <arg name=\"im\" direction=\"in\" type=\"s\"/>\n"
+    "    </method>\n"
+    "    <method name=\"ReloadConfig\">\n"
+    "    </method>\n"
+    "    <method name=\"Restart\">\n"
+    "    </method>\n"
+    "    <method name=\"Configure\">\n"
+    "    </method>\n"
+    "    <method name=\"ConfigureAddon\">\n"
+    "      <arg name=\"addon\" direction=\"in\" type=\"s\"/>\n"
+    "    </method>\n"
+    "    <method name=\"GetIMAddon\">\n"
+    "      <arg name=\"im\" direction=\"in\" type=\"s\"/>\n"
+    "      <arg name=\"addon\" direction=\"out\" type=\"s\"/>\n"
+    "    </method>\n"
     "    <property access=\"readwrite\" type=\"a(sssb)\" name=\"IMList\">\n"
     "      <annotation name=\"org.freedesktop.DBus.Property.EmitsChangedSignal\" value=\"true\"/>"
     "    </property>\n"
@@ -168,13 +207,29 @@ const char * ic_introspection_xml =
     "    </method>\n"
     "    <method name=\"Reset\">\n"
     "    </method>\n"
+    "    <method name=\"CommitPreedit\">\n"
+    "    </method>\n"
+    "    <method name=\"MouseEvent\">\n"
+    "      <arg name=\"x\" direction=\"in\" type=\"i\"/>\n"
+    "    </method>\n"
     "    <method name=\"SetCursorLocation\">\n"
     "      <arg name=\"x\" direction=\"in\" type=\"i\"/>\n"
     "      <arg name=\"y\" direction=\"in\" type=\"i\"/>\n"
     "    </method>\n"
+    "    <method name=\"SetCursorRect\">\n"
+    "      <arg name=\"x\" direction=\"in\" type=\"i\"/>\n"
+    "      <arg name=\"y\" direction=\"in\" type=\"i\"/>\n"
+    "      <arg name=\"w\" direction=\"in\" type=\"i\"/>\n"
+    "      <arg name=\"h\" direction=\"in\" type=\"i\"/>\n"
+    "    </method>\n"
     "    <method name=\"SetCapacity\">\n"
     "      <arg name=\"caps\" direction=\"in\" type=\"u\"/>\n"
     "    </method>\n"
+    "    <method name=\"SetSurroundingText\">\n"
+    "      <arg name=\"text\" direction=\"in\" type=\"s\"/>\n"
+    "      <arg name=\"cursor\" direction=\"in\" type=\"u\"/>\n"
+    "      <arg name=\"anchor\" direction=\"in\" type=\"u\"/>\n"
+    "    </method>\n"
     "    <method name=\"DestroyIC\">\n"
     "    </method>\n"
     "    <method name=\"ProcessKeyEvent\">\n"
@@ -196,6 +251,10 @@ const char * ic_introspection_xml =
     "      <arg name=\"str\" type=\"s\"/>\n"
     "      <arg name=\"cursorpos\" type=\"i\"/>\n"
     "    </signal>\n"
+    "    <signal name=\"DeleteSurroundingText\">\n"
+    "      <arg name=\"offset\" type=\"i\"/>\n"
+    "      <arg name=\"nchar\" type=\"u\"/>\n"
+    "    </signal>\n"
     "    <signal name=\"UpdateFormattedPreedit\">\n"
     "      <arg name=\"str\" type=\"a(si)\"/>\n"
     "      <arg name=\"cursorpos\" type=\"i\"/>\n"
@@ -228,12 +287,14 @@ FcitxFrontend frontend = {
     IPCCommitString,
     IPCForwardKey,
     IPCSetWindowOffset,
-    IPCGetWindowPosition,
+    IPCGetWindowRect,
     IPCUpdatePreedit,
     IPCUpdateClientSideUI,
     NULL,
     IPCCheckICFromSameApplication,
-    NULL
+    IPCGetPid,
+    IPCDeleteSurroundingText,
+    IPCGetSurroundingText
 };
 
 FCITX_EXPORT_API
@@ -326,6 +387,36 @@ void IPCCreateIC(void* arg, FcitxInputContext* context, void* priv)
 
         boolean arg0 = context->state != IS_CLOSED;
 
+        dbus_error_free(&error);
+        dbus_message_append_args(reply,
+                                 DBUS_TYPE_INT32, &ipcic->id,
+                                 DBUS_TYPE_BOOLEAN, &arg0,
+                                 DBUS_TYPE_UINT32, &arg1,
+                                 DBUS_TYPE_UINT32, &arg2,
+                                 DBUS_TYPE_UINT32, &arg3,
+                                 DBUS_TYPE_UINT32, &arg4,
+                                 DBUS_TYPE_INVALID);
+    } else if (dbus_message_is_method_call(message, FCITX_IM_DBUS_INTERFACE, "CreateICv3")) {
+
+        DBusError error;
+        dbus_error_init(&error);
+        char* appname;
+        int icpid;
+        if (!dbus_message_get_args(message, &error, DBUS_TYPE_STRING, &appname, DBUS_TYPE_INT32, &icpid, DBUS_TYPE_INVALID))
+            ipcic->appname = NULL;
+        else {
+            if (strlen(appname) == 0)
+                ipcic->appname = NULL;
+            else
+                ipcic->appname = strdup(appname);
+        }
+        ipcic->pid = icpid;
+
+        if (config->shareState == ShareState_PerProgram)
+            FcitxInstanceSetICStateFromSameApplication(ipc->owner, ipc->frontendid, context);
+
+        boolean arg0 = context->state != IS_CLOSED;
+
         dbus_error_free(&error);
         dbus_message_append_args(reply,
                                  DBUS_TYPE_INT32, &ipcic->id,
@@ -364,6 +455,8 @@ void IPCDestroyIC(void* arg, FcitxInputContext* context)
     dbus_connection_unregister_object_path(ipc->conn, GetIPCIC(context)->path);
     if (ipcic->appname)
         free(ipcic->appname);
+    if (ipcic->surroundingText)
+        free(ipcic->surroundingText);
     free(context->privateic);
     context->privateic = NULL;
 }
@@ -441,21 +534,24 @@ void IPCSetWindowOffset(void* arg, FcitxInputContext* ic, int x, int y)
     ic->offset_y = y;
 }
 
-void IPCGetWindowPosition(void* arg, FcitxInputContext* ic, int* x, int* y)
+void IPCGetWindowRect(void* arg, FcitxInputContext* ic, int* x, int* y, int* w, int* h)
 {
     *x = ic->offset_x;
     *y = ic->offset_y;
+    *w = GetIPCIC(ic)->width;
+    *h = GetIPCIC(ic)->height;
 }
 
 
 static DBusHandlerResult IPCDBusEventHandler(DBusConnection *connection, DBusMessage *msg, void *user_data)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) user_data;
+    FcitxInstance* instance = ipc->owner;
     if (dbus_message_is_method_call(msg, DBUS_INTERFACE_INTROSPECTABLE, "Introspect")) {
         DBusMessage *reply = dbus_message_new_method_return(msg);
 
         dbus_message_append_args(reply, DBUS_TYPE_STRING, &im_introspection_xml, DBUS_TYPE_INVALID);
-        dbus_connection_send(ipc->conn, reply, NULL);
+        dbus_connection_send(connection, reply, NULL);
         dbus_message_unref(reply);
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, DBUS_INTERFACE_PROPERTIES, "Get")) {
@@ -473,13 +569,89 @@ static DBusHandlerResult IPCDBusEventHandler(DBusConnection *connection, DBusMes
     } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateICv2")) {
         FcitxInstanceCreateIC(ipc->owner, ipc->frontendid, msg);
         return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateICv3")) {
+        FcitxInstanceCreateIC(ipc->owner, ipc->frontendid, msg);
+        return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "Exit")) {
         DBusMessage *reply = dbus_message_new_method_return(msg);
-        dbus_connection_send(ipc->conn, reply, NULL);
+        dbus_connection_send(connection, reply, NULL);
         dbus_message_unref(reply);
-        dbus_connection_flush(ipc->conn);
+        dbus_connection_flush(connection);
         FcitxInstanceEnd(ipc->owner);
         return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "GetCurrentIM")) {
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        FcitxIM* im = FcitxInstanceGetCurrentIM(ipc->owner);
+        const char* name = "";
+        if (im) {
+            name = im->uniqueName;
+        }
+        dbus_message_append_args(reply, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "SetCurrentIM")) {
+        DBusError error;
+        dbus_error_init(&error);
+        char* imname = NULL;
+        if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &imname, DBUS_TYPE_INVALID)) {
+            FcitxInstanceSwitchIMByName(instance, imname);
+            DBusMessage *reply = dbus_message_new_method_return(msg);
+            dbus_connection_send(connection, reply, NULL);
+            dbus_message_unref(reply);
+        }
+        dbus_error_free(&error);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "GetIMAddon")) {
+        DBusError error;
+        dbus_error_init(&error);
+        char* imname = NULL;
+        if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &imname, DBUS_TYPE_INVALID)) {
+            FcitxIM* im = FcitxInstanceGetIMFromIMList(instance, IMAS_Enable, imname);
+            const char* name = "";
+            if (im)
+                name = im->owner->name;
+            DBusMessage *reply = dbus_message_new_method_return(msg);
+            dbus_message_append_args(reply, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
+            dbus_connection_send(connection, reply, NULL);
+            dbus_message_unref(reply);
+        }
+        dbus_error_free(&error);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "Configure")) {
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        fcitx_utils_launch_configure_tool();
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "ConfigureAddon")) {
+        DBusError error;
+        dbus_error_init(&error);
+        char* addonname = NULL;
+        if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &addonname, DBUS_TYPE_INVALID)) {
+            DBusMessage *reply = dbus_message_new_method_return(msg);
+            dbus_connection_send(connection, reply, NULL);
+            dbus_message_unref(reply);
+            if (addonname) {
+                fcitx_utils_launch_configure_tool_for_addon(addonname);
+            }
+        }
+        dbus_error_free(&error);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "ReloadConfig")) {
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        dbus_connection_flush(connection);
+        FcitxInstanceReloadConfig(instance);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "Restart")) {
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        dbus_connection_flush(connection);
+        fcitx_utils_launch_restart();
+        return DBUS_HANDLER_RESULT_HANDLED;
     }
     return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
@@ -497,50 +669,74 @@ static DBusHandlerResult IPCICDBusEventHandler(DBusConnection *connection, DBusM
         DBusMessage *reply = dbus_message_new_method_return(msg);
 
         dbus_message_append_args(reply, DBUS_TYPE_STRING, &ic_introspection_xml, DBUS_TYPE_INVALID);
-        dbus_connection_send(ipc->conn, reply, NULL);
+        dbus_connection_send(connection, reply, NULL);
         dbus_message_unref(reply);
         result = DBUS_HANDLER_RESULT_HANDLED;
     }
 
+    /* ic is not NULL */
     if (result == DBUS_HANDLER_RESULT_NOT_YET_HANDLED && ic) {
         DBusError error;
         dbus_error_init(&error);
         if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "EnableIC")) {
             FcitxInstanceEnableIM(ipc->owner, ic, false);
             DBusMessage *reply = dbus_message_new_method_return(msg);
-            dbus_connection_send(ipc->conn, reply, NULL);
+            dbus_connection_send(connection, reply, NULL);
             dbus_message_unref(reply);
             result = DBUS_HANDLER_RESULT_HANDLED;
         } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "CloseIC")) {
             FcitxInstanceCloseIM(ipc->owner, ic);
             DBusMessage *reply = dbus_message_new_method_return(msg);
-            dbus_connection_send(ipc->conn, reply, NULL);
+            dbus_connection_send(connection, reply, NULL);
             dbus_message_unref(reply);
             result = DBUS_HANDLER_RESULT_HANDLED;
         } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "FocusIn")) {
             IPCICFocusIn(ipc, ic);
             DBusMessage *reply = dbus_message_new_method_return(msg);
-            dbus_connection_send(ipc->conn, reply, NULL);
+            dbus_connection_send(connection, reply, NULL);
             dbus_message_unref(reply);
             result = DBUS_HANDLER_RESULT_HANDLED;
         } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "FocusOut")) {
             IPCICFocusOut(ipc, ic);
             DBusMessage *reply = dbus_message_new_method_return(msg);
-            dbus_connection_send(ipc->conn, reply, NULL);
+            dbus_connection_send(connection, reply, NULL);
             dbus_message_unref(reply);
             result = DBUS_HANDLER_RESULT_HANDLED;
         } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "Reset")) {
             IPCICReset(ipc, ic);
             DBusMessage *reply = dbus_message_new_method_return(msg);
-            dbus_connection_send(ipc->conn, reply, NULL);
+            dbus_connection_send(connection, reply, NULL);
+            dbus_message_unref(reply);
+            result = DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "CommitPreedit")) {
+            IPCICCommitPreedit(ipc, ic);
+            DBusMessage *reply = dbus_message_new_method_return(msg);
+            dbus_connection_send(connection, reply, NULL);
+            dbus_message_unref(reply);
+            result = DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "MouseEvent")) {
+            DBusMessage *reply = dbus_message_new_method_return(msg);
+            dbus_connection_send(connection, reply, NULL);
             dbus_message_unref(reply);
             result = DBUS_HANDLER_RESULT_HANDLED;
         } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "SetCursorLocation")) {
             int x, y;
             if (dbus_message_get_args(msg, &error, DBUS_TYPE_INT32, &x, DBUS_TYPE_INT32, &y, DBUS_TYPE_INVALID)) {
-                IPCICSetCursorLocation(ipc, ic, x, y);
+                IPCICSetCursorRect(ipc, ic, x, y, 0, 0);
                 DBusMessage *reply = dbus_message_new_method_return(msg);
-                dbus_connection_send(ipc->conn, reply, NULL);
+                dbus_connection_send(connection, reply, NULL);
+                dbus_message_unref(reply);
+            }
+            result = DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "SetCursorRect")) {
+            int x, y, w, h;
+            if (dbus_message_get_args(msg, &error,
+                DBUS_TYPE_INT32, &x, DBUS_TYPE_INT32, &y,
+                DBUS_TYPE_INT32, &w, DBUS_TYPE_INT32, &h,
+                DBUS_TYPE_INVALID)) {
+                IPCICSetCursorRect(ipc, ic, x, y, w, h);
+                DBusMessage *reply = dbus_message_new_method_return(msg);
+                dbus_connection_send(connection, reply, NULL);
                 dbus_message_unref(reply);
             }
             result = DBUS_HANDLER_RESULT_HANDLED;
@@ -548,15 +744,40 @@ static DBusHandlerResult IPCICDBusEventHandler(DBusConnection *connection, DBusM
             uint32_t flags;
             if (dbus_message_get_args(msg, &error, DBUS_TYPE_UINT32, &flags, DBUS_TYPE_INVALID)) {
                 ic->contextCaps = flags;
+                if (!(ic->contextCaps & CAPACITY_SURROUNDING_TEXT)) {
+                    if (GetIPCIC(ic)->surroundingText)
+                        free(GetIPCIC(ic)->surroundingText);
+                    GetIPCIC(ic)->surroundingText = NULL;
+                }
                 DBusMessage *reply = dbus_message_new_method_return(msg);
-                dbus_connection_send(ipc->conn, reply, NULL);
+                dbus_connection_send(connection, reply, NULL);
                 dbus_message_unref(reply);
             }
             result = DBUS_HANDLER_RESULT_HANDLED;
-        } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "FcitxInstanceDestroyIC")) {
+        } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "SetSurroundingText")) {
+            char* text;
+            uint32_t cursor, anchor;
+            if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &text,  DBUS_TYPE_UINT32, &cursor, DBUS_TYPE_UINT32, &anchor, DBUS_TYPE_INVALID)) {
+                FcitxIPCIC* ipcic = GetIPCIC(ic);
+                if (!ipcic->surroundingText || strcmp(ipcic->surroundingText, text) != 0 || cursor != ipcic->cursor || anchor != ipcic->anchor)
+                {
+                    if (ipcic->surroundingText) {
+                        free(ipcic->surroundingText);
+                    }
+                    ipcic->surroundingText = strdup(text);
+                    ipcic->cursor = cursor;
+                    ipcic->anchor = anchor;
+                    FcitxInstanceNotifyUpdateSurroundingText(ipc->owner, ic);
+                }
+                DBusMessage *reply = dbus_message_new_method_return(msg);
+                dbus_connection_send(connection, reply, NULL);
+                dbus_message_unref(reply);
+            }
+            result = DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "DestroyIC")) {
             FcitxInstanceDestroyIC(ipc->owner, ipc->frontendid, &id);
             DBusMessage *reply = dbus_message_new_method_return(msg);
-            dbus_connection_send(ipc->conn, reply, NULL);
+            dbus_connection_send(connection, reply, NULL);
             dbus_message_unref(reply);
             result = DBUS_HANDLER_RESULT_HANDLED;
         } else if (dbus_message_is_method_call(msg, FCITX_IC_DBUS_INTERFACE, "ProcessKeyEvent")) {
@@ -576,9 +797,9 @@ static DBusHandlerResult IPCICDBusEventHandler(DBusConnection *connection, DBusM
                 DBusMessage *reply = dbus_message_new_method_return(msg);
 
                 dbus_message_append_args(reply, DBUS_TYPE_INT32, &ret, DBUS_TYPE_INVALID);
-                dbus_connection_send(ipc->conn, reply, NULL);
+                dbus_connection_send(connection, reply, NULL);
                 dbus_message_unref(reply);
-                dbus_connection_flush(ipc->conn);
+                dbus_connection_flush(connection);
             }
 
             result = DBUS_HANDLER_RESULT_HANDLED;
@@ -594,8 +815,9 @@ static int IPCProcessKey(FcitxIPCFrontend* ipc, FcitxInputContext* callic, const
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(ipc->owner);
     FcitxInputState* input = FcitxInstanceGetInputState(ipc->owner);
 
-    if (ic == NULL) {
+    if (ic == NULL || ic->frontendid != callic->frontendid || GetIPCIC(ic)->id != GetIPCIC(callic)->id) {
         FcitxInstanceSetCurrentIC(ipc->owner, callic);
+        FcitxUIOnInputFocus(ipc->owner);
     }
     ic = callic;
     FcitxKeySym sym;
@@ -615,6 +837,9 @@ static int IPCProcessKey(FcitxIPCFrontend* ipc, FcitxInputContext* callic, const
         FcitxInstanceEnableIM(ipc->owner, ic, false);
         return 1;
     }
+    else if (ic->state == IS_CLOSED) {
+        return 0;
+    }
 
     FcitxInputStateSetKeyCode(input, keycode);
     FcitxInputStateSetKeySym(input, originsym);
@@ -637,9 +862,15 @@ static void IPCICFocusIn(FcitxIPCFrontend* ipc, FcitxInputContext* ic)
     if (ic == NULL)
         return;
 
+    FcitxInputContext* oldic = FcitxInstanceGetCurrentIC(ipc->owner);
+
+    if (oldic && oldic != ic)
+        FcitxUICommitPreedit(ipc->owner);
+
     if (!FcitxInstanceSetCurrentIC(ipc->owner, ic))
         return;
 
+
     if (ic) {
         FcitxUIOnInputFocus(ipc->owner);
     } else {
@@ -654,9 +885,10 @@ static void IPCICFocusOut(FcitxIPCFrontend* ipc, FcitxInputContext* ic)
 {
     FcitxInputContext* currentic = FcitxInstanceGetCurrentIC(ipc->owner);
     if (ic && ic == currentic) {
+        FcitxUICommitPreedit(ipc->owner);
         FcitxUICloseInputWindow(ipc->owner);
-        FcitxUIOnInputUnFocus(ipc->owner);
         FcitxInstanceSetCurrentIC(ipc->owner, NULL);
+        FcitxUIOnInputUnFocus(ipc->owner);
     }
 
     return;
@@ -673,10 +905,24 @@ static void IPCICReset(FcitxIPCFrontend* ipc, FcitxInputContext* ic)
     return;
 }
 
-static void IPCICSetCursorLocation(FcitxIPCFrontend* ipc, FcitxInputContext* ic, int x, int y)
+static void IPCICCommitPreedit(FcitxIPCFrontend* ipc, FcitxInputContext* ic)
+{
+    FcitxInputContext* currentic = FcitxInstanceGetCurrentIC(ipc->owner);
+    if (ic && ic == currentic) {
+        FcitxUICommitPreedit(ipc->owner);
+        FcitxUICloseInputWindow(ipc->owner);
+        FcitxInstanceResetInput(ipc->owner);
+    }
+
+    return;
+}
+
+static void IPCICSetCursorRect(FcitxIPCFrontend* ipc, FcitxInputContext* ic, int x, int y, int w, int h)
 {
     ic->offset_x = x;
     ic->offset_y = y;
+    GetIPCIC(ic)->width = w;
+    GetIPCIC(ic)->height = h;
     FcitxUIMoveInputWindow(ipc->owner);
 
     return;
@@ -691,9 +937,9 @@ void IPCUpdatePreedit(void* arg, FcitxInputContext* ic)
         DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                         FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                         "UpdateFormattedPreedit"); // name of the signal
-        
+
         FcitxMessages* clientPreedit = FcitxInputStateGetClientPreedit(input);
-        
+
         DBusMessageIter args, array, sub;
         dbus_message_iter_init_append(msg, &args);
         dbus_message_iter_open_container(&args, DBUS_TYPE_ARRAY, "(si)", &array);
@@ -713,7 +959,7 @@ void IPCUpdatePreedit(void* arg, FcitxInputContext* ic)
                 free(needtofree);
         }
         dbus_message_iter_close_container(&args, &array);
-        
+
         int iCursorPos = FcitxInputStateGetClientCursorPos(input);
         dbus_message_iter_append_basic(&args, DBUS_TYPE_INT32, &iCursorPos);
 
@@ -751,6 +997,44 @@ void IPCUpdatePreedit(void* arg, FcitxInputContext* ic)
     }
 }
 
+void IPCDeleteSurroundingText(void* arg, FcitxInputContext* ic, int offset, unsigned int size)
+{
+    FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
+    dbus_uint32_t serial = 0; // unique number to associate replies with requests
+    DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
+                       FCITX_IC_DBUS_INTERFACE, // interface name of the signal
+                       "DeleteSurroundingText"); // name of the signal
+
+    dbus_message_append_args(msg, DBUS_TYPE_INT32, &offset, DBUS_TYPE_UINT32, &size, DBUS_TYPE_INVALID);
+
+    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
+        FcitxLog(DEBUG, "Out Of Memory!");
+    }
+    dbus_connection_flush(ipc->conn);
+    dbus_message_unref(msg);
+}
+
+
+boolean IPCGetSurroundingText(void* arg, FcitxInputContext* ic, char** str, unsigned int *cursor, unsigned int *anchor)
+{
+    FcitxIPCIC* ipcic = GetIPCIC(ic);
+
+    if (!ipcic->surroundingText)
+        return false;
+
+    if (str)
+        *str = strdup(ipcic->surroundingText);
+
+    if (cursor)
+        *cursor = ipcic->cursor;
+
+    if (anchor)
+        *anchor = ipcic->anchor;
+
+    return true;
+}
+
+
 void IPCUpdateClientSideUI(void* arg, FcitxInputContext* ic)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
@@ -1062,6 +1346,9 @@ void IPCUpdateIMList(void* arg)
     dbus_message_unref(msg);
 }
 
-
+pid_t IPCGetPid(void* arg, FcitxInputContext* ic)
+{
+    return GetIPCIC(ic)->pid;
+}
 
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
