@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2011~2011 by CSSlayer                                   *
+ *   Copyright (C) 2011~2012 by CSSlayer                                   *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -32,7 +32,7 @@
 #include "fcitx-config/hotkey.h"
 #include "module/dbus/dbusstuff.h"
 #include "frontend/ipc/ipc.h"
-#include "fcitx-compose-data.h"
+#include "im/keyboard/fcitx-compose-data.h"
 #include "fcitx-input-context.h"
 #include "keyserver_x11.h"
 
@@ -42,6 +42,7 @@
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/Xatom.h>
+#include <unistd.h>
 static const int XKeyPress = KeyPress;
 static const int XKeyRelease = KeyRelease;
 #undef KeyPress
@@ -50,6 +51,14 @@ static const int XKeyRelease = KeyRelease;
 #undef FocusOut
 #endif
 
+#ifndef Q_LIKELY
+#define Q_LIKELY(x) (x)
+#endif
+
+#ifndef Q_UNLIKELY
+#define Q_UNLIKELY(x) (x)
+#endif
+
 typedef struct _FcitxComposeTableCompact FcitxComposeTableCompact;
 struct _FcitxComposeTableCompact {
     const quint32 *data;
@@ -96,6 +105,25 @@ compare_seq(const void *key, const void *value)
     return 0;
 }
 
+static boolean
+get_boolean_env(const char *name,
+                 boolean defval)
+{
+    const char *value = getenv(name);
+
+    if (value == NULL)
+        return defval;
+
+    if (strcmp(value, "") == 0 ||
+        strcmp(value, "0") == 0 ||
+        strcmp(value, "false") == 0 ||
+        strcmp(value, "False") == 0 ||
+        strcmp(value, "FALSE") == 0)
+        return false;
+
+    return true;
+}
+
 static const uint fcitx_compose_ignore[] = {
     FcitxKey_Shift_L,
     FcitxKey_Shift_R,
@@ -120,18 +148,6 @@ typedef QInputMethodEvent::Attribute QAttribute;
 
 static bool key_filtered = false;
 
-static boolean FcitxIsHotKey(FcitxKeySym sym, int state, FcitxHotkey * hotkey);
-
-boolean FcitxIsHotKey(FcitxKeySym sym, int state, FcitxHotkey * hotkey)
-{
-    state &= FcitxKeyState_Ctrl_Alt_Shift;
-    if (hotkey[0].sym && sym == hotkey[0].sym && (hotkey[0].state == state))
-        return true;
-    if (hotkey[1].sym && sym == hotkey[1].sym && (hotkey[1].state == state))
-        return true;
-    return false;
-}
-
 QFcitxInputContext::QFcitxInputContext()
     : m_connection(QDBusConnection::sessionBus()),
       m_dbusproxy(0),
@@ -139,34 +155,18 @@ QFcitxInputContext::QFcitxInputContext()
       m_icproxy(0),
       m_id(0),
       m_path(""),
-      m_enable(false),
       m_has_focus(false),
-      m_slave(0),
-      m_n_compose(0)
+      m_n_compose(0),
+      m_cursorPos(0),
+      m_useSurroundingText(false)
 {
-#if defined(Q_WS_X11)
-    /* slave has too much limitation, ibus compose by hand is better, so m_slave will be NULL then */
-#if 0
-    m_slave = QInputContextFactory::create("xims", 0);
-#endif
-#endif
-
     FcitxFormattedPreedit::registerMetaType();
 
     memset(m_compose_buffer, 0, sizeof(uint) * (FCITX_MAX_COMPOSE_LEN + 1));
 
-    if (m_slave) {
-        qDebug() << "slave created";
-        m_slave->setParent(this);
-        connect(m_slave, SIGNAL(destroyed(QObject*)), this, SLOT(destroySlaveContext()));
-    }
-
     m_dbusproxy = new org::freedesktop::DBus("org.freedesktop.DBus", "/org/freedesktop/DBus", m_connection, this);
     connect(m_dbusproxy, SIGNAL(NameOwnerChanged(QString, QString, QString)), this, SLOT(imChanged(QString, QString, QString)));
 
-    m_triggerKey[0].sym = m_triggerKey[1].sym = (FcitxKeySym) 0;
-    m_triggerKey[0].state = m_triggerKey[1].state = 0;
-
     m_serviceName = QString("%1-%2").arg(FCITX_DBUS_SERVICE).arg(fcitx_utils_get_display_number());
 
     createInputContext();
@@ -196,12 +196,21 @@ QString QFcitxInputContext::language()
     return "";
 }
 
+void QFcitxInputContext::commitPreedit()
+{
+    if (m_commitPreedit.length() > 0) {
+        QInputMethodEvent e;
+        e.setCommitString(m_commitPreedit);
+        m_commitPreedit.clear();
+        sendEvent(e);
+    }
+}
+
 void QFcitxInputContext::reset()
 {
+    commitPreedit();
     if (isValid())
         m_icproxy->Reset();
-    if (m_slave)
-        m_slave->reset();
 }
 
 void QFcitxInputContext::update()
@@ -211,12 +220,35 @@ void QFcitxInputContext::update()
         return;
     }
 
+    if (m_useSurroundingText) {
+        QVariant var = widget->inputMethodQuery(Qt::ImSurroundingText);
+        QVariant var1 = widget->inputMethodQuery(Qt::ImCursorPosition);
+        QVariant var2 = widget->inputMethodQuery(Qt::ImAnchorPosition);
+        if (var.isValid() && var1.isValid() && !m_capacity.testFlag(CAPACITY_PASSWORD) ) {
+            addCapacity(CAPACITY_SURROUNDING_TEXT);
+            QString text = var.toString();
+            int cursor = var1.toInt();
+            int anchor;
+            if (var2.isValid())
+                anchor = var2.toInt();
+            else
+                anchor = cursor;
+            m_icproxy->SetSurroundingText(text, cursor, anchor);
+        }
+        else {
+            removeCapacity(CAPACITY_SURROUNDING_TEXT);
+        }
+    }
+
     QRect rect = widget->inputMethodQuery(Qt::ImMicroFocus).toRect();
 
     QPoint topleft = widget->mapToGlobal(QPoint(0, 0));
     rect.translate(topleft);
 
-    m_icproxy->SetCursorLocation(rect.x(), rect.y() + rect.height());
+    if (m_rect != rect) {
+        m_rect = rect;
+        m_icproxy->SetCursorRect(rect.x(), rect.y(), rect.width(), rect.height());
+    }
 }
 
 
@@ -290,6 +322,20 @@ bool QFcitxInputContext::filterEvent(const QEvent* event)
 
 }
 
+void QFcitxInputContext::mouseHandler(int x, QMouseEvent* event)
+{
+    if ((event->type() == QEvent::MouseButtonPress
+        || event->type() == QEvent::MouseButtonRelease)
+        && (x <= 0 || x >= m_preedit.length())
+    )
+    {
+        commitPreedit();
+        if (isValid())
+            m_icproxy->Reset();
+    }
+}
+
+
 QKeyEvent* QFcitxInputContext::createKeyEvent(uint keyval, uint state, int type)
 {
     Q_UNUSED(keyval);
@@ -340,9 +386,6 @@ void QFcitxInputContext::setFocusWidget(QWidget* w)
     }
 
     QInputContext::setFocusWidget(w);
-    if (m_slave) {
-        m_slave->setFocusWidget(w);
-    }
 
     bool has_focus = (w != NULL);
 
@@ -360,8 +403,6 @@ void QFcitxInputContext::setFocusWidget(QWidget* w)
 void QFcitxInputContext::widgetDestroyed(QWidget* w)
 {
     QInputContext::widgetDestroyed(w);
-    if (m_slave)
-        m_slave->widgetDestroyed(w);
     if (!isValid())
         return;
     if (w == focusWidget())
@@ -388,8 +429,11 @@ bool QFcitxInputContext::x11FilterEvent(QWidget* keywidget, XEvent* event)
     else
         removeCapacity(CAPACITY_PASSWORD);
 
-    if (!isValid() || (event->type != XKeyRelease && event->type != XKeyPress)) {
-        return x11FilterEventFallback(keywidget, event, 0);
+    if (Q_UNLIKELY(event->xkey.state & FcitxKeyState_IgnoredMask))
+        return false;
+
+    if (Q_UNLIKELY(!isValid() || (event->type != XKeyRelease && event->type != XKeyPress))) {
+        return x11FilterEventFallback(event, 0);
     }
 
     KeySym sym = 0;
@@ -397,18 +441,6 @@ bool QFcitxInputContext::x11FilterEvent(QWidget* keywidget, XEvent* event)
     memset(strbuf, 0, 64);
     XLookupString(&event->xkey, strbuf, 64, &sym, NULL);
 
-
-    if (!m_enable) {
-        FcitxKeySym fcitxsym;
-        uint fcitxstate;
-        FcitxHotkeyGetKey((FcitxKeySym) sym, event->xkey.state, &fcitxsym, &fcitxstate);
-        if (!FcitxIsHotKey(fcitxsym, fcitxstate, m_triggerKey)) {
-            return x11FilterEventFallback(keywidget, event, sym);
-        }
-    }
-
-    m_icproxy->FocusIn();
-
     QDBusPendingReply< int > result = this->m_icproxy->ProcessKeyEvent(
                                           sym,
                                           event->xkey.keycode,
@@ -416,34 +448,58 @@ bool QFcitxInputContext::x11FilterEvent(QWidget* keywidget, XEvent* event)
                                           (event->type == XKeyPress) ? FCITX_PRESS_KEY : FCITX_RELEASE_KEY,
                                           event->xkey.time
                                       );
-    {
+    if (Q_LIKELY(m_syncMode)) {
         QEventLoop loop;
         QDBusPendingCallWatcher watcher(result);
         loop.connect(&watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(quit()));
-        loop.exec(QEventLoop::ExcludeUserInputEvents | QEventLoop::WaitForMoreEvents);
-    }
+        loop.exec(QEventLoop::ExcludeUserInputEvents);
 
-    if (result.isError() || result.value() <= 0) {
-        return x11FilterEventFallback(keywidget, event, sym);
-    } else {
-        update();
+        if (result.isError() || result.value() <= 0) {
+            QTimer::singleShot(0, this, SLOT(updateIM()));
+            return x11FilterEventFallback(event, sym);
+        } else {
+            update();
+            return true;
+        }
+        return false;
+    }
+    else {
+        ProcessKeyWatcher* watcher = new ProcessKeyWatcher(event, sym, result);
+        connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(x11ProcessKeyEventCallback(QDBusPendingCallWatcher*)));
         return true;
     }
-    return false;
 }
 
-bool QFcitxInputContext::x11FilterEventFallback(QWidget *keywidget, XEvent *event, KeySym sym)
+void QFcitxInputContext::x11ProcessKeyEventCallback(QDBusPendingCallWatcher* watcher)
 {
-    if (m_slave && m_slave->x11FilterEvent(keywidget, event))
-        return true;
-    else {
-        if (event->type == XKeyPress || event->type == XKeyRelease) {
-            if (processCompose(sym, event->xkey.state, (event->type == XKeyPress) ? FCITX_PRESS_KEY : FCITX_RELEASE_KEY)) {
-                return true;
-            }
-        }
-        return QInputContext::x11FilterEvent(keywidget, event);
+    ProcessKeyWatcher* pkwatcher = static_cast<ProcessKeyWatcher*>(watcher);
+    QDBusPendingReply< int > result(*watcher);
+    bool r = false;
+    if (result.isError() || result.value() <= 0) {
+        QTimer::singleShot(0, this, SLOT(updateIM()));
+        r = x11FilterEventFallback(pkwatcher->event, pkwatcher->sym);
+    } else {
+        update();
+        r = true;
     }
+    if (r) {
+        free(pkwatcher->event);
+        delete pkwatcher;
+    }
+    else {
+        pkwatcher->event->xkey.state |= FcitxKeyState_IgnoredMask;
+        QTimer::singleShot(0, pkwatcher, SLOT(processEvent()));
+    }
+}
+
+bool QFcitxInputContext::x11FilterEventFallback(XEvent *event, KeySym sym)
+{
+    if (event->type == XKeyPress || event->type == XKeyRelease) {
+        if (processCompose(sym, event->xkey.state, (event->type == XKeyPress) ? FCITX_PRESS_KEY : FCITX_RELEASE_KEY)) {
+            return true;
+        }
+    }
+    return false;
 }
 
 #endif // Q_WS_X11
@@ -463,9 +519,6 @@ void QFcitxInputContext::imChanged(const QString& service, const QString& oldown
                 delete m_icproxy;
                 m_icproxy = NULL;
             }
-            m_enable = false;
-            m_triggerKey[0].sym = m_triggerKey[1].sym = (FcitxKeySym) 0;
-            m_triggerKey[0].state = m_triggerKey[1].state = 0;
         }
 
         /* new rise */
@@ -476,13 +529,14 @@ void QFcitxInputContext::imChanged(const QString& service, const QString& oldown
 
 void QFcitxInputContext::createInputContext()
 {
+    m_rect = QRect();
     m_improxy = new org::fcitx::Fcitx::InputMethod(m_serviceName, FCITX_IM_DBUS_PATH, m_connection, this);
 
     if (!m_improxy->isValid())
         return;
 
     char* name = fcitx_utils_get_process_name();
-    QDBusPendingReply< int, bool, uint, uint, uint, uint > result = m_improxy->CreateICv2(name);
+    QDBusPendingReply< int, bool, uint, uint, uint, uint > result = m_improxy->CreateICv3(name, getpid());
     free(name);
     QDBusPendingCallWatcher* watcher = new QDBusPendingCallWatcher(result);
     connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), this, SLOT(createInputContextFinished(QDBusPendingCallWatcher*)));
@@ -495,19 +549,12 @@ void QFcitxInputContext::createInputContextFinished(QDBusPendingCallWatcher* wat
         qWarning() << result.error();
     else {
         this->m_id = qdbus_cast<int>(result.argumentAt(0));
-        this->m_enable = qdbus_cast<bool>(result.argumentAt(1));
-        m_triggerKey[0].sym = (FcitxKeySym) qdbus_cast<uint>(result.argumentAt(2));
-        m_triggerKey[0].state = qdbus_cast<uint>(result.argumentAt(3));
-        m_triggerKey[1].sym = (FcitxKeySym) qdbus_cast<uint>(result.argumentAt(4));
-        m_triggerKey[1].state = qdbus_cast<uint>(result.argumentAt(5));
         this->m_path = QString(FCITX_IC_DBUS_PATH_QSTRING).arg(m_id);
         m_icproxy = new org::fcitx::Fcitx::InputContext(m_serviceName, m_path, m_connection, this);
-        connect(m_icproxy, SIGNAL(CloseIM()), this, SLOT(closeIM()));
         connect(m_icproxy, SIGNAL(CommitString(QString)), this, SLOT(commitString(QString)));
-        connect(m_icproxy, SIGNAL(EnableIM()), this, SLOT(enableIM()));
         connect(m_icproxy, SIGNAL(ForwardKey(uint, uint, int)), this, SLOT(forwardKey(uint, uint, int)));
-        connect(m_icproxy, SIGNAL(UpdatePreedit(QString, int)), this, SLOT(updatePreedit(QString, int)));
         connect(m_icproxy, SIGNAL(UpdateFormattedPreedit(FcitxFormattedPreeditList,int)), this, SLOT(updateFormattedPreedit(FcitxFormattedPreeditList,int)));
+        connect(m_icproxy, SIGNAL(DeleteSurroundingText(int,uint)), this, SLOT(deleteSurroundingText(int,uint)));
 
         if (m_icproxy->isValid() && focusWidget() != NULL)
             m_icproxy->FocusIn();
@@ -515,7 +562,13 @@ void QFcitxInputContext::createInputContextFinished(QDBusPendingCallWatcher* wat
         QFlags<FcitxCapacityFlags> flag;
         flag |= CAPACITY_PREEDIT;
         flag |= CAPACITY_FORMATTED_PREEDIT;
-        
+        flag |= CAPACITY_CLIENT_UNFOCUS_COMMIT;
+        m_useSurroundingText = get_boolean_env("FCITX_QT_ENABLE_SURROUNDING_TEXT", false);
+        if (m_useSurroundingText)
+            flag |= CAPACITY_SURROUNDING_TEXT;
+
+        m_syncMode = get_boolean_env("FCITX_QT_USE_SYNC", true);
+
         addCapacity(flag, true);
     }
     delete watcher;
@@ -525,23 +578,8 @@ void QFcitxInputContext::updateCapacity()
 {
     if (!m_icproxy || !m_icproxy->isValid())
         return;
-    
+
     QDBusPendingReply< void > result = m_icproxy->SetCapacity((uint) m_capacity);
-    
-    QEventLoop loop;
-    QDBusPendingCallWatcher watcher(result);
-    loop.connect(&watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(quit()));
-    loop.exec(QEventLoop::ExcludeUserInputEvents | QEventLoop::WaitForMoreEvents);
-}
-
-void QFcitxInputContext::closeIM()
-{
-    this->m_enable = false;
-}
-
-void QFcitxInputContext::enableIM()
-{
-    this->m_enable = true;
 }
 
 void QFcitxInputContext::commitString(const QString& str)
@@ -552,30 +590,18 @@ void QFcitxInputContext::commitString(const QString& str)
     update();
 }
 
-void QFcitxInputContext::updatePreedit(const QString& str, int cursorPos)
-{
-    QByteArray array = str.toUtf8();
-    array.truncate(cursorPos);
-    cursorPos = QString::fromUtf8(array).length();
-
-    QList<QAttribute> attrList;
-    QTextCharFormat format;
-    format.setUnderlineStyle(QTextCharFormat::DashUnderline);
-    attrList.append(QAttribute(QInputMethodEvent::Cursor, cursorPos, 1, 0));
-    attrList.append(QAttribute(QInputMethodEvent::TextFormat, 0, str.length(), format));
-    QInputMethodEvent event(str, attrList);
-    sendEvent(event);
-    update();
-}
-
 void QFcitxInputContext::updateFormattedPreedit(const FcitxFormattedPreeditList& preeditList, int cursorPos)
 {
-    QString str;
+    m_preeditList = preeditList;
+    m_cursorPos = cursorPos;
+    QString str, commitStr;
     int pos = 0;
     QList<QAttribute> attrList;
     Q_FOREACH(const FcitxFormattedPreedit& preedit, preeditList)
     {
         str += preedit.string();
+        if (!(preedit.format() & MSG_DONOT_COMMIT_WHEN_UNFOCUS))
+            commitStr += preedit.string();
         QTextCharFormat format;
         if ((preedit.format() & MSG_NOUNDERLINE) == 0) {
             format.setUnderlineStyle(QTextCharFormat::DashUnderline);
@@ -593,28 +619,37 @@ void QFcitxInputContext::updateFormattedPreedit(const FcitxFormattedPreeditList&
         attrList.append(QAttribute(QInputMethodEvent::TextFormat, pos, preedit.string().length(), format));
         pos += preedit.string().length();
     }
-    
+
     QByteArray array = str.toUtf8();
     array.truncate(cursorPos);
     cursorPos = QString::fromUtf8(array).length();
 
     attrList.append(QAttribute(QInputMethodEvent::Cursor, cursorPos, 1, 0));
     QInputMethodEvent event(str, attrList);
+    m_preedit = str;
+    m_commitPreedit = commitStr;
     sendEvent(event);
     update();
 }
 
+void QFcitxInputContext::deleteSurroundingText(int offset, uint nchar)
+{
+    QInputMethodEvent event;
+    event.setCommitString("", offset, nchar);
+    sendEvent(event);
+    update();
+}
 
 void QFcitxInputContext::forwardKey(uint keyval, uint state, int type)
 {
     QWidget* widget = focusWidget();
-    if (widget != NULL) {
+    if (Q_LIKELY(widget != NULL)) {
         key_filtered = true;
 #ifdef Q_WS_X11
         const WId window_id = widget->winId();
         Display* x11_display = QX11Info::display();
 
-        XEvent* xevent = createXEvent(x11_display, window_id, keyval, state, type);
+        XEvent* xevent = createXEvent(x11_display, window_id, keyval, state | FcitxKeyState_IgnoredMask, type);
         qApp->x11ProcessEvent(xevent);
         free(xevent);
 #else
@@ -663,14 +698,6 @@ bool QFcitxInputContext::isValid()
     return m_icproxy != NULL && m_icproxy->isValid();
 }
 
-void QFcitxInputContext::destroySlaveContext()
-{
-    if (m_slave) {
-        m_slave->deleteLater();
-        m_slave = 0;
-    }
-}
-
 bool
 QFcitxInputContext::processCompose(uint keyval, uint state, FcitxKeyEventType event)
 {
@@ -869,4 +896,10 @@ QFcitxInputContext::checkCompactTable(const FcitxComposeTableCompact *table)
     }
     return false;
 }
+
+void QFcitxInputContext::updateIM()
+{
+    update();
+}
+
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
