@@ -39,6 +39,7 @@
 #include "IC.h"
 #include "xim.h"
 #include "ximhandler.h"
+#include "ximqueue.h"
 #include "module/x11/x11stuff.h"
 #include "fcitx-config/xdg.h"
 #include "fcitx/hook.h"
@@ -50,8 +51,10 @@ static void XimCloseIM(void* arg, FcitxInputContext* ic);
 static void XimCommitString(void* arg, FcitxInputContext* ic, const char* str);
 static void XimForwardKey(void* arg, FcitxInputContext* ic, FcitxKeyEventType event, FcitxKeySym sym, unsigned int state);
 static void XimSetWindowOffset(void* arg, FcitxInputContext* ic, int x, int y);
-static void XimGetWindowPosition(void* arg, FcitxInputContext* ic, int* x, int* y);
+static void XimGetWindowRect(void* arg, FcitxInputContext* ic, int* x, int* y, int* w, int* h);
 static void XimUpdatePreedit(void* arg, FcitxInputContext* ic);
+// static pid_t XimGetPid(void* arg, FcitxInputContext* ic);
+// static pid_t XimFindApplicationPid(FcitxXimFrontend* xim, Window w);
 
 static Bool XimProtocolHandler(XIMS _ims, IMProtocol * call_data);
 
@@ -90,11 +93,13 @@ FcitxFrontend frontend = {
     XimCommitString,
     XimForwardKey,
     XimSetWindowOffset,
-    XimGetWindowPosition,
+    XimGetWindowRect,
     XimUpdatePreedit,
     NULL,
     NULL,
     XimCheckICFromSameApplication,
+    NULL,
+    NULL,
     NULL
 };
 
@@ -137,6 +142,8 @@ void* XimCreate(FcitxInstance* instance, int frontendid)
         return NULL;
     }
 
+    FcitxAddon* ximaddon = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(instance), "fcitx-xim");
+    xim->x11addon = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(instance), "fcitx-x11");
     xim->iScreen = DefaultScreen(xim->display);
     xim->owner = instance;
     xim->frontendid = frontendid;
@@ -162,6 +169,7 @@ void* XimCreate(FcitxInstance* instance, int frontendid)
             imname = DEFAULT_IMNAME;
         }
     }
+    XimQueueInit(xim);
 
     if (GetXimConfigDesc() == NULL)
         xim->bUseOnTheSpotStyle = false;
@@ -240,6 +248,9 @@ void* XimCreate(FcitxInstance* instance, int frontendid)
         free(xim);
         return NULL;
     }
+
+    AddFunction(ximaddon, XimConsumeQueue);
+
     return xim;
 }
 
@@ -318,8 +329,6 @@ Bool XimProtocolHandler(XIMS _ims, IMProtocol * call_data)
         return XIMUnsetFocusHandler(ximfrontend, (IMChangeICStruct *) call_data);;
     case XIM_RESET_IC:
         return True;
-    case XIM_TRIGGER_NOTIFY:
-        return XIMTriggerNotifyHandler(ximfrontend, (IMTriggerNotifyStruct *) call_data);
     case XIM_PREEDIT_START_REPLY:
         return 0;
     case XIM_PREEDIT_CARET_REPLY:
@@ -335,8 +344,10 @@ boolean XimDestroy(void* arg)
 {
     FcitxXimFrontend* xim = (FcitxXimFrontend*) arg;
 
-    if (xim->ims)
+    if (xim->ims) {
         IMCloseIM(xim->ims);
+        xim->ims = NULL;
+    }
 
     return true;
 }
@@ -344,55 +355,43 @@ boolean XimDestroy(void* arg)
 void XimEnableIM(void* arg, FcitxInputContext* ic)
 {
     FcitxXimFrontend* xim = (FcitxXimFrontend*) arg;
-    IMChangeFocusStruct call_data;
     FcitxXimIC* ximic = (FcitxXimIC*) ic->privateic;
-    call_data.connect_id = ximic->connect_id;
-    call_data.icid = ximic->id;
-    IMPreeditStart(xim->ims, (XPointer) &call_data);
+    IMChangeFocusStruct* call_data = fcitx_utils_new(IMChangeFocusStruct);
+    call_data->connect_id = ximic->connect_id;
+    call_data->icid = ximic->id;
+    XimPendingCall(xim, XCT_PREEDIT_START, (XPointer) call_data);
 }
 
 void XimCloseIM(void* arg, FcitxInputContext* ic)
 {
     FcitxXimFrontend* xim = (FcitxXimFrontend*) arg;
-    IMChangeFocusStruct call_data;
+    IMChangeFocusStruct* call_data = fcitx_utils_new(IMChangeFocusStruct);
     FcitxXimIC* ximic = (FcitxXimIC*) ic->privateic;
-    call_data.connect_id = ximic->connect_id;
-    call_data.icid = ximic->id;
-    IMPreeditEnd(xim->ims, (XPointer) &call_data);
+    call_data->connect_id = ximic->connect_id;
+    call_data->icid = ximic->id;
+    XimPendingCall(xim, XCT_PREEDIT_END, (XPointer) call_data);
 }
 
 void XimCommitString(void* arg, FcitxInputContext* ic, const char* str)
 {
     FcitxXimFrontend* xim = (FcitxXimFrontend*) arg;
     XTextProperty tp;
-    IMCommitStruct cms;
     FcitxXimIC* ximic = (FcitxXimIC*) ic->privateic;
 
     /* avoid Seg fault */
     if (!ic)
         return;
 
-    /*
-     * I'm not sure whether xim should commit string before preedit done
-     * but this can fix opera's crash in specific input box
-     * quite strange.
-     */
-    if (GetXimIC(ic)->bPreeditStarted == true) {
-        XimPreeditCallbackDraw(xim, GetXimIC(ic), "", 0);
-        XimPreeditCallbackDone(xim, GetXimIC(ic));
-        GetXimIC(ic)->bPreeditStarted = false;
-    }
 
     Xutf8TextListToTextProperty(xim->display, (char **) &str, 1, XCompoundTextStyle, &tp);
+    IMCommitStruct* cms = fcitx_utils_new(IMCommitStruct);
 
-    memset(&cms, 0, sizeof(IMCommitStruct));
-    cms.major_code = XIM_COMMIT;
-    cms.icid = ximic->id;
-    cms.connect_id = ximic->connect_id;
-    cms.flag = XimLookupChars;
-    cms.commit_string = (char *) tp.value;
-    IMCommitString(xim->ims, (XPointer) & cms);
-    XFree(tp.value);
+    cms->major_code = XIM_COMMIT;
+    cms->icid = ximic->id;
+    cms->connect_id = ximic->connect_id;
+    cms->flag = XimLookupChars;
+    cms->commit_string = (char *) tp.value;
+    XimPendingCall(xim, XCT_COMMIT, (XPointer) cms);
 }
 
 void XimForwardKey(void *arg, FcitxInputContext* ic, FcitxKeyEventType event, FcitxKeySym sym, unsigned int state)
@@ -438,10 +437,12 @@ void XimSetWindowOffset(void* arg, FcitxInputContext* ic, int x, int y)
     }
 }
 
-void XimGetWindowPosition(void* arg, FcitxInputContext* ic, int* x, int* y)
+void XimGetWindowRect(void* arg, FcitxInputContext* ic, int* x, int* y, int* w, int* h)
 {
     *x = ic->offset_x;
     *y = ic->offset_y;
+    *w = 0;
+    *h = 0;
 }
 
 void XimUpdatePreedit(void* arg, FcitxInputContext* ic)
@@ -466,10 +467,63 @@ void XimUpdatePreedit(void* arg, FcitxInputContext* ic)
         GetXimIC(ic)->bPreeditStarted = true;
     }
     if (strlen(strPreedit) != 0) {
-        XimPreeditCallbackDraw(xim, GetXimIC(ic), strPreedit, FcitxInputStateGetCursorPos(input));
+        XimPreeditCallbackDraw(xim, GetXimIC(ic), strPreedit, FcitxInputStateGetClientCursorPos(input));
     }
 
     free(strPreedit);
 }
 
+#if 0
+pid_t XimGetPid(void* arg, FcitxInputContext* ic)
+{
+    FcitxXimFrontend* xim = (FcitxXimFrontend*) arg;
+    Window w = None;
+    FcitxXimIC* ximic = GetXimIC(ic);
+    if (ximic->focus_win)
+        w = ximic->focus_win;
+    else if (ximic->client_win)
+        w = ximic->client_win;
+
+    return XimFindApplicationPid(xim, w);
+}
+
+pid_t XimFindApplicationPid(FcitxXimFrontend* xim, Window w) {
+    if (w == DefaultRootWindow(xim->display))
+        return 0;
+
+    Atom actual_type;
+    int actual_format;
+    unsigned long nitems;
+    unsigned long bytes;
+    unsigned char* prop;
+    int status = XGetWindowProperty(
+        xim->display, w, XInternAtom(xim->display, "_NET_WM_PID", True), 0,
+        1024L, False, AnyPropertyType, &actual_type, &actual_format, &nitems,
+        &bytes, (unsigned char**) &prop);
+    if (status != 0) {
+        if (status == BadRequest)
+            return 0;
+        return 0;
+    }
+    if (!prop) {
+        Window parent;
+        Window root;
+        Window* children = NULL;
+        unsigned int sz = 0;
+        status = XQueryTree(xim->display, w, &root, &parent, &children, &sz);
+        if (status != 0) {
+            if (status == BadRequest)
+                return 0;
+            return 0;
+        }
+        if (children)
+            XFree(children);
+        return XimFindApplicationPid(xim, parent);
+    } else {
+        // TODO: is this portable?
+        return prop[1] * 256 + prop[0];
+    }
+}
+#endif
+
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
