@@ -27,6 +27,7 @@
 #include <unistd.h>
 #include <sys/stat.h>
 #include <assert.h>
+#include <ctype.h>
 
 #include "fcitx/fcitx.h"
 #include "fcitx-utils/utils.h"
@@ -50,8 +51,8 @@
 #include "fcitx/context.h"
 #include "spdata.h"
 
-#define TEMP_FILE       "FCITX_DICT_TEMP"
 #define PY_INDEX_MAGIC_NUMBER 0xf7462e34
+#define PINYIN_TEMP_FILE "pinyin_XXXXXX"
 
 FCITX_EXPORT_API
 FcitxIMClass ime = {
@@ -91,6 +92,7 @@ void *PYCreate(FcitxInstance* instance)
     InitMHPY(&pystate->pyconfig.MHPY_C, MHPY_C_TEMPLATE);
     InitMHPY(&pystate->pyconfig.MHPY_S, MHPY_S_TEMPLATE);
     InitPYTable(&pystate->pyconfig);
+    InitPYSplitData(&pystate->pyconfig);
     if (!LoadPYConfig(&pystate->pyconfig)) {
         free(pystate->pyconfig.MHPY_C);
         free(pystate->pyconfig.MHPY_S);
@@ -100,7 +102,7 @@ void *PYCreate(FcitxInstance* instance)
     }
 
     PinyinMigration();
-    
+
     pystate->pool = fcitx_memory_pool_create();
 
     FcitxInstanceRegisterIM(instance,
@@ -256,25 +258,25 @@ void LoadPYPhraseDict(FcitxPinyinState* pystate, FILE *fp, boolean isSystem, boo
         for (k = 0; k < count; k++) {
             if (!isSystem)
                 phrase = (PyPhrase *) fcitx_utils_malloc0(sizeof(PyUsrPhrase));
-            
+
             fread(&iLen, sizeof(int), 1, fp);
-            
+
             if (isSystem)
                 phrase->strMap = (char* ) fcitx_memory_pool_alloc(pystate->pool, sizeof(char) * (iLen + 1));
             else
                 phrase->strMap = (char *) fcitx_utils_malloc0(sizeof(char) * (iLen + 1));
             fread(phrase->strMap, sizeof(char) * iLen, 1, fp);
             phrase->strMap[iLen] = '\0';
-            
+
             fread(&iLen, sizeof(int), 1, fp);
-            
+
             if (isSystem)
                 phrase->strPhrase = (char* ) fcitx_memory_pool_alloc(pystate->pool, sizeof(char) * (iLen + 1));
             else
                 phrase->strPhrase = (char *) fcitx_utils_malloc0(sizeof(char) * (iLen + 1));
             fread(phrase->strPhrase, sizeof(char) * iLen, 1, fp);
             phrase->strPhrase[iLen] = '\0';
-            
+
             fread(&iLen, sizeof(unsigned int), 1, fp);
             phrase->iIndex = iLen;
             if (iLen > pystate->iCounter)
@@ -363,13 +365,13 @@ boolean LoadPYOtherDict(FcitxPinyinState* pystate)
         FcitxStringHashSet *curStr = sset;
         while (curStr) {
             char *filename;
-                
+
             if (strcmp(curStr->name, PY_PHRASE_FILE) != 0
                 && strcmp(curStr->name, PY_USERPHRASE_FILE) != 0
                 && strcmp(curStr->name, PY_SYMBOL_FILE) != 0
-                && strcmp(curStr->name, PY_BASE_FILE) != 0 
+                && strcmp(curStr->name, PY_BASE_FILE) != 0
                 && strcmp(curStr->name, PY_FREQ_FILE) != 0) {
-                
+
                 fp = FcitxXDGGetFileWithPrefix("pinyin", curStr->name, "r", &filename);
                 FcitxLog(INFO, _("Load extra dict: %s"), filename);
                 free(filename);
@@ -378,7 +380,7 @@ boolean LoadPYOtherDict(FcitxPinyinState* pystate)
             }
             curStr = curStr->hh.next;
         }
-        
+
         fcitx_utils_free_string_hash_set(sset);
 
         pystate->iOrigCounter = pystate->iCounter;
@@ -471,30 +473,36 @@ boolean LoadPYOtherDict(FcitxPinyinState* pystate)
     //下面读取特殊符号表
     fp = FcitxXDGGetFileWithPrefix("pinyin", PY_SYMBOL_FILE, "r", NULL);
     if (fp) {
-        char strTxt[256];
-        char str1[MAX_PY_PHRASE_LENGTH * MAX_PY_LENGTH + 1], str2[MAX_PY_PHRASE_LENGTH * UTF8_MAX_LENGTH + 1];
-        char *str;
+        char* buf = NULL, *buf1 = NULL;
+        size_t len;
 
-        for (;;) {
-            if (!fgets(strTxt, 255, fp))
-                break;
-            i = strlen(strTxt) - 1;
-            if (strTxt[0] == '#')
+        while (getline(&buf, &len, fp) != -1) {
+            if (buf1)
+                free(buf1);
+            buf1 = fcitx_utils_trim(buf);
+            char *p = buf1;
+
+            while (*p && !isspace(*p))
+                p ++;
+            if (*p == '\0' || *p == '#')
                 continue;
-            if (strTxt[i] == '\n')
-                strTxt[i] = '\0';
-            str = strTxt;
-            while (*str == ' ' || *str == '\t')
-                str++;
-            if (!strlen(str))
+
+            while (isspace(*p)) {
+                *p = '\0';
+                p ++;
+            }
+
+            if (strlen(buf1) >= MAX_PY_PHRASE_LENGTH * MAX_PY_LENGTH)
+                continue;
+
+            if (strlen(p) >= MAX_PY_PHRASE_LENGTH * UTF8_MAX_LENGTH)
                 continue;
-            sscanf(str, "%s %s", str1, str2);
 
             //首先看看str1是否已经在列表中
             pyFreqTemp = pystate->pyFreq->next;
             pPyFreq = pystate->pyFreq;
             while (pyFreqTemp) {
-                if (!strcmp(pyFreqTemp->strPY, str1))
+                if (!strcmp(pyFreqTemp->strPY, buf1))
                     break;
                 pPyFreq = pPyFreq->next;
                 pyFreqTemp = pyFreqTemp->next;
@@ -502,7 +510,7 @@ boolean LoadPYOtherDict(FcitxPinyinState* pystate)
 
             if (!pyFreqTemp) {
                 pyFreqTemp = (PyFreq *) fcitx_utils_malloc0(sizeof(PyFreq));
-                strcpy(pyFreqTemp->strPY, str1);
+                strcpy(pyFreqTemp->strPY, buf1);
                 pyFreqTemp->next = NULL;
                 pyFreqTemp->iCount = 0;
                 pyFreqTemp->bIsSym = true;
@@ -516,7 +524,7 @@ boolean LoadPYOtherDict(FcitxPinyinState* pystate)
             }
 
             HZTemp = (HZ *) fcitx_utils_malloc0(sizeof(HZ));
-            strcpy(HZTemp->strHZ, str2);
+            strcpy(HZTemp->strHZ, p);
             HZTemp->next = NULL;
             pyFreqTemp->iCount++;
 
@@ -527,6 +535,11 @@ boolean LoadPYOtherDict(FcitxPinyinState* pystate)
             pHZ->next = HZTemp;
         }
 
+        if (buf)
+            free(buf);
+        if (buf1)
+            free(buf1);
+
         fclose(fp);
     }
 
@@ -611,12 +624,13 @@ INPUT_RETURN_VALUE DoPYInput(void* arg, FcitxKeySym sym, unsigned int state)
             for (i = 0; i < pystate->iPYSelected; i++)
                 val += fcitx_utf8_strlen(pystate->pySelected[i].strHZ);
 
+            retVal = IRV_DISPLAY_CANDWORDS;
             if (pystate->findMap.iHZCount > (MAX_WORDS_USER_INPUT - val)) {
                 UpdateFindString(pystate, val);
                 ParsePY(&pystate->pyconfig, pystate->strFindString, &pystate->findMap, PY_PARSE_INPUT_USER, pystate->bSP);
+                retVal = IRV_DO_NOTHING;
             }
 
-            retVal = IRV_DISPLAY_CANDWORDS;
         } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)) {
             if (pystate->iPYSelected) {
                 char strTemp[MAX_USER_INPUT + 1];
@@ -696,7 +710,7 @@ INPUT_RETURN_VALUE DoPYInput(void* arg, FcitxKeySym sym, unsigned int state)
         } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_LEFT)) {
             if (!FcitxInputStateGetRawInputBufferSize(input))
                 return IRV_TO_PROCESS;
-            if (pystate->iPYInsertPoint < 2) {
+            if (pystate->iPYInsertPoint <= 0) {
                 if (pystate->iPYSelected) {
                     char strTemp[MAX_USER_INPUT + 1];
 
@@ -708,9 +722,6 @@ INPUT_RETURN_VALUE DoPYInput(void* arg, FcitxKeySym sym, unsigned int state)
                     pystate->iPYSelected--;
                     ParsePY(&pystate->pyconfig, pystate->strFindString, &pystate->findMap, PY_PARSE_INPUT_USER, pystate->bSP);
 
-                    retVal = IRV_DISPLAY_CANDWORDS;
-                } else if (pystate->iPYInsertPoint) {
-                    pystate->iPYInsertPoint--;
                     retVal = IRV_DISPLAY_CANDWORDS;
                 } else
                     retVal = IRV_DO_NOTHING;
@@ -924,10 +935,13 @@ void CalculateCursorPosition(FcitxPinyinState* pystate)
     FcitxInputState* input = FcitxInstanceGetInputState(pystate->owner);
 
     int iCursorPos = 0;
+    int hzLen = 0;
 
     for (i = 0; i < pystate->iPYSelected; i++)
         iCursorPos += strlen(pystate->pySelected[i].strHZ);
 
+    hzLen = iCursorPos;
+
     if (pystate->iPYInsertPoint > strlen(pystate->strFindString))
         pystate->iPYInsertPoint = strlen(pystate->strFindString);
     iTemp = pystate->iPYInsertPoint;
@@ -943,7 +957,11 @@ void CalculateCursorPosition(FcitxPinyinState* pystate)
         iTemp -= strlen(pystate->findMap.strPYParsed[i]);
     }
     FcitxInputStateSetCursorPos(input, iCursorPos);
-    FcitxInputStateSetClientCursorPos(input, 0);
+
+    if (pystate->pyconfig.bFixCursorAtHead)
+        FcitxInputStateSetClientCursorPos(input, 0);
+    else
+        FcitxInputStateSetClientCursorPos(input, hzLen);
 }
 
 /*
@@ -1747,8 +1765,16 @@ void SavePYUserPhrase(FcitxPinyinState* pystate)
     FILE *fp;
     PyPhrase *phrase;
     PYFA* PYFAList = pystate->PYFAList;
+    int fd;
+
+    FcitxXDGGetFileUserWithPrefix("pinyin", "", "w", NULL);
+    FcitxXDGGetFileUserWithPrefix("pinyin", PINYIN_TEMP_FILE, NULL, &tempfile);
+    fd = mkstemp(tempfile);
+    fp = NULL;
+
+    if (fd > 0)
+        fp = fdopen(fd, "w");
 
-    fp = FcitxXDGGetFileUserWithPrefix("pinyin", TEMP_FILE, "w", &tempfile);
     if (!fp) {
         FcitxLog(ERROR, _("Cannot Save User Pinyin Database: %s"), tempfile);
         free(tempfile);
@@ -1803,8 +1829,16 @@ void SavePYFreq(FcitxPinyinState *pystate)
     FILE *fp;
     PyFreq *pPyFreq;
     HZ *hz;
+    int fd;
+
+    FcitxXDGGetFileUserWithPrefix("pinyin", "", "w", NULL);
+    FcitxXDGGetFileUserWithPrefix("pinyin", PINYIN_TEMP_FILE, NULL, &tempfile);
+    fd = mkstemp(tempfile);
+    fp = NULL;
+
+    if (fd > 0)
+        fp = fdopen(fd, "w");
 
-    fp = FcitxXDGGetFileUserWithPrefix("pinyin", TEMP_FILE, "w", &tempfile);
     if (!fp) {
         FcitxLog(ERROR, _("Cannot Save Frequent word: %s"), tempfile);
         free(tempfile);
@@ -1868,8 +1902,16 @@ void SavePYIndex(FcitxPinyinState *pystate)
     char *tempfile;
     FILE *fp;
     PYFA* PYFAList = pystate->PYFAList;
+    int fd;
+
+    FcitxXDGGetFileUserWithPrefix("pinyin", "", "w", NULL);
+    FcitxXDGGetFileUserWithPrefix("pinyin", PINYIN_TEMP_FILE, NULL, &tempfile);
+    fd = mkstemp(tempfile);
+    fp = NULL;
+
+    if (fd > 0)
+        fp = fdopen(fd, "w");
 
-    fp = FcitxXDGGetFileUserWithPrefix("pinyin", TEMP_FILE, "w", &tempfile);
     if (!fp) {
         FcitxLog(ERROR, _("Cannot Save Pinyin Index: %s"), tempfile);
         free(tempfile);
