@@ -44,6 +44,8 @@
 #include "module/punc/punc.h"
 #include "tablepinyinwrapper.h"
 
+#define MAX_TABLE_INPUT 50
+
 static void FreeTableConfig(void *v);
 const UT_icd table_icd = {sizeof(TableMetaData), NULL , NULL, FreeTableConfig};
 const UT_icd tableCand_icd = {sizeof(TABLECANDWORD*), NULL, NULL, NULL };
@@ -53,6 +55,7 @@ typedef struct _TableCandWordSortContext {
 
 static void *TableCreate(FcitxInstance* instance);
 static int TableCandCmp(const void* a, const void* b, void* arg);
+static INPUT_RETURN_VALUE TableKeyBlocker(void* arg, FcitxKeySym sym, unsigned int state);
 
 FCITX_EXPORT_API
 FcitxIMClass ime = {
@@ -84,8 +87,8 @@ void *TableCreate(FcitxInstance* instance)
             TableGetCandWords,
             TablePhraseTips,
             SaveTableIM,
-            NULL,
-            NULL,
+            ReloadTableConfig,
+            TableKeyBlocker,
             table->iPriority,
             table->langCode
         );
@@ -98,7 +101,7 @@ void *TableCreate(FcitxInstance* instance)
 void SaveTableIM(void *arg)
 {
     TableMetaData* table = (TableMetaData*) arg;
-    
+
     if (!table->tableDict)
         return;
     if (table->tableDict && table->tableDict->iTableChanged)
@@ -158,6 +161,7 @@ void LoadTableInfo(FcitxTableState *tbl)
             FcitxConfigBindSync((FcitxGenericConfig*)t);
             FcitxLog(DEBUG, _("Table Config %s is %s"), string->name, (t->bEnabled) ? "Enabled" : "Disabled");
             if (t->bEnabled) {
+                t->confName = strdup(string->name);
                 t->owner = tbl;
                 tbl->iTableCount ++;
             }
@@ -186,6 +190,10 @@ boolean TableInit(void *arg)
     boolean flag = true;
     FcitxInstanceSetContext(tbl->owner, CONTEXT_IM_KEYBOARD_LAYOUT, table->kbdlayout);
     FcitxInstanceSetContext(tbl->owner, CONTEXT_SHOW_REMIND_STATUS, &flag);
+    if (table->bUseAlternativePageKey) {
+        FcitxInstanceSetContext(tbl->owner, CONTEXT_ALTERNATIVE_PREVPAGE_KEY, table->hkAlternativePrevPage);
+        FcitxInstanceSetContext(tbl->owner, CONTEXT_ALTERNATIVE_NEXTPAGE_KEY, table->hkAlternativeNextPage);
+    }
     FcitxAddon* pyaddon = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(tbl->owner), "fcitx-pinyin");
     tbl->pyaddon = pyaddon;
     if (pyaddon == NULL)
@@ -217,9 +225,10 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
     FcitxInputState *input = FcitxInstanceGetInputState(instance);
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(instance);
     char* strCodeInput = FcitxInputStateGetRawInputBuffer(input);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
 
-    FcitxCandidateWordSetChoose(FcitxInputStateGetCandidateList(input), table->strChoose);
-    FcitxCandidateWordSetPageSize(FcitxInputStateGetCandidateList(input), config->iMaxCandWord);
+    FcitxCandidateWordSetChoose(candList, table->strChoose);
+    FcitxCandidateWordSetPageSize(candList, config->iMaxCandWord);
 
     tbl->iTableIMIndex = utarray_eltidx(tbl->table, table);
     if (tbl->iTableIMIndex != tbl->iCurrentTableLoaded) {
@@ -250,9 +259,16 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
     }
 
     retVal = IRV_DO_NOTHING;
-    if (state == FcitxKeyState_None && (IsInputKey(table->tableDict, sym) || IsEndKey(table, sym) || sym == table->cMatchingKey || sym == table->cPinyin)) {
+    if (state == FcitxKeyState_None &&
+        (IsInputKey(table->tableDict, sym)
+         || IsEndKey(table, sym)
+         || (table->bUseMatchingKey && sym == table->cMatchingKey)
+         || (table->bUsePY && sym == table->cPinyin)
+         || (strCodeInput[0] == table->cPinyin && table->bUsePY && sym == FcitxKey_apostrophe)
+        )
+       ) {
         if (FcitxInputStateGetIsInRemind(input))
-            FcitxCandidateWordReset(FcitxInputStateGetCandidateList(input));
+            FcitxCandidateWordReset(candList);
         FcitxInputStateSetIsInRemind(input, false);
 
         /* it's not in special state */
@@ -268,7 +284,14 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                     retVal = IRV_DO_NOTHING;
             } else {
                 /* length is not too large */
-                if ((FcitxInputStateGetRawInputBufferSize(input) < table->tableDict->iCodeLength) || (table->tableDict->bHasPinyin && FcitxInputStateGetRawInputBufferSize(input) < table->tableDict->iPYCodeLength)) {
+                if (((FcitxInputStateGetRawInputBufferSize(input) < table->tableDict->iCodeLength)
+                    || (table->tableDict->bHasPinyin && FcitxInputStateGetRawInputBufferSize(input) < table->tableDict->iPYCodeLength)
+                    || (((FcitxCandidateWordPageCount(candList) == 0
+                        && table->bNoMatchDontCommit) || !table->bUseAutoSend)
+                        && FcitxInputStateGetRawInputBufferSize(input) >= table->tableDict->iCodeLength
+                    ))
+                    && FcitxInputStateGetRawInputBufferSize(input) <= MAX_TABLE_INPUT
+                ) {
                     strCodeInput[FcitxInputStateGetRawInputBufferSize(input)] = (char) sym;
                     strCodeInput[FcitxInputStateGetRawInputBufferSize(input) + 1] = '\0';
                     FcitxInputStateSetRawInputBufferSize(input, FcitxInputStateGetRawInputBufferSize(input) + 1);
@@ -279,14 +302,14 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                         char        *strTemp;
                         char        *strLastFirstCand;
                         CANDTYPE     lastFirstCandType;
-                        int          lastPageCount = FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input));
+                        int          lastPageCount = FcitxCandidateWordPageCount(candList);
 
                         strLastFirstCand = (char *)NULL;
                         lastFirstCandType = CT_AUTOPHRASE;
-                        if (FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input)) != 0) {
+                        if (FcitxCandidateWordPageCount(candList) != 0) {
                             // to realize auto-sending HZ to client
                             FcitxCandidateWord *candWord = NULL;
-                            candWord = FcitxCandidateWordGetCurrentWindow(FcitxInputStateGetCandidateList(input));
+                            candWord = FcitxCandidateWordGetCurrentWindow(candList);
                             if (candWord->owner == table) {
                                 TABLECANDWORD* tableCandWord = candWord->priv;
                                 lastFirstCandType = tableCandWord->flag;
@@ -300,27 +323,35 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                         FcitxModuleFunctionArg farg;
                         int key = FcitxInputStateGetRawInputBuffer(input)[0];
                         farg.args[0] = &key;
-                        strTemp = InvokeFunction(instance, FCITX_PUNC, GETPUNC, farg);
+                        if (!table->bIgnorePunc)
+                            strTemp = InvokeFunction(instance, FCITX_PUNC, GETPUNC, farg);
+                        else
+                            strTemp = NULL;
                         if (IsEndKey(table, sym)) {
                             if (FcitxInputStateGetRawInputBufferSize(input) == 1)
                                 return IRV_TO_PROCESS;
 
-                            if (FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input)) == 0) {
+                            if (FcitxCandidateWordPageCount(candList) == 0) {
                                 FcitxInputStateSetRawInputBufferSize(input, 0);
                                 return IRV_CLEAN;
                             }
 
-                            if (FcitxCandidateWordGetCurrentWindowSize(FcitxInputStateGetCandidateList(input)) == 1) {
-                                retVal = FcitxCandidateWordChooseByIndex(FcitxInputStateGetCandidateList(input), 0);
+                            if (FcitxCandidateWordGetCurrentWindowSize(candList) == 1) {
+                                retVal = FcitxCandidateWordChooseByIndex(candList, 0);
                                 return retVal;
                             }
 
                             return IRV_DISPLAY_CANDWORDS;
-                        } else if (table->iTableAutoSendToClientWhenNone
-                                   && (FcitxInputStateGetRawInputBufferSize(input) == (table->iTableAutoSendToClientWhenNone + 1))
-                                   && FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input)) == 0) {
+                        } else if (table->bUseAutoSend
+                                   && table->iTableAutoSendToClientWhenNone
+                                   && (!(retVal & IRV_FLAG_PENDING_COMMIT_STRING))
+                                   && (FcitxInputStateGetRawInputBufferSize(input) >= (table->iTableAutoSendToClientWhenNone + 1))
+                                   && FcitxCandidateWordPageCount(candList) == 0) {
                             if (strLastFirstCand && (lastFirstCandType != CT_AUTOPHRASE)) {
                                 FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), strLastFirstCand);
+                            } else if (table->bSendRawPreedit) {
+                                strCodeInput[FcitxInputStateGetRawInputBufferSize(input) - 1] = '\0';
+                                FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), strCodeInput);
                             }
                             retVal = IRV_DISPLAY_CANDWORDS;
                             FcitxInputStateSetRawInputBufferSize(input, 1);
@@ -332,20 +363,21 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                         }
                     }
                 } else {
-                    if (table->iTableAutoSendToClient) {
+                    if (table->bUseAutoSend && table->iTableAutoSendToClient) {
                         retVal = IRV_DISPLAY_CANDWORDS;
-                        if (FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input))) {
-                            FcitxCandidateWord* candWord = FcitxCandidateWordGetCurrentWindow(FcitxInputStateGetCandidateList(input));
+                        if (FcitxCandidateWordPageCount(candList)) {
+                            FcitxCandidateWord* candWord = FcitxCandidateWordGetCurrentWindow(candList);
                             if (candWord->owner == table) {
                                 TABLECANDWORD* tableCandWord = candWord->priv;
                                 if (tableCandWord->flag != CT_AUTOPHRASE) {
                                     INPUT_RETURN_VALUE ret = TableGetCandWord(table, candWord);
                                     if (ret & IRV_FLAG_PENDING_COMMIT_STRING) {
                                         FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), FcitxInputStateGetOutputString(input));
-                                        FcitxInstanceIncreateInputCharacterCount(instance, fcitx_utf8_strlen(FcitxInputStateGetOutputString(input)));
                                     }
                                 }
                             }
+                        } else if (table->bSendRawPreedit) {
+                            FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), strCodeInput);
                         }
 
                         FcitxInputStateSetRawInputBufferSize(input, 1);
@@ -439,7 +471,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                 TableResetStatus(table);
                 FcitxInstanceCleanInputWindowUp(instance);
                 FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
-                FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
+                FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT | MSG_DONOT_COMMIT_WHEN_UNFOCUS, "%s", FcitxInputStateGetRawInputBuffer(input));
                 retVal = IRV_DISPLAY_CANDWORDS;
             } else
                 return IRV_CLEAN;
@@ -447,13 +479,13 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
             int iKey;
             iKey = FcitxHotkeyCheckChooseKey(sym, state, table->strChoose);
 
-            if (FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input)) == 0)
+            if (FcitxCandidateWordPageCount(candList) == 0)
                 return IRV_TO_PROCESS;
 
-            if (FcitxCandidateWordGetByIndex(FcitxInputStateGetCandidateList(input), iKey) == NULL)
+            if (FcitxCandidateWordGetByIndex(candList, iKey) == NULL)
                 return IRV_DO_NOTHING;
             else {
-                FcitxCandidateWord* candWord = FcitxCandidateWordGetByIndex(FcitxInputStateGetCandidateList(input), iKey);
+                FcitxCandidateWord* candWord = FcitxCandidateWordGetByIndex(candList, iKey);
                 if (candWord->owner == table && tbl->bIsTableDelPhrase) {
                     TableDelPhraseByIndex(table, candWord->priv);
                     tbl->bIsTableDelPhrase = false;
@@ -467,7 +499,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
             }
         } else if (!tbl->bIsTableDelPhrase && !tbl->bIsTableAdjustOrder) {
             if (FcitxHotkeyIsHotKey(sym, state, tbl->hkTableAdjustOrder)) {
-                if (FcitxCandidateWordGetListSize(FcitxInputStateGetCandidateList(input)) < 2 || FcitxInputStateGetIsInRemind(input))
+                if (FcitxCandidateWordGetListSize(candList) < 2 || FcitxInputStateGetIsInRemind(input))
                     return IRV_DO_NOTHING;
 
                 tbl->bIsTableAdjustOrder = true;
@@ -475,7 +507,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                 FcitxMessagesAddMessageAtLast(FcitxInputStateGetAuxUp(input), MSG_TIPS, _("Choose the phrase to be put in the front, Press Escape to Cancel"));
                 retVal = IRV_DISPLAY_MESSAGE;
             } else if (FcitxHotkeyIsHotKey(sym, state, tbl->hkTableDelPhrase)) {
-                if (!FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input)) || FcitxInputStateGetIsInRemind(input))
+                if (!FcitxCandidateWordPageCount(candList) || FcitxInputStateGetIsInRemind(input))
                     return IRV_DO_NOTHING;
 
                 tbl->bIsTableDelPhrase = true;
@@ -492,19 +524,30 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                 strCodeInput[FcitxInputStateGetRawInputBufferSize(input)] = '\0';
 
                 if (FcitxInputStateGetRawInputBufferSize(input) == 1 && strCodeInput[0] == table->cPinyin && table->bUsePY) {
-                    FcitxCandidateWordReset(FcitxInputStateGetCandidateList(input));
+                    FcitxCandidateWordReset(candList);
                     retVal = IRV_DISPLAY_LAST;
                 } else if (FcitxInputStateGetRawInputBufferSize(input))
                     retVal = IRV_DISPLAY_CANDWORDS;
                 else
                     retVal = IRV_CLEAN;
-            } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_SPACE)) {
+            } else if (FcitxHotkeyIsHotKey(sym, state, table->hkCommitKey)) {
                 if (FcitxInputStateGetRawInputBufferSize(input) == 1 && strCodeInput[0] == table->cPinyin && table->bUsePY)
                     retVal = IRV_COMMIT_STRING;
                 else {
-                    retVal = FcitxCandidateWordChooseByIndex(FcitxInputStateGetCandidateList(input), 0);
-                    if (retVal == IRV_TO_PROCESS)
+                    if (FcitxCandidateWordPageCount(candList) == 0 && table->bCommitKeyCommitWhenNoMatch) {
+                        FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), FcitxInputStateGetRawInputBuffer(input));
+                        FcitxInputStateSetRawInputBufferSize(input, 0);
+                        FcitxInputStateGetRawInputBuffer(input)[0] = '\0';
+                        FcitxInputStateSetIsInRemind(input, false);
+                        FcitxInstanceCleanInputWindow(instance);
+                        FcitxUIUpdateInputWindow(instance);
                         retVal = IRV_DO_NOTHING;
+                    }
+                    else {
+                        retVal = FcitxCandidateWordChooseByIndex(candList, 0);
+                        if (retVal == IRV_TO_PROCESS)
+                            retVal = IRV_DO_NOTHING;
+                    }
                 }
             } else {
                 return IRV_TO_PROCESS;
@@ -541,7 +584,7 @@ INPUT_RETURN_VALUE TableGetCandWord(void* arg, FcitxCandidateWord* candWord)
             UpdateHZLastInput(table, FcitxInputStateGetOutputString(input));
 
         if (tbl->pCurCandRecord)
-            TableUpdateHitFrequency(table->tableDict, tbl->pCurCandRecord);
+            TableUpdateHitFrequency(table, tbl->pCurCandRecord);
     }
 
     return retVal;
@@ -570,7 +613,6 @@ INPUT_RETURN_VALUE _TableGetCandWord(TableMetaData* table, TABLECANDWORD* tableC
     else
         tbl->pCurCandRecord = (RECORD *)NULL;
 
-    table->tableDict->iTableChanged++;
     if (table->tableDict->iTableChanged >= TABLE_AUTO_SAVE_AFTER)
         SaveTableDict(table);
 
@@ -653,7 +695,7 @@ INPUT_RETURN_VALUE TableGetPinyinCandWords(TableMetaData* table)
 
     FcitxInstanceCleanInputWindowUp(instance);
     FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
-    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
+    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT | MSG_DONOT_COMMIT_WHEN_UNFOCUS, "%s", FcitxInputStateGetRawInputBuffer(input));
     FcitxInputStateSetCursorPos(input, FcitxInputStateGetRawInputBufferSize(input));
     FcitxInputStateSetClientCursorPos(input, 0);
 
@@ -693,6 +735,7 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
     RECORD         *recTemp;
     FcitxInstance *instance = tbl->owner;
     FcitxInputState *input = FcitxInstanceGetInputState(instance);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
 
     if (FcitxInputStateGetRawInputBuffer(input)[0] == '\0')
         return IRV_TO_PROCESS;
@@ -711,7 +754,10 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
             FcitxMessagesSetMessageCount(FcitxInputStateGetPreedit(input), 0);
             FcitxMessagesSetMessageCount(FcitxInputStateGetClientPreedit(input), 0);
             FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
-            FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
+            FcitxMessageType type = MSG_INPUT;
+            if (!table->bSendRawPreedit)
+                type |= MSG_DONOT_COMMIT_WHEN_UNFOCUS;
+            FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), type, "%s", FcitxInputStateGetRawInputBuffer(input));
             FcitxInputStateSetCursorPos(input, strlen(FcitxInputStateGetRawInputBuffer(input)));
             FcitxInputStateSetClientCursorPos(input, 0);
         }
@@ -779,7 +825,7 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
                     } else
                         totallen += 1;
                 }
-                
+
                 candWord.strExtra = fcitx_utils_malloc0(sizeof(char) * (totallen + 2));
                 candWord.strExtra[0] = ' ';
                 for (i = 0; i < codelen; i ++) {
@@ -798,7 +844,7 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
             candWord.extraType = MSG_CODE;
         }
 
-        FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
+        FcitxCandidateWordAppend(candList, &candWord);
     }
     utarray_clear(&candTemp);
 
@@ -826,34 +872,45 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
         candWord.strExtra = NULL;
         candWord.wordType = MSG_USERPHR;
 
-        FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
+        FcitxCandidateWordAppend(candList, &candWord);
     }
 
     utarray_done(&candTemp);
 
-    if (FcitxInputStateGetRawInputBufferSize(input)) {
-        FcitxMessagesSetMessageCount(FcitxInputStateGetPreedit(input), 0);
-        FcitxMessagesSetMessageCount(FcitxInputStateGetClientPreedit(input), 0);
-        FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
-        FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
-        FcitxInputStateSetCursorPos(input, strlen(FcitxInputStateGetRawInputBuffer(input)));
-        FcitxInputStateSetClientCursorPos(input, 0);
-    }
-
     INPUT_RETURN_VALUE retVal = IRV_DISPLAY_CANDWORDS;
 
-    if (table->iTableAutoSendToClient && (FcitxInputStateGetRawInputBufferSize(input) >= table->iTableAutoSendToClient)) {
-        if (FcitxCandidateWordGetCurrentWindowSize(FcitxInputStateGetCandidateList(input)) == 1) {  //如果只有一个候选词，则送到客户程序中
-            FcitxCandidateWord* candWord = FcitxCandidateWordGetCurrentWindow(FcitxInputStateGetCandidateList(input));
+    if (table->bUseAutoSend && table->iTableAutoSendToClient && (FcitxInputStateGetRawInputBufferSize(input) >= table->iTableAutoSendToClient)) {
+        if (FcitxCandidateWordGetCurrentWindowSize(candList) == 1) {  //如果只有一个候选词，则送到客户程序中
+            FcitxCandidateWord* candWord = FcitxCandidateWordGetCurrentWindow(candList);
             if (candWord->owner == table) {
                 TABLECANDWORD* tableCandWord = candWord->priv;
                 if (tableCandWord->flag != CT_AUTOPHRASE || (tableCandWord->flag == CT_AUTOPHRASE && !table->iSaveAutoPhraseAfter))
                     if (!(tableCandWord->flag == CT_NORMAL && tableCandWord->candWord.record->type == RECORDTYPE_PINYIN))
-                        retVal = FcitxCandidateWordChooseByIndex(FcitxInputStateGetCandidateList(input), 0);
+                        retVal = FcitxCandidateWordChooseByIndex(candList, 0);
             }
         }
     }
 
+    if (FcitxInputStateGetRawInputBufferSize(input)) {
+        FcitxMessagesSetMessageCount(FcitxInputStateGetPreedit(input), 0);
+        FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
+        FcitxInputStateSetCursorPos(input, strlen(FcitxInputStateGetRawInputBuffer(input)));
+        FcitxCandidateWord* candWord = NULL;
+        if (table->bFirstCandidateAsPreedit && (candWord = FcitxCandidateWordGetFirst(candList))) {
+            FcitxMessagesSetMessageCount(FcitxInputStateGetClientPreedit(input), 0);
+            FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", candWord->strWord);
+            FcitxInputStateSetClientCursorPos(input, 0);
+        }
+        else {
+            FcitxMessagesSetMessageCount(FcitxInputStateGetClientPreedit(input), 0);
+            FcitxMessageType type = MSG_INPUT;
+            if (!table->bSendRawPreedit)
+                type |= MSG_DONOT_COMMIT_WHEN_UNFOCUS;
+            FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), type, "%s", FcitxInputStateGetRawInputBuffer(input));
+            FcitxInputStateSetClientCursorPos(input, 0);
+        }
+    }
+
     return retVal;
 }
 
@@ -877,7 +934,7 @@ void TableAdjustOrderByIndex(TableMetaData* table, TABLECANDWORD* tableCandWord)
 {
     RECORD         *recTemp;
     int             iTemp;
-    
+
     FcitxTableState* tbl = table->owner;
 
     recTemp = tableCandWord->candWord.record;
@@ -1030,7 +1087,7 @@ INPUT_RETURN_VALUE TableGetFHCandWords(TableMetaData* table)
 
     FcitxInstanceCleanInputWindowUp(instance);
     FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
-    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
+    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT | MSG_DONOT_COMMIT_WHEN_UNFOCUS, "%s", FcitxInputStateGetRawInputBuffer(input));
     FcitxInputStateSetCursorPos(input, FcitxInputStateGetRawInputBufferSize(input));
     FcitxInputStateSetClientCursorPos(input, 0);
 
@@ -1141,6 +1198,8 @@ void FreeTableConfig(void *v)
     if (!table)
         return;
     FcitxConfigFreeConfigFile(table->config.configFile);
+    FcitxHotkeyFree(table->hkAlternativePrevPage);
+    FcitxHotkeyFree(table->hkAlternativeNextPage);
     free(table->strPath);
     free(table->strSymbolFile);
     free(table->uniqueName);
@@ -1149,6 +1208,7 @@ void FreeTableConfig(void *v)
     free(table->strEndCode);
     free(table->strSymbol);
     free(table->strChoose);
+    fcitx_utils_free(table->confName);
 }
 
 int TableCandCmp(const void* a, const void* b, void *arg)
@@ -1178,4 +1238,71 @@ int TableCandCmp(const void* a, const void* b, void *arg)
     return 0;
 }
 
+INPUT_RETURN_VALUE TableKeyBlocker(void* arg, FcitxKeySym sym, unsigned int state)
+{
+    TableMetaData* table = arg;
+    FcitxInstance* instance = table->owner->owner;
+    FcitxInputState* input = FcitxInstanceGetInputState(instance);
+
+    do {
+        if (!table->bCommitAndPassByInvalidKey)
+            break;
+        if (!FcitxHotkeyIsHotKeySimple(sym, state))
+            break;
+
+        if (FcitxCandidateWordPageCount(FcitxInputStateGetCandidateList(input))) {
+            FcitxCandidateWord* candWord = FcitxCandidateWordGetCurrentWindow(FcitxInputStateGetCandidateList(input));
+            if (candWord->owner != table)
+                break;
+            TABLECANDWORD* tableCandWord = candWord->priv;
+            if (tableCandWord->flag == CT_AUTOPHRASE)
+                break;
+            INPUT_RETURN_VALUE ret = TableGetCandWord(table, candWord);
+            if (!(ret & IRV_FLAG_PENDING_COMMIT_STRING))
+                break;
+            FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), FcitxInputStateGetOutputString(input));
+        }
+        else if (table->bSendRawPreedit) {
+            FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), FcitxInputStateGetRawInputBuffer(input));
+        }
+        FcitxInputStateSetRawInputBufferSize(input, 0);
+        FcitxInputStateGetRawInputBuffer(input)[0] = '\0';
+        FcitxInputStateSetIsInRemind(input, false);
+        FcitxInstanceCleanInputWindow(instance);
+        FcitxUIUpdateInputWindow(instance);
+        return IRV_FLAG_FORWARD_KEY;
+    } while (0);
+    return FcitxStandardKeyBlocker(input, sym, state);
+}
+
+/* we should try not to break the existing on even if the config file is not exist anymore */
+void ReloadTableConfig(void* arg)
+{
+    TableMetaData* table = arg;
+    size_t len = 0;
+    int i = 0;
+    char** tablePath = FcitxXDGGetPathWithPrefix(&len, "table");
+
+    char **paths = fcitx_utils_malloc0(sizeof(char*) * len);
+    for (i = 0; i < len ; i ++)
+        paths[i] = NULL;
+
+    for (i = len - 1; i >= 0; i--) {
+        asprintf(&paths[i], "%s/%s", tablePath[len - i - 1], table->confName);
+        FcitxLog(DEBUG, "Load Table Config File:%s", paths[i]);
+    }
+    FcitxConfigFile* cfile = FcitxConfigParseMultiConfigFile(paths, len, GetTableConfigDesc());
+    if (cfile) {
+        TableMetaDataConfigBind(table, cfile, GetTableConfigDesc());
+        FcitxConfigBindSync((FcitxGenericConfig*)table);
+    }
+    for (i = 0; i < len ; i ++) {
+        free(paths[i]);
+        paths[i] = NULL;
+    }
+
+    free(paths);
+    FcitxXDGFreePath(tablePath);
+}
+
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
