@@ -10,14 +10,13 @@
 #include "fcitx-utils/utf8.h"
 #include "tabledict.h"
 
-#define TEMP_FILE       "FCITX_TABLE_TEMP"
-
+#define TABLE_TEMP_FILE "table_XXXXXX"
 const int iInternalVersion = INTERNAL_VERSION;
 
 boolean LoadTableDict(TableMetaData* tableMetaData)
 {
     char            strCode[MAX_CODE_LENGTH + 1];
-    char            strHZ[PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1];
+    char           *strHZ = 0;
     FILE           *fpDict;
     RECORD         *recTemp;
     unsigned int    i = 0;
@@ -102,9 +101,14 @@ boolean LoadTableDict(TableMetaData* tableMetaData)
     }
 
     iRecordIndex = 0;
+    size_t bufSize = 0;
     for (i = 0; i < tableDict->iRecordCount; i++) {
         fread(strCode, sizeof(char), tableDict->iPYCodeLength + 1, fpDict);
         fread(&iTemp, sizeof(unsigned int), 1, fpDict);
+        if (iTemp > bufSize) {
+            bufSize = iTemp;
+            strHZ = realloc(strHZ, bufSize);
+        }
         fread(strHZ, sizeof(char), iTemp, fpDict);
         recTemp = (RECORD *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(RECORD));
         recTemp->strCode = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iPYCodeLength + 1));
@@ -139,7 +143,7 @@ boolean LoadTableDict(TableMetaData* tableMetaData)
                 tableSingleHZ = tableDict->tableSingleHZ;
             else if (recTemp->type == RECORDTYPE_CONSTRUCT)
                 tableSingleHZ = tableDict->tableSingleHZCons;
-            
+
             if (tableSingleHZ) {
                 iTemp = CalHZIndex(recTemp->strHZ);
                 if (iTemp < SINGLE_HZ_COUNT) {
@@ -154,14 +158,16 @@ boolean LoadTableDict(TableMetaData* tableMetaData)
 
         if (recTemp->type == RECORDTYPE_PINYIN)
             tableDict->bHasPinyin = true;
-        
+
         if (recTemp->type == RECORDTYPE_PROMPT && strlen(recTemp->strCode) == 1)
             tableDict->promptCode[(uint8_t) recTemp->strCode[0]] = recTemp;
-            
+
         tableDict->currentRecord->next = recTemp;
         recTemp->prev = tableDict->currentRecord;
         tableDict->currentRecord = recTemp;
     }
+    if (strHZ)
+        free(strHZ);
 
     tableDict->currentRecord->next = tableDict->recordHead;
     tableDict->recordHead->prev = tableDict->currentRecord;
@@ -248,20 +254,29 @@ void SaveTableDict(TableMetaData *tableMetaData)
     FILE           *fpDict;
     unsigned int    iTemp;
     unsigned int    i;
+    int             fd;
     int8_t          cTemp;
     TableDict      *tableDict = tableMetaData->tableDict;
 
     if (!tableDict->iTableChanged)
         return;
 
-    fpDict = FcitxXDGGetFileUserWithPrefix("table", TEMP_FILE, "w", &tempfile);
+    // make ~/.config/fcitx/table/ dir
+    FcitxXDGGetFileUserWithPrefix("table", "", "w", NULL);
+    FcitxXDGGetFileUserWithPrefix("table", TABLE_TEMP_FILE, NULL, &tempfile);
+    fd = mkstemp(tempfile);
+    fpDict = NULL;
+
+    if (fd > 0)
+        fpDict = fdopen(fd, "w");
+
     if (!fpDict) {
         FcitxLog(ERROR, _("Save dict error"));
         free(tempfile);
         return;
     }
 
-    //写入版本号--如果第一个字为0,表示后面那个字节为版本号，为了与老版本兼容
+    // write version number
     iTemp = 0;
     fwrite(&iTemp, sizeof(unsigned int), 1, fpDict);
     fwrite(&iInternalVersion, sizeof(char), 1, fpDict);
@@ -276,7 +291,7 @@ void SaveTableDict(TableMetaData *tableMetaData)
     fwrite(tableDict->strIgnoreChars, sizeof(char), iTemp + 1, fpDict);
 
     fwrite(&(tableDict->bRule), sizeof(unsigned char), 1, fpDict);
-    if (tableDict->bRule) { //表示有组词规则
+    if (tableDict->bRule) { // table contains rule
         for (i = 0; i < tableDict->iCodeLength - 1; i++) {
             fwrite(&(tableDict->rule[i].iFlag), sizeof(unsigned char), 1, fpDict);
             fwrite(&(tableDict->rule[i].iWords), sizeof(unsigned char), 1, fpDict);
@@ -317,8 +332,16 @@ void SaveTableDict(TableMetaData *tableMetaData)
     tableDict->iTableChanged = 0;
 
     if (tableDict->autoPhrase) {
-        //保存上次的自动词组信息
-        fpDict = FcitxXDGGetFileUserWithPrefix("table", TEMP_FILE, "w", &tempfile);
+        // Save auto phrase
+        // make ~/.config/fcitx/table/ dir
+        FcitxXDGGetFileUserWithPrefix("table", "", "w", NULL);
+        FcitxXDGGetFileUserWithPrefix("table", TABLE_TEMP_FILE, NULL, &tempfile);
+        fd = mkstemp(tempfile);
+        fpDict = NULL;
+
+        if (fd > 0)
+            fpDict = fdopen(fd, "w");
+
         if (fpDict) {
             fwrite(&tableDict->iAutoPhrase, sizeof(int), 1, fpDict);
             for (i = 0; i < tableDict->iAutoPhrase; i++) {
@@ -441,6 +464,7 @@ void TableCreateAutoPhrase(TableMetaData* tableMetaData, char iCount)
                 tableDict->insertPoint->iSelected = 0;
                 tableDict->insertPoint = tableDict->insertPoint->next;
             }
+            tableDict->iTableChanged++;
 
         _next:
             continue;
@@ -473,7 +497,7 @@ boolean TableCreatePhraseCode(TableDict* tableDict, char *strHZ)
         if (tableDict->rule[i].iWords == i2 && tableDict->rule[i].iFlag == i1)
             break;
     }
-    
+
     if (i == tableDict->iCodeLength - 1)
         return true;
 
@@ -491,7 +515,7 @@ boolean TableCreatePhraseCode(TableDict* tableDict, char *strHZ)
         }
 
         int hzIndex = CalHZIndex(strTemp);
-        
+
         if (tableDict->tableSingleHZ[hzIndex]) {
             if (tableDict->tableSingleHZCons[hzIndex])
                 recTemp = tableDict->tableSingleHZCons[hzIndex];
@@ -561,6 +585,7 @@ void TableInsertPhrase(TableDict* tableDict, const char *strCode, const char *st
     dictNew->next = insertPoint;
 
     tableDict->iRecordCount++;
+    tableDict->iTableChanged++;
 }
 
 /*
@@ -587,12 +612,16 @@ void TableDelPhrase(TableDict* tableDict, RECORD * record)
      */
 
     tableDict->iRecordCount--;
+    tableDict->iTableChanged++;
 }
 
-void TableUpdateHitFrequency(TableDict* tableDict, RECORD * record)
+void TableUpdateHitFrequency(TableMetaData* tableMetaData, RECORD * record)
 {
-    record->iHit++;
-    record->iIndex = ++tableDict->iTableIndex;
+    if (tableMetaData->tableOrder != AD_NO) {
+        tableMetaData->tableDict->iTableChanged++;
+        record->iHit++;
+        record->iIndex = ++tableMetaData->tableDict->iTableIndex;
+    }
 }
 
 int TableCompareCode(const TableMetaData* tableMetaData, const char *strUser, const char *strDict)
