@@ -51,6 +51,7 @@ struct _FcitxInputState {
     int iCodeInputCount;
     char strCodeInput[MAX_USER_INPUT + 1];
     char strStringGet[MAX_USER_INPUT + 1];
+    char strLastCommit[MAX_USER_INPUT + 1];
     boolean bIsInRemind;
 
     time_t dummy;
@@ -68,7 +69,7 @@ struct _FcitxInputState {
     FcitxMessages* msgAuxUp;
     FcitxMessages* msgAuxDown;
     FcitxMessages* msgClientPreedit;
-    
+
     uint32_t keycode;
     uint32_t keysym;
     uint32_t keystate;
@@ -116,12 +117,14 @@ boolean IMIsInIMNameList(UT_array* imList, FcitxIM* ime);
 
 void FcitxInstanceLoadIM(struct _FcitxInstance* instance, FcitxAddon* addon);
 
-void FcitxInstanceSwitchIMInternal(struct _FcitxInstance* instance, int index, boolean skipZero);
+void FcitxInstanceSwitchIMInternal(struct _FcitxInstance* instance, int index, boolean skipZero, boolean updateGlobal);
 
 FcitxConfigFileDesc* GetIMConfigDesc();
 
 int IMPriorityCmp(const void *a, const void *b);
 
+void FcitxInstanceUpdateCurrentIM(struct _FcitxInstance* instance, boolean force);
+
 #endif
 
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
