@@ -52,7 +52,60 @@
 #include "addon-internal.h"
 #include "context-internal.h"
 
-static const char* GetStateName(INPUT_RETURN_VALUE retVal);
+
+static const FcitxHotkey* switchKey1[] = {
+    FCITX_RCTRL,
+    FCITX_RSHIFT,
+    FCITX_LSHIFT,
+    FCITX_LCTRL,
+    FCITX_ALT_LSHIFT,
+    FCITX_ALT_RSHIFT,
+    FCITX_RCTRL,
+    FCITX_RSHIFT,
+    FCITX_NONE_KEY,
+};
+
+static const FcitxHotkey* switchKey2[] = {
+    FCITX_NONE_KEY,
+    FCITX_NONE_KEY,
+    FCITX_NONE_KEY,
+    FCITX_NONE_KEY,
+    FCITX_NONE_KEY,
+    FCITX_NONE_KEY,
+    FCITX_LCTRL,
+    FCITX_LSHIFT,
+    FCITX_NONE_KEY
+};
+
+static const FcitxHotkey* imSWNextKey1[] = {
+    FCITX_LCTRL_LSHIFT,
+    FCITX_LALT_LSHIFT,
+    FCITX_LCTRL_LSUPER,
+    FCITX_LALT_LSUPER,
+};
+
+static const FcitxHotkey* imSWNextKey2[] = {
+    FCITX_LCTRL_LSHIFT2,
+    FCITX_LALT_LSHIFT2,
+    FCITX_LCTRL_LSUPER2,
+    FCITX_LALT_LSUPER2,
+};
+
+static const FcitxHotkey* imSWPrevKey1[] = {
+    FCITX_RCTRL_RSHIFT,
+    FCITX_RALT_RSHIFT,
+    FCITX_RCTRL_RSUPER,
+    FCITX_RALT_RSUPER,
+};
+
+static const FcitxHotkey* imSWPrevKey2[] = {
+    FCITX_RCTRL_RSHIFT2,
+    FCITX_RALT_RSHIFT2,
+    FCITX_RCTRL_RSUPER2,
+    FCITX_RALT_RSUPER2,
+};
+
+
 static const UT_icd ime_icd = {sizeof(FcitxIM), NULL , NULL, NULL};
 static const UT_icd imclass_icd = {sizeof(FcitxAddon*), NULL , NULL, NULL};
 static boolean IMMenuAction(FcitxUIMenu* menu, int index);
@@ -61,13 +114,14 @@ static void FcitxInstanceEnableIMInternal(FcitxInstance* instance, FcitxInputCon
 static void FcitxInstanceCloseIMInternal(FcitxInstance* instance, FcitxInputContext* ic);
 static void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputContext* ic, FcitxContextState objectState);
 static void FreeIMEntry(FcitxIMEntry* entry);
-static INPUT_RETURN_VALUE FcitxStandardKeyBlocker(FcitxInputState* input, FcitxKeySym key, unsigned int state);
+static boolean FcitxInstanceCheckICFromSameApplication (FcitxInstance* instance, FcitxInputContext* rec, FcitxInputContext* ic);
 
 FCITX_GETTER_VALUE(FcitxInputState, IsInRemind, bIsInRemind, boolean)
 FCITX_SETTER(FcitxInputState, IsInRemind, bIsInRemind, boolean)
 FCITX_GETTER_VALUE(FcitxInputState, IsDoInputOnly, bIsDoInputOnly, boolean)
 FCITX_SETTER(FcitxInputState, IsDoInputOnly, bIsDoInputOnly, boolean)
 FCITX_GETTER_VALUE(FcitxInputState, OutputString, strStringGet, char*)
+FCITX_GETTER_VALUE(FcitxInputState, LastCommitString, strLastCommit, const char*)
 FCITX_GETTER_VALUE(FcitxInputState, RawInputBuffer, strCodeInput, char*)
 FCITX_GETTER_VALUE(FcitxInputState, CursorPos, iCursorPos, int)
 FCITX_SETTER(FcitxInputState, CursorPos, iCursorPos, int)
@@ -179,9 +233,26 @@ void FcitxInstanceSaveAllIM(FcitxInstance* instance)
     }
 }
 
-static const char* GetStateName(INPUT_RETURN_VALUE retVal)
-{
-    return "unknown";
+
+static boolean FcitxInstanceCheckICFromSameApplication (FcitxInstance* instance, FcitxInputContext* rec, FcitxInputContext* ic) {
+    if (rec->frontendid != ic->frontendid)
+        return false;
+    if (rec == ic)
+        return true;
+    FcitxInputContext2* rec2 = (FcitxInputContext2*) rec;
+    FcitxInputContext2* ic2 = (FcitxInputContext2*) ic;
+    if (rec2->imname || ic2->imname)
+        return false;
+    UT_array* frontends = &instance->frontends;
+    FcitxAddon** pfrontend = (FcitxAddon**) utarray_eltptr(frontends, ic->frontendid);
+    if (pfrontend) {
+        FcitxFrontend* frontend = (*pfrontend)->frontend;
+        if (frontend->CheckICFromSameApplication
+            && frontend->CheckICFromSameApplication((*pfrontend)->addonInstance, rec, ic))
+            return true;
+    }
+
+    return false;
 }
 
 void FcitxInstanceLoadIM(FcitxInstance* instance, FcitxAddon* addon)
@@ -219,10 +290,13 @@ void FcitxInstanceLoadIM(FcitxInstance* instance, FcitxAddon* addon)
             dlclose(handle);
             return;
         }
+
+        instance->currentIMAddon = addon;
         if ((addon->addonInstance = imclass->Create(instance)) == NULL) {
             dlclose(handle);
             return;
         }
+        instance->currentIMAddon = NULL;
         addon->imclass = imclass;
         utarray_push_back(&instance->imeclasses, &addon);
     }
@@ -278,8 +352,44 @@ void FcitxInstanceRegisterIM(FcitxInstance *instance,
                        const char *langCode
                       )
 {
-    if (priority <= 0)
+    FcitxIMIFace iface;
+    memset(&iface, 0, sizeof(FcitxIMIFace));
+    iface.Init = Init;
+    iface.ResetIM = ResetIM;
+    iface.DoInput = DoInput;
+    iface.GetCandWords = GetCandWords;
+    iface.PhraseTips = PhraseTips;
+    iface.Save = Save;
+    iface.ReloadConfig = ReloadConfig;
+    iface.KeyBlocker = KeyBlocker;
+    FcitxInstanceRegisterIMv2(instance,
+                       imclass,
+                       uniqueName,
+                       name,
+                       iconName,
+                       iface,
+                       priority,
+                       langCode
+                      );
+}
+
+FCITX_EXPORT_API
+void FcitxInstanceRegisterIMv2(FcitxInstance *instance,
+                       void *imclass,
+                       const char* uniqueName,
+                       const char* name,
+                       const char* iconName,
+                       FcitxIMIFace iface,
+                       int priority,
+                       const char *langCode
+                      )
+{
+    if (priority < 0)
         return ;
+
+    if (priority == PRIORITY_MAGIC_FIRST)
+        priority = 0;
+
     UT_array* imes = &instance->availimes ;
     FcitxIM* entry;
 
@@ -300,20 +410,23 @@ void FcitxInstanceRegisterIM(FcitxInstance *instance,
     entry->uniqueName = strdup(uniqueName);
     entry->strName = strdup(name);
     entry->strIconName = strdup(iconName);
-    entry->Init = Init;
-    entry->ResetIM = ResetIM;
-    entry->DoInput = DoInput;
-    entry->GetCandWords = GetCandWords;
-    entry->PhraseTips = PhraseTips;
-    entry->Save = Save;
-    entry->ReloadConfig = ReloadConfig;
-    entry->KeyBlocker = KeyBlocker;
+    entry->Init = iface.Init;
+    entry->ResetIM = iface.ResetIM;
+    entry->DoInput = iface.DoInput;
+    entry->GetCandWords = iface.GetCandWords;
+    entry->PhraseTips = iface.PhraseTips;
+    entry->Save = iface.Save;
+    entry->ReloadConfig = iface.ReloadConfig;
+    entry->KeyBlocker = iface.KeyBlocker;
+    entry->UpdateSurroundingText = iface.UpdateSurroundingText;
+    entry->DoReleaseInput = iface.DoReleaseInput;
     entry->klass = imclass;
     entry->iPriority = priority;
     if (langCode)
         strncpy(entry->langCode, langCode, LANGCODE_LENGTH);
     entry->langCode[LANGCODE_LENGTH] = 0;
     entry->initialized = true;
+    entry->owner = instance->currentIMAddon;
 }
 
 CONFIG_DESC_DEFINE(GetIMConfigDesc, "inputmethod.desc")
@@ -406,11 +519,14 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
     FcitxInputState *input = instance->input;
 
     FcitxGlobalConfig *fc = instance->config;
-    
-    if (FcitxInstanceGetCurrentIC(instance) == NULL)
+
+    if (instance->CurrentIC == NULL)
         return IRV_TO_PROCESS;
-    
-    if (FcitxInstanceGetCurrentIC(instance)->contextCaps & CAPACITY_PASSWORD)
+
+    if (instance->CurrentIC->contextCaps & CAPACITY_PASSWORD)
+        return IRV_TO_PROCESS;
+
+    if (currentIM == NULL)
         return IRV_TO_PROCESS;
 
     /*
@@ -445,21 +561,15 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
                     }
 
                     input->keyReleased = KR_OTHER;
-                } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, FCITX_LCTRL_LSHIFT) || FcitxHotkeyIsHotKey(sym, state, FCITX_LCTRL_LSHIFT2))) {
-                    if (FcitxInstanceGetCurrentState(instance) == IS_ACTIVE) {
-                        if (input->keyReleased == KR_SWITCH_IM)
-                            FcitxInstanceSwitchIM(instance, -1);
-                    } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkTrigger)) {
-                            FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
+                } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWNextKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWNextKey2[fc->iIMSwitchKey]))) {
+                    if (input->keyReleased == KR_SWITCH_IM) {
+                        FcitxInstanceSwitchIMByIndex(instance, -1);
                     }
-                } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, FCITX_RCTRL_RSHIFT) || FcitxHotkeyIsHotKey(sym, state, FCITX_RCTRL_RSHIFT2))) {
-                    if (FcitxInstanceGetCurrentState(instance) == IS_ACTIVE) {
-                        if (input->keyReleased == KR_SWITCH_IM_REVERSE)
-                            FcitxInstanceSwitchIM(instance, -2);
-                    } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkTrigger)) {
-                            FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
+                } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWPrevKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWPrevKey2[fc->iIMSwitchKey]))) {
+                    if (input->keyReleased == KR_SWITCH_IM_REVERSE) {
+                        FcitxInstanceSwitchIMByIndex(instance, -2);
                     }
-                } else if (FcitxHotkeyIsHotKey(sym, state, fc->switchKey) && input->keyReleased == KR_SWITCH && !fc->bDoubleSwitchKey) {
+                } else if ((FcitxHotkeyIsHotKey(sym, state, switchKey1[fc->iSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, switchKey2[fc->iSwitchKey])) && input->keyReleased == KR_SWITCH && !fc->bDoubleSwitchKey) {
                     retVal = IRV_DONOT_PROCESS;
                     if (fc->bSendTextWhenSwitchEng) {
                         if (input->iCodeInputCount != 0) {
@@ -470,7 +580,7 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
                     input->keyReleased = KR_OTHER;
                     if (FcitxInstanceGetCurrentState(instance) == IS_INACTIVE)
                         FcitxInstanceShowInputSpeed(instance);
-                    FcitxInstanceChangeIMState(instance, FcitxInstanceGetCurrentIC(instance));
+                    FcitxInstanceChangeIMState(instance, instance->CurrentIC);
                 }
             }
         }
@@ -499,41 +609,58 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
                     return IRV_DO_NOTHING;
                 }
             }
-            if (!FcitxHotkeyIsHotKey(sym, state, fc->switchKey))
-                input->keyReleased = KR_OTHER;
+            if (!(FcitxHotkeyIsHotKey(sym, state, switchKey1[fc->iSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, switchKey2[fc->iSwitchKey])))
+                    input->keyReleased = KR_OTHER;
             else {
                 if ((input->keyReleased == KR_SWITCH)
                         && (timestamp - input->lastKeyPressedTime < fc->iTimeInterval)
                         && fc->bDoubleSwitchKey) {
-                    FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), FcitxInputStateGetRawInputBuffer(input));
-                    FcitxInstanceChangeIMState(instance, FcitxInstanceGetCurrentIC(instance));
+                    FcitxInstanceCommitString(instance, instance->CurrentIC, FcitxInputStateGetRawInputBuffer(input));
+                    FcitxInstanceChangeIMState(instance, instance->CurrentIC);
                 }
             }
 
-            if (FcitxHotkeyIsHotKey(sym, state, fc->switchKey)) {
+            if (FcitxHotkeyIsHotKey(sym, state, switchKey1[fc->iSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, switchKey2[fc->iSwitchKey])) {
                 input->keyReleased = KR_SWITCH;
-                retVal = IRV_DO_NOTHING;
-            } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, FCITX_LCTRL_LSHIFT) || FcitxHotkeyIsHotKey(sym, state, FCITX_LCTRL_LSHIFT2))) {
+                retVal = IRV_DONOT_PROCESS;
+            } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWNextKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWNextKey2[fc->iIMSwitchKey]))) {
                 input->keyReleased = KR_SWITCH_IM;
-                retVal = IRV_DO_NOTHING;
-            } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, FCITX_RCTRL_RSHIFT) || FcitxHotkeyIsHotKey(sym, state, FCITX_RCTRL_RSHIFT2))) {
+                retVal = IRV_DONOT_PROCESS;
+            } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWPrevKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWPrevKey2[fc->iIMSwitchKey]))) {
                 input->keyReleased = KR_SWITCH_IM_REVERSE;
-                retVal = IRV_DO_NOTHING;
+                retVal = IRV_DONOT_PROCESS;
             } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkTrigger)) {
                 /* trigger key has the highest priority, so we check it first */
                 if (FcitxInstanceGetCurrentState(instance) == IS_INACTIVE) {
-                    FcitxInstanceChangeIMState(instance, FcitxInstanceGetCurrentIC(instance));
+                    FcitxInstanceChangeIMState(instance, instance->CurrentIC);
                     FcitxInstanceShowInputSpeed(instance);
                 } else
-                    FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
+                    FcitxInstanceCloseIM(instance, instance->CurrentIC);
 
                 retVal = IRV_DO_NOTHING;
+            } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkActivate)) {
+                if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
+                    FcitxInstanceEnableIM(instance, instance->CurrentIC, false);
+                    retVal = IRV_DO_NOTHING;
+                }
+            } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkInactivate)) {
+                if (FcitxInstanceGetCurrentState(instance) == IS_ACTIVE) {
+                    FcitxInstanceCloseIM(instance, instance->CurrentIC);
+                    retVal = IRV_DO_NOTHING;
+                }
             }
         }
     }
 
-    if (retVal == IRV_TO_PROCESS && event != FCITX_PRESS_KEY)
-        retVal = IRV_DONOT_PROCESS;
+    if (retVal == IRV_TO_PROCESS && event == FCITX_RELEASE_KEY) {
+         if (currentIM && currentIM->DoReleaseInput) {
+            retVal = currentIM->DoReleaseInput(currentIM->klass, sym, state);
+            if (retVal == IRV_TO_PROCESS)
+                retVal = IRV_DONOT_PROCESS;
+        }
+        else
+            retVal = IRV_DONOT_PROCESS;
+    }
 
     /* the key processed before this phase is very important, we don't let any interrupt */
     if (FcitxInstanceGetCurrentStatev2(instance) == IS_ACTIVE
@@ -544,7 +671,7 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
         }
 
         if (retVal == IRV_TO_PROCESS) {
-            if (!FcitxHotkeyIsHotKey(sym, state, FCITX_LCTRL_LSHIFT) && currentIM) {
+            if (!FcitxHotkeyIsHotKey(sym, state, imSWNextKey1[fc->iIMSwitchKey]) && currentIM) {
                 retVal = currentIM->DoInput(currentIM->klass, sym, state);
             }
         }
@@ -622,11 +749,11 @@ INPUT_RETURN_VALUE FcitxInstanceDoInputCallback(
         const FcitxHotkey* hkPrevPage = FcitxInstanceGetContextHotkey(instance, CONTEXT_ALTERNATIVE_PREVPAGE_KEY);
         if (hkPrevPage == NULL)
             hkPrevPage = fc->hkPrevPage;
-        
+
         const FcitxHotkey* hkNextPage = FcitxInstanceGetContextHotkey(instance, CONTEXT_ALTERNATIVE_NEXTPAGE_KEY);
         if (hkNextPage == NULL)
             hkNextPage = fc->hkNextPage;
-        
+
         if (FcitxHotkeyIsHotKey(sym, state, hkPrevPage)) {
             if (FcitxCandidateWordGoPrevPage(input->candList))
                 retVal = IRV_DISPLAY_CANDWORDS;
@@ -638,7 +765,7 @@ INPUT_RETURN_VALUE FcitxInstanceDoInputCallback(
 
     if (FcitxInstanceGetCurrentStatev2(instance) == IS_ACTIVE && !input->bIsDoInputOnly && retVal == IRV_TO_PROCESS) {
         FcitxInstanceProcessPostInputFilter(instance, sym, state, &retVal);
-        
+
         if (retVal == IRV_TO_PROCESS) {
             if (currentIM && currentIM->KeyBlocker)
                 retVal = currentIM->KeyBlocker(currentIM->klass, sym, state);
@@ -651,8 +778,6 @@ INPUT_RETURN_VALUE FcitxInstanceDoInputCallback(
         retVal = FcitxInstanceProcessHotkey(instance, sym, state);
     }
 
-    FcitxLog(DEBUG, "ProcessKey Return State: %s", GetStateName(retVal));
-
     FcitxInstanceProcessInputReturnValue(instance, retVal);
 
     return retVal;
@@ -669,8 +794,7 @@ void FcitxInstanceProcessInputReturnValue(
     FcitxGlobalConfig *fc = instance->config;
 
     if (retVal & IRV_FLAG_PENDING_COMMIT_STRING) {
-        FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), FcitxInputStateGetOutputString(input));
-        instance->iHZInputed += (int)(fcitx_utf8_strlen(FcitxInputStateGetOutputString(input)));
+        FcitxInstanceCommitString(instance, instance->CurrentIC, FcitxInputStateGetOutputString(input));
     }
 
     if (retVal & IRV_FLAG_DO_PHRASE_TIPS) {
@@ -691,7 +815,7 @@ void FcitxInstanceProcessInputReturnValue(
     if (retVal & IRV_FLAG_DISPLAY_LAST) {
         FcitxInstanceCleanInputWindow(instance);
         FcitxMessagesAddMessageAtLast(input->msgAuxUp, MSG_INPUT, "%c", FcitxInputStateGetRawInputBuffer(input)[0]);
-        FcitxMessagesAddMessageAtLast(input->msgAuxDown, MSG_TIPS, "%s", FcitxInputStateGetOutputString(input));
+        FcitxMessagesAddMessageAtLast(input->msgAuxDown, MSG_TIPS, "%s", FcitxInputStateGetLastCommitString(input));
     }
 
     if (retVal & IRV_FLAG_UPDATE_INPUT_WINDOW)
@@ -714,16 +838,66 @@ void FcitxInstanceForwardKey(FcitxInstance* instance, FcitxInputContext *ic, Fci
 FCITX_EXPORT_API
 void FcitxInstanceSwitchIM(FcitxInstance* instance, int index)
 {
-    FcitxInstanceSwitchIMInternal(instance, index, instance->config->firstAsInactive);
+    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
+    if (ic)
+        FcitxInstanceSetLocalIMName(instance, ic, NULL);
+    FcitxInstanceSwitchIMInternal(instance, index, true, true);
 }
 
-void FcitxInstanceSwitchIMInternal(FcitxInstance* instance, int index, boolean skipZero)
+FCITX_EXPORT_API
+void FcitxInstanceSwitchIMByName(FcitxInstance* instance, const char* name)
+{
+    FcitxIM* im = FcitxInstanceGetIMFromIMList(instance, IMAS_Enable, name);
+    if (im) {
+        int idx = FcitxInstanceGetIMIndexByName(instance, name);
+        if (idx == 0)
+            FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
+        else {
+            FcitxInstanceSwitchIM(instance, idx);
+            if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
+                FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
+            }
+        }
+    }
+}
+
+FCITX_EXPORT_API
+void FcitxInstanceSwitchIMByIndex(FcitxInstance* instance, int index)
+{
+    UT_array* imes = &instance->imes;
+    int iIMCount = utarray_len(imes);
+    /* less than -2, invalid, set to zero
+     * -2 scroll back
+     * -1 scroll forward
+     * 0~positive select
+     */
+    if (index < -2 || index >= iIMCount)
+        return;
+    else if (index == -2) {
+        if (instance->iIMIndex > 0)
+            index = instance->iIMIndex -1;
+        else
+            index = iIMCount - 1;
+    } else if (index == -1) {
+        if (instance->iIMIndex >= (iIMCount - 1))
+            index = 0;
+        else
+            index = instance->iIMIndex + 1;
+    }
+    if (index == 0)
+        FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
+    else {
+        FcitxInstanceSwitchIM(instance, index);
+        if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
+            FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
+        }
+    }
+}
+
+void FcitxInstanceSwitchIMInternal(FcitxInstance* instance, int index, boolean skipZero, boolean updateGlobal)
 {
     UT_array* imes = &instance->imes;
     int iIMCount = utarray_len(imes);
-    int oldIMIndex = instance->iIMIndex;
-    if (instance->iIMIndex == 0 && instance->lastIMIndex != 0)
-        oldIMIndex = instance->lastIMIndex;
 
     FcitxInstanceCleanInputWindow(instance);
     FcitxInstanceResetInput(instance);
@@ -754,8 +928,12 @@ void FcitxInstanceSwitchIMInternal(FcitxInstance* instance, int index, boolean s
     } else if (index >= 0)
         instance->iIMIndex = index;
 
-    if (skipZero && instance->iIMIndex == 0)
-        instance->iIMIndex = 1;
+    if (skipZero && instance->iIMIndex == 0) {
+        if (index == -2)
+            instance->iIMIndex = iIMCount -1;
+        else
+            instance->iIMIndex = 1;
+    }
 
     if (instance->iIMIndex >= iIMCount || instance->iIMIndex < 0)
         newIM = NULL;
@@ -789,17 +967,13 @@ void FcitxInstanceSwitchIMInternal(FcitxInstance* instance, int index, boolean s
         newIM->Init(newIM->klass);
     }
 
-    if (instance->config->firstAsInactive && instance->iIMIndex != 0)
-        instance->lastIMIndex = instance->iIMIndex;
-
-    int newIMIndex = instance->iIMIndex;
-    if (instance->iIMIndex == 0 && instance->lastIMIndex != 0)
-        newIMIndex = instance->lastIMIndex;
+    if (newIM && updateGlobal) {
+        fcitx_utils_free(instance->globalIMName);
+        instance->globalIMName = strdup(newIM->uniqueName);
+        FcitxProfileSave(instance->profile);
+    }
 
     FcitxInstanceResetInput(instance);
-
-    if (newIMIndex != oldIMIndex)
-        FcitxProfileSave(instance->profile);
 }
 
 /**
@@ -951,6 +1125,65 @@ FcitxIM* FcitxInstanceGetCurrentIM(FcitxInstance* instance)
     return pcurrentIM;
 }
 
+FCITX_EXPORT_API
+void FcitxInstanceSetLocalIMName(FcitxInstance* instance, FcitxInputContext* ic, const char* imname)
+{
+    FcitxInputContext2* ic2 = (FcitxInputContext2*) ic;
+    if (ic2->imname) {
+        free(ic2->imname);
+        ic2->imname = NULL;
+    }
+
+    if (imname)
+        ic2->imname = strdup(imname);
+
+    if (ic == FcitxInstanceGetCurrentIC(instance))
+        FcitxInstanceUpdateCurrentIM(instance, false);
+}
+
+void FcitxInstanceUpdateCurrentIM(FcitxInstance* instance, boolean force) {
+    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
+    if (!ic && !force)
+        return;
+    FcitxInputContext2* ic2 = (FcitxInputContext2*) ic;
+    int globalIndex = FcitxInstanceGetIMIndexByName(instance, instance->globalIMName);
+    boolean forceSwtich = force;
+    /* fix it here */
+    if (globalIndex == 0) {
+        UT_array* ime = &instance->imes;
+        FcitxIM* im = (FcitxIM*) utarray_eltptr(ime, 1);
+        if (im) {
+            fcitx_utils_string_swap(&instance->globalIMName, im->uniqueName);
+            globalIndex = 1;
+            forceSwtich = true;
+        }
+    }
+    int targetIMIndex = 0;
+    boolean skipZero = false;
+
+    if (ic2 && ic2->imname) {
+        FcitxIM* im = FcitxInstanceGetIMFromIMList(instance, IMAS_Enable, ic2->imname);
+        if (!im) {
+            free(ic2->imname);
+            ic2->imname = NULL;
+        }
+    }
+
+    if (ic && ic->state != IS_ACTIVE) {
+        targetIMIndex = 0;
+    }
+    else {
+        if (ic2 && ic2->imname)
+            targetIMIndex = FcitxInstanceGetIMIndexByName(instance, ic2->imname);
+        else
+            targetIMIndex = globalIndex;
+        skipZero = true;
+    }
+
+    if (forceSwtich || targetIMIndex != instance->iIMIndex)
+        FcitxInstanceSwitchIMInternal(instance, targetIMIndex, skipZero, forceSwtich);
+}
+
 FCITX_EXPORT_API
 void FcitxInstanceEnableIM(FcitxInstance* instance, FcitxInputContext* ic, boolean keepState)
 {
@@ -966,14 +1199,7 @@ void FcitxInstanceEnableIM(FcitxInstance* instance, FcitxInputContext* ic, boole
             if (instance->config->shareState == ShareState_All)
                 flag = true;
             else {
-                UT_array* frontends = &instance->frontends;
-                FcitxAddon** pfrontend = (FcitxAddon**) utarray_eltptr(frontends, ic->frontendid);
-                if (pfrontend) {
-                    FcitxFrontend* frontend = (*pfrontend)->frontend;
-                    if (frontend->CheckICFromSameApplication &&
-                            frontend->CheckICFromSameApplication((*pfrontend)->addonInstance, rec, ic))
-                        flag = true;
-                }
+                flag = FcitxInstanceCheckICFromSameApplication(instance, rec, ic);
             }
 
             if (flag && (rec == ic || !(rec->contextCaps & CAPACITY_CLIENT_SIDE_CONTROL_STATE)))
@@ -986,13 +1212,8 @@ void FcitxInstanceEnableIM(FcitxInstance* instance, FcitxInputContext* ic, boole
         FcitxInstanceEnableIMInternal(instance, ic, keepState);
         break;
     }
-    
-    if (instance->config->firstAsInactive
-        && FcitxInstanceGetCurrentState(instance) == IS_ACTIVE
-        && instance->iIMIndex == 0
-    ) {
-        FcitxInstanceSwitchIM(instance, instance->lastIMIndex);
-    }
+
+    FcitxInstanceUpdateCurrentIM(instance, false);
     instance->input->keyReleased = KR_OTHER;
 }
 
@@ -1011,7 +1232,7 @@ void FcitxInstanceEnableIMInternal(FcitxInstance* instance, FcitxInputContext* i
     if (oldstate == IS_CLOSED)
         frontend->EnableIM((*pfrontend)->addonInstance, ic);
 
-    if (ic == FcitxInstanceGetCurrentIC(instance)) {
+    if (ic == instance->CurrentIC) {
         if (oldstate == IS_CLOSED)
             FcitxUIOnTriggerOn(instance);
         if (!keepState)
@@ -1024,12 +1245,13 @@ void FcitxInstanceCloseIM(FcitxInstance* instance, FcitxInputContext* ic)
 {
     if (ic == NULL)
         return;
-    
-    if (instance->config->firstAsInactive && !(ic->contextCaps & CAPACITY_CLIENT_SIDE_CONTROL_STATE)) {
-        FcitxInstanceChangeIMState(instance, ic);
+
+    if (!(ic->contextCaps & CAPACITY_CLIENT_SIDE_CONTROL_STATE)) {
+        if (ic->state == IS_ACTIVE)
+            FcitxInstanceChangeIMState(instance, ic);
         return;
     }
-    
+
     instance->globalState = IS_CLOSED;
     switch (instance->config->shareState) {
     case ShareState_All:
@@ -1040,14 +1262,7 @@ void FcitxInstanceCloseIM(FcitxInstance* instance, FcitxInputContext* ic)
             if (instance->config->shareState == ShareState_All)
                 flag = true;
             else {
-                UT_array* frontends = &instance->frontends;
-                FcitxAddon** pfrontend = (FcitxAddon**) utarray_eltptr(frontends, ic->frontendid);
-                if (pfrontend) {
-                    FcitxFrontend* frontend = (*pfrontend)->frontend;
-                    if (frontend->CheckICFromSameApplication &&
-                            frontend->CheckICFromSameApplication((*pfrontend)->addonInstance, rec, ic))
-                        flag = true;
-                }
+                flag = FcitxInstanceCheckICFromSameApplication(instance, rec, ic);
             }
 
             if (flag && (rec == ic || !(rec->contextCaps & CAPACITY_CLIENT_SIDE_CONTROL_STATE)))
@@ -1074,9 +1289,10 @@ void FcitxInstanceCloseIMInternal(FcitxInstance* instance, FcitxInputContext* ic
     ic->state = IS_CLOSED;
     frontend->CloseIM((*pfrontend)->addonInstance, ic);
 
-    if (ic == FcitxInstanceGetCurrentIC(instance)) {
+    if (ic == instance->CurrentIC) {
         FcitxUIOnTriggerOff(instance);
         FcitxUICloseInputWindow(instance);
+        FcitxInstanceResetInput(instance);
     }
 }
 
@@ -1095,14 +1311,6 @@ void FcitxInstanceChangeIMState(FcitxInstance* instance, FcitxInputContext* ic)
         objectState = IS_ACTIVE;
     else
         objectState = IS_INACTIVE;
-    
-    if (instance->config->firstAsInactive) {
-        if (objectState == IS_ACTIVE)
-            FcitxInstanceSwitchIM(instance, instance->lastIMIndex);
-        else {
-            FcitxInstanceSwitchIMInternal(instance, 0, false);
-        }
-    }
 
     instance->globalState = objectState;
     switch (instance->config->shareState) {
@@ -1114,14 +1322,7 @@ void FcitxInstanceChangeIMState(FcitxInstance* instance, FcitxInputContext* ic)
             if (instance->config->shareState == ShareState_All)
                 flag = true;
             else {
-                UT_array* frontends = &instance->frontends;
-                FcitxAddon** pfrontend = (FcitxAddon**) utarray_eltptr(frontends, ic->frontendid);
-                if (pfrontend) {
-                    FcitxFrontend* frontend = (*pfrontend)->frontend;
-                    if (frontend->CheckICFromSameApplication &&
-                            frontend->CheckICFromSameApplication((*pfrontend)->addonInstance, rec, ic))
-                        flag = true;
-                }
+                flag = FcitxInstanceCheckICFromSameApplication(instance, rec, ic);
             }
 
             if (flag && (rec == ic || !(rec->contextCaps & CAPACITY_CLIENT_SIDE_CONTROL_STATE)))
@@ -1134,6 +1335,8 @@ void FcitxInstanceChangeIMState(FcitxInstance* instance, FcitxInputContext* ic)
         FcitxInstanceChangeIMStateInternal(instance, ic, objectState);
         break;
     }
+
+    FcitxInstanceUpdateCurrentIM(instance, false);
 }
 
 void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputContext* ic, FcitxContextState objectState)
@@ -1143,7 +1346,7 @@ void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputConte
     if (ic->state == objectState)
         return;
     ic->state = objectState;
-    if (ic == FcitxInstanceGetCurrentIC(instance)) {
+    if (ic == instance->CurrentIC) {
         if (objectState == IS_ACTIVE) {
             FcitxInstanceResetInput(instance);
         } else {
@@ -1156,7 +1359,7 @@ void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputConte
 void FcitxInstanceInitIMMenu(FcitxInstance* instance)
 {
     FcitxMenuInit(&instance->imMenu);
-    instance->imMenu.candStatusBind = strdup("im");
+    instance->imMenu.candStatusBind = NULL;
     instance->imMenu.name = strdup(_("Input Method"));
 
     instance->imMenu.UpdateMenu = UpdateIMMenuItem;
@@ -1168,7 +1371,15 @@ void FcitxInstanceInitIMMenu(FcitxInstance* instance)
 boolean IMMenuAction(FcitxUIMenu *menu, int index)
 {
     FcitxInstance* instance = (FcitxInstance*) menu->priv;
-    FcitxInstanceSwitchIM(instance, index);
+
+    if (index == 0)
+        FcitxInstanceCloseIM(instance, instance->CurrentIC);
+    else {
+        FcitxInstanceSwitchIM(instance, index);
+        if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
+            FcitxInstanceEnableIM(instance, instance->CurrentIC, false);
+        }
+    }
     return true;
 }
 
@@ -1197,9 +1408,13 @@ void FcitxInstanceShowInputSpeed(FcitxInstance* instance)
     input->bShowCursor = false;
 
     FcitxInstanceCleanInputWindow(instance);
+
+    FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
+    if (!im)
+        return;
+
     if (instance->config->bShowVersion) {
         FcitxMessagesAddMessageAtLast(input->msgAuxUp, MSG_TIPS, "FCITX " VERSION);
-        FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
         if (im) {
             FcitxMessagesAddMessageAtLast(input->msgAuxUp, MSG_TIPS, " %s", im->strName);
         }
@@ -1310,6 +1525,32 @@ int FcitxInstanceGetIMIndexByName(FcitxInstance* instance, const char* imName)
     return index;
 }
 
+FCITX_EXPORT_API
+void FcitxInstanceNotifyUpdateSurroundingText(FcitxInstance* instance, FcitxInputContext* ic)
+{
+    if (!ic)
+        return;
+    if (ic != instance->CurrentIC)
+        return;
+
+    FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
+    if (!im)
+        return;
+
+    if (im->UpdateSurroundingText)
+        im->UpdateSurroundingText(im->klass);
+}
+
+void UnusedIMItemFreeAll(UnusedIMItem* item)
+{
+    UnusedIMItem *cur;
+    while (item) {
+        cur = item;
+        HASH_DEL(item, cur);
+        fcitx_utils_free(cur->name);
+    }
+}
+
 FCITX_EXPORT_API
 void FcitxInstanceUpdateIMList(FcitxInstance* instance)
 {
@@ -1319,6 +1560,8 @@ void FcitxInstanceUpdateIMList(FcitxInstance* instance)
     UT_array* imList = fcitx_utils_split_string(instance->profile->imList, ',');
     utarray_sort(&instance->availimes, IMPriorityCmp);
     utarray_clear(&instance->imes);
+    UnusedIMItemFreeAll(instance->unusedItem);
+    instance->unusedItem = NULL;
 
     char** pstr;
     FcitxIM* ime;
@@ -1331,11 +1574,21 @@ void FcitxInstanceUpdateIMList(FcitxInstance* instance)
             ime = NULL;
             *pos = '\0';
             pos ++;
-            if (strcmp(pos, "True") == 0)
-                ime = FcitxInstanceGetIMFromIMList(instance, IMAS_Disable, str);
-
-            if (ime)
+            ime = FcitxInstanceGetIMFromIMList(instance, IMAS_Disable, str);
+            boolean status = (strcmp(pos, "True") == 0);
+            if (status && ime)
                 utarray_push_back(&instance->imes, ime);
+
+            if (!ime) {
+                UnusedIMItem* item;
+                HASH_FIND_STR(instance->unusedItem, str, item);
+                if (!item) {
+                    item = fcitx_utils_new(UnusedIMItem);
+                    item->name = strdup(str);
+                    item->status = status;
+                    HASH_ADD_KEYPTR(hh, instance->unusedItem, item->name, strlen(item->name), item);
+                }
+            }
         }
     }
 
@@ -1344,18 +1597,20 @@ void FcitxInstanceUpdateIMList(FcitxInstance* instance)
             ime = (FcitxIM*) utarray_next(&instance->availimes, ime)) {
         if (!IMIsInIMNameList(imList, ime)) {
             /* ok, make all im priority larger than 100 disable by default */
-            if (ime->iPriority < PRIORITY_DISABLE)
+            if (ime->iPriority == 0)
+                utarray_push_front(&instance->imes, ime);
+            else if (ime->iPriority < PRIORITY_DISABLE)
                 utarray_push_back(&instance->imes, ime);
         }
     }
 
     utarray_free(imList);
 
-    instance->iIMIndex = FcitxInstanceGetIMIndexByName(instance, instance->profile->imName);
-
-    FcitxInstanceSwitchIM(instance, instance->iIMIndex);
+    FcitxInstanceUpdateCurrentIM(instance, true);
     FcitxInstanceProcessUpdateIMListHook(instance);
-    FcitxProfileSave(instance->profile);
+
+    if (instance->globalIMName)
+        FcitxProfileSave(instance->profile);
 }
 
 FCITX_EXPORT_API
@@ -1401,6 +1656,7 @@ void FreeIMEntry(FcitxIMEntry* entry)
     free(entry);
 }
 
+FCITX_EXPORT_API
 INPUT_RETURN_VALUE FcitxStandardKeyBlocker(FcitxInputState* input, FcitxKeySym key, unsigned int state)
 {
     if (FcitxInputStateGetRawInputBufferSize(input) != 0
