@@ -30,6 +30,25 @@
  *
  *  Public Header for Input Method Develop
  *
+ * The input method is key event centric application. When a key event comes to fcitx,
+ * the process handling the keyboard event can be separated into 7 phases, PreInput, DoInput, Update Candidates, Prev/Next Page
+ * PostInput, Hotkey, and key blocker.
+ * 
+ * The input method engine process key event inside "DoInput".
+ * 
+ * Each phase will change the INPUT_RETURN_VALUE, if INPUT_RETURN_VALUE is non-zero (non IRV_TO_PROCESS), the following
+ * phases will not run.
+ * 
+ * If a key event goes through all phase and still the state is IRV_TO_PROCESS, it will be forwarded.
+ * 
+ * When it comes to update candidates, if the flag contains IRV_FLAG_UPDATE_CANDIDATE_WORDS, it will trigger the GetCandWords
+ * of input method engine and clean up all stiring in the input window, after that, it will trigger update candidates hook.
+ * 
+ * Key blocker is useful if you want to post input phase to do something, but you don't want forward key if they do nothing.
+ * There is an default implemention inside fcitx, it will blocks key when raw input buffer is not empty. Those keys
+ * contains direction key(left/right..), key will cause something input (a,b,c...), key will cause cursor move (home/end...).
+ * 
+ * DoInput, Update Candidates, Key Blocker are belongs to input method engine, other can be registered by other addon.
  */
 #ifndef _FCITX_IME_H_
 #define _FCITX_IME_H_
@@ -65,11 +84,14 @@ extern "C" {
 /** when input method priority is larger than 100, it will be disabled by default after install */
 #define PRIORITY_DISABLE 100
 
+/** due to backward compatible, this priority will be the most priority one */
+#define PRIORITY_MAGIC_FIRST 0xf1527
+
     struct _FcitxInputContext;
     struct _FcitxInstance;
     struct _FcitxAddon;
     struct _FcitxCandidateWordList;
-    
+
     /** input method available status */
     typedef enum _FcitxIMAvailableStatus {
         IMAS_Enable,
@@ -121,6 +143,26 @@ extern "C" {
     typedef void (*FcitxIMSave)(void *arg); /**< FcitxIMSave */
     typedef void (*FcitxIMReloadConfig)(void *arg); /**< FcitxIMReloadConfig */
     typedef INPUT_RETURN_VALUE (*FcitxIMKeyBlocker)(void* arg, FcitxKeySym, unsigned int); /**< FcitxIMKeyBlocker */
+    typedef void (*FcitxIMUpdateSurroundingText)(void* arg); /**< FcitxIMKeyBlocker */
+
+    /**
+     * a more fexible interface for input method
+     *
+     * @since 4.2.3
+     **/
+    typedef struct _FcitxIMIFace {
+        FcitxIMResetIM ResetIM /**< Reset input method state */;
+        FcitxIMDoInput DoInput /**< process key input */;
+        FcitxIMGetCandWords GetCandWords; /**< get candidate words */
+        FcitxIMPhraseTips PhraseTips; /**< don't use it */
+        FcitxIMSave Save; /**< force save input method data */
+        FcitxIMInit Init; /**< called when switch to this input method */
+        FcitxIMReloadConfig ReloadConfig; /**< reload configuration */
+        FcitxIMKeyBlocker KeyBlocker; /**< block unused key */
+        FcitxIMUpdateSurroundingText UpdateSurroundingText; /**< surrounding text update trigger */
+        FcitxIMDoInput DoReleaseInput;
+        void* padding[63]; /**< padding */
+    } FcitxIMIFace;
 
     /**
      * Fcitx Input method instance
@@ -196,7 +238,11 @@ extern "C" {
          **/
         FcitxIMKeyBlocker KeyBlocker;
 
-        void* padding[10]; /**< padding */
+        FcitxIMUpdateSurroundingText UpdateSurroundingText; /**< called when surrounding text updated */
+
+        FcitxIMDoInput DoReleaseInput;
+
+        void* padding[8]; /**< padding */
     } FcitxIM;
 
     /** a key event is press or release */
@@ -225,6 +271,16 @@ extern "C" {
      **/
     char* FcitxInputStateGetOutputString(FcitxInputState* input);
 
+    /**
+     * @brief get last commit string
+     *
+     * @param input input state
+     * @return const char*
+     *
+     * @since 4.2.3
+     **/
+    const char* FcitxInputStateGetLastCommitString(FcitxInputState * input);
+
     /**
      * get current input method
      *
@@ -341,6 +397,34 @@ extern "C" {
                            int priority,
                            const char *langCode
                           );
+
+    /**
+     * register a new input method
+     *
+     * @param instance fcitx instance
+     * @param imclass pointer to input method class
+     * @param uniqueName uniqueName which cannot be duplicated to others
+     * @param name input method name
+     * @param iconName icon name
+     * @param iface interface
+     * @param priority order of this input method
+     * @param langCode language code for this input method
+     * @return void
+     *
+     * @see FcitxInstanceRegisterIMv2
+     *
+     * @since 4.2.3
+     **/
+    void FcitxInstanceRegisterIMv2(struct _FcitxInstance *instance,
+                       void *imclass,
+                       const char* uniqueName,
+                       const char* name,
+                       const char* iconName,
+                       FcitxIMIFace iface,
+                       int priority,
+                       const char *langCode
+                      );
+
     /**
      * process a key event, should only used by frontend
      *
@@ -371,8 +455,8 @@ extern "C" {
         long unsigned int timestamp,
         FcitxKeySym sym,
         unsigned int state);
-    
-    
+
+
     /**
      * @brief choose candidate by index
      *
@@ -413,7 +497,11 @@ extern "C" {
     void FcitxInstanceReloadConfig(struct _FcitxInstance* instance);
 
     /**
-     * switch to input method by index
+     * switch to input method by index, if index is zero, it will be skipped
+     *
+     * @deprecated
+     *
+     * @see FcitxInstanceSwitchIMByIndex
      *
      * @param instance fcitx instance
      * @param index input method index
@@ -421,6 +509,33 @@ extern "C" {
      **/
     void FcitxInstanceSwitchIM(struct _FcitxInstance* instance, int index);
 
+
+    /**
+     * switch to a input method by name, name need to be valid, otherwise it have no effect
+     * And if the index is zero, the state will automatically change to inactive
+     *
+     * @param instance fcitx instance
+     * @param name ...
+     * @return void
+     *
+     * @since 4.2.4
+     **/
+    void FcitxInstanceSwitchIMByName(struct _FcitxInstance* instance, const char* name);
+
+    /**
+     * switch to a input method by index, index need to be valid, otherwise it have no effect
+     * And if the object index is zero, the state will automatically change to inactive
+     * -1 means scroll forward, and -2 means scroll back.
+     *
+     *
+     * @param instance fcitx instance
+     * @param name ...
+     * @return void
+     *
+     * @since 4.2.4
+     **/
+    void FcitxInstanceSwitchIMByIndex(struct _FcitxInstance* instance, int index);
+
     /**
      * check is choose key or not, if so, return the choose index
      *
@@ -441,14 +556,14 @@ extern "C" {
      * @return int
      **/
     int FcitxHotkeyCheckChooseKeyAndModifier(FcitxKeySym sym, int state, const char* strChoose, int candState);
-    
+
     /**
      * get im index by im name
      *
      * @param instance fcitx instance
      * @param imName im name
      * @return int im index
-     * 
+     *
      * @since 4.2
      **/
     int FcitxInstanceGetIMIndexByName(struct _FcitxInstance* instance, const char* imName);
@@ -619,7 +734,7 @@ extern "C" {
      * @return void
      **/
     void FcitxInputStateSetLastIsSingleChar(FcitxInputState* input, int lastIsSingleChar);
-    
+
     /**
      * set keycode for current key event
      *
@@ -628,7 +743,7 @@ extern "C" {
      * @return void
      **/
     void FcitxInputStateSetKeyCode( FcitxInputState* input, uint32_t value );
-    
+
     /**
      * set keysym for current key event
      *
@@ -642,11 +757,11 @@ extern "C" {
      * set keystate for current key state
      *
      * @param input input state
-     * @param value state
+     * @param state key state
      * @return void
      **/
     void FcitxInputStateSetKeyState( FcitxInputState* input, uint32_t state );
-    
+
     /**
      * get keycode for current key event
      *
@@ -654,7 +769,7 @@ extern "C" {
      * @return uint32_t
      **/
     uint32_t FcitxInputStateGetKeyCode( FcitxInputState* input);
-    
+
     /**
      * get keysym for current key event
      *
@@ -689,6 +804,34 @@ extern "C" {
      **/
     void FcitxInstanceUpdateIMList(struct _FcitxInstance* instance);
 
+    /**
+     * notify surrounding text changed to im
+     *
+     * @param instance instance
+     * @param ic ic
+     * @return void
+     **/
+    void FcitxInstanceNotifyUpdateSurroundingText(struct _FcitxInstance* instance, struct _FcitxInputContext* ic);
+
+    /**
+     * an standard key blocker, block all the key that cause cursor move when raw input buffer is not empty.
+     *
+     * @param input input state
+     * @param key keysym
+     * @param state key state
+     * @return INPUT_RETURN_VALUE
+     **/
+    INPUT_RETURN_VALUE FcitxStandardKeyBlocker(FcitxInputState* input, FcitxKeySym key, unsigned int state);
+
+    /**
+     * set local input method name
+     *
+     * @param ic name
+     * @param imname im name
+     * @return void
+     **/
+    void FcitxInstanceSetLocalIMName(struct _FcitxInstance* instance, struct _FcitxInputContext* ic, const char* imname);
+
 #ifdef __cplusplus
 }
 #endif
