@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2011~2011 by CSSlayer                                   *
+ *   Copyright (C) 2011~2012 by CSSlayer                                   *
  *   wengxt@gmail.com                                                      *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
@@ -36,6 +36,12 @@
 #include "addon.h"
 #include "context.h"
 
+typedef struct _UnusedIMItem {
+    char* name;
+    boolean status;
+    UT_hash_handle hh;
+} UnusedIMItem;
+
 struct _FcitxInstance {
     pthread_mutex_t fcitxMutex;
     UT_array uistats;
@@ -56,6 +62,7 @@ struct _FcitxInstance {
     UT_array imeclasses;
     UT_array imes;
     UT_array frontends;
+    UT_array modules;
     UT_array eventmodules;
 
     struct _FcitxInputContext *CurrentIC;
@@ -97,14 +104,18 @@ struct _FcitxInstance {
     FcitxAddon* uifallback;
 
     FcitxAddon* uinormal;
-    
+
     FcitxContext* context;
-    
+
     boolean tryReplace;
-    
-    int lastIMIndex;
-    
+
+    char* globalIMName;
+
     char* fallbackuiName;
+
+    FcitxAddon* currentIMAddon;
+
+    UnusedIMItem* unusedItem;
 };
 
 #endif
