@@ -158,27 +158,13 @@ void FcitxUILoad(FcitxInstance* instance)
 
     instance->ui = instance->uinormal;
 
-    if (instance->ui == NULL)
+    if (instance->ui == NULL) {
         FcitxLog(ERROR, "no usable user interface.");
-    else {
-        do {
-            FcitxAddon* fallbackAddon = FcitxAddonsGetAddonByName(&instance->addons, addon->uifallback);
-            if (fallbackAddon == NULL)
-                break;
-
-            if (!fallbackAddon->bEnabled)
-                break;
-            
-            instance->fallbackuiName = strdup(addon->uifallback);
-
-            /* ui is fallback */
-            if (FcitxUILoadInternal(instance, fallbackAddon)) {
-                instance->uifallback = fallbackAddon;
-                if (instance->uifallback->ui->Suspend)
-                    instance->uifallback->ui->Suspend(instance->uifallback->addonInstance);
-            }
-        } while (0);
+        return;
     }
+
+    if (addon->uifallback)
+        instance->fallbackuiName = strdup(addon->uifallback);
 }
 
 boolean FcitxUILoadInternal(FcitxInstance* instance, FcitxAddon* addon)
@@ -283,7 +269,7 @@ void FcitxMessagesAddMessageAtLast(FcitxMessages* message, FcitxMessageType type
 }
 
 FCITX_EXPORT_API
-void FcitxMessagesSetMessage(FcitxMessages* message, int position, FcitxMessageType type, const char* fmt, ...)
+void FcitxMessagesSetMessage(FcitxMessages* message, int position, int type, const char* fmt, ...)
 {
     va_list ap;
     va_start(ap, fmt);
@@ -301,7 +287,7 @@ void FcitxMessagesSetMessageText(FcitxMessages* message, int position, const cha
 }
 
 FCITX_EXPORT_API
-void FcitxMessagesSetMessageV(FcitxMessages* message, int position, FcitxMessageType type, const char* fmt, va_list ap)
+void FcitxMessagesSetMessageV(FcitxMessages* message, int position, int type, const char* fmt, va_list ap)
 {
     if (position < MAX_MESSAGE_COUNT) {
         vsnprintf(message->msg[position].strMsg, MESSAGE_MAX_LENGTH, fmt, ap);
@@ -400,7 +386,7 @@ void FcitxUIUpdateStatus(FcitxInstance* instance, const char* name)
         FcitxUIComplexStatus *compstatus = FcitxUIGetComplexStatusByName(instance, name);
         if (!compstatus)
             return;
-        
+
         if (compstatus->toggleStatus)
             compstatus->toggleStatus(compstatus->arg);
         if (UI_FUNC_IS_VALID(UpdateComplexStatus))
@@ -418,7 +404,7 @@ void FcitxUISetStatusString(FcitxInstance* instance, const char* name, const cha
         compstatus = FcitxUIGetComplexStatusByName(instance, name);
         if (!compstatus)
             return;
-        
+
         pShort = &compstatus->shortDescription;
         pLong = &compstatus->longDescription;
     }
@@ -426,16 +412,16 @@ void FcitxUISetStatusString(FcitxInstance* instance, const char* name, const cha
         pShort = &status->shortDescription;
         pLong = &status->longDescription;
     }
-    
+
     if (*pShort)
         free(*pShort);
-    
+
     if (*pLong)
         free(*pLong);
-    
+
     *pShort = strdup(shortDesc);
     *pLong = strdup(longDesc);
-    
+
     if (status){
         if (UI_FUNC_IS_VALID(UpdateStatus))
             instance->ui->ui->UpdateStatus(instance->ui->addonInstance, status);
@@ -454,7 +440,7 @@ void FcitxUISetStatusVisable(FcitxInstance* instance, const char* name, boolean
         FcitxUIComplexStatus *compstatus = FcitxUIGetComplexStatusByName(instance, name);
         if (!compstatus)
             return;
-        
+
         if (compstatus->visible != visible) {
             compstatus->visible = visible;
 
@@ -463,7 +449,7 @@ void FcitxUISetStatusVisable(FcitxInstance* instance, const char* name, boolean
         }
         return;
     }
-    
+
     if (status->visible != visible) {
         status->visible = visible;
 
@@ -491,7 +477,7 @@ void FcitxUIRegisterStatus(
     status.longDescription = strdup(longDesc);
     status.getCurrentStatus = getStatus;
     status.toggleStatus = toggleStatus;
-    status.arg = arg;    
+    status.arg = arg;
     status.visible = true;
 
     UT_array* uistats = &instance->uistats;
@@ -522,7 +508,7 @@ void FcitxUIRegisterComplexStatus(
     compstatus.longDescription = strdup(longDesc);
     compstatus.getIconName = getIconName;
     compstatus.toggleStatus = toggleStatus;
-    compstatus.arg = arg;    
+    compstatus.arg = arg;
     compstatus.visible = true;
 
     UT_array* uicompstats = &instance->uicompstats;
@@ -543,6 +529,7 @@ void FcitxUIRegisterMenu(FcitxInstance* instance, FcitxUIMenu* menu)
         return ;
 
     menu->mark = -1;
+    menu->visible = true;
 
     utarray_push_back(uimenus, &menu);
     if (UI_FUNC_IS_VALID(RegisterMenu))
@@ -552,7 +539,7 @@ void FcitxUIRegisterMenu(FcitxInstance* instance, FcitxUIMenu* menu)
 }
 
 FCITX_EXPORT_API
-void FcitxMenuAddMenuItem(FcitxUIMenu* menu, const char* string, FcitxMenuItemType type, FcitxUIMenu* subMenu)
+void FcitxMenuAddMenuItemWithData(FcitxUIMenu* menu, const char* string, FcitxMenuItemType type, FcitxUIMenu* subMenu, void* arg)
 {
     FcitxMenuItem shell;
     memset(&shell, 0, sizeof(FcitxMenuItem));
@@ -566,6 +553,7 @@ void FcitxMenuAddMenuItem(FcitxUIMenu* menu, const char* string, FcitxMenuItemTy
         shell.tipstr = NULL;
 
     shell.type = type;
+    shell.data = arg;
 
     shell.isselect = false;
 
@@ -575,6 +563,13 @@ void FcitxMenuAddMenuItem(FcitxUIMenu* menu, const char* string, FcitxMenuItemTy
     utarray_push_back(&menu->shell, &shell);
 }
 
+
+FCITX_EXPORT_API
+void FcitxMenuAddMenuItem(FcitxUIMenu* menu, const char* string, FcitxMenuItemType type, FcitxUIMenu* subMenu)
+{
+    FcitxMenuAddMenuItemWithData(menu, string, type, subMenu, NULL);
+}
+
 FCITX_EXPORT_API
 void FcitxMenuClear(FcitxUIMenu* menu)
 {
@@ -590,15 +585,8 @@ void FcitxUIOnInputFocus(FcitxInstance* instance)
     FcitxInstanceProcessInputFocusHook(instance);
 
     FcitxInstanceResetInput(instance);
-    if (instance->config->firstAsInactive) {
-        if (FcitxInstanceGetCurrentState(instance) == IS_ACTIVE)
-            FcitxInstanceSwitchIM(instance, instance->lastIMIndex);
-        else if (FcitxInstanceGetCurrentState(instance) == IS_INACTIVE) {
-            if (instance->iIMIndex != 0)
-                instance->lastIMIndex = instance->iIMIndex;
-            FcitxInstanceSwitchIMInternal(instance, 0, false);
-        }
-    }
+
+    FcitxInstanceUpdateCurrentIM(instance, false);
 
     FcitxUICloseInputWindow(instance);
 }
@@ -612,6 +600,26 @@ void FcitxUIOnInputUnFocus(struct _FcitxInstance* instance)
     FcitxInstanceProcessInputUnFocusHook(instance);
 }
 
+FCITX_EXPORT_API
+void FcitxUICommitPreedit(struct _FcitxInstance* instance)
+{
+    if (instance->CurrentIC
+        && !instance->config->bDontCommitPreeditWhenUnfocus
+        && !(instance->CurrentIC->contextCaps & CAPACITY_CLIENT_UNFOCUS_COMMIT)) {
+        FcitxInputState* input = FcitxInstanceGetInputState(instance);
+        FcitxMessages* clientPreedit = FcitxInputStateGetClientPreedit(input);
+
+        if (FcitxMessagesGetMessageCount(clientPreedit) > 0) {
+            char* str = FcitxUIMessagesToCStringForCommit(clientPreedit);
+            if (str[0]) {
+                FcitxInstanceCommitString(instance, instance->CurrentIC, str);
+            }
+            free(str);
+        }
+        FcitxMessagesSetMessageCount(clientPreedit, 0);
+    }
+}
+
 FCITX_EXPORT_API
 void FcitxUIOnTriggerOn(FcitxInstance* instance)
 {
@@ -717,7 +725,9 @@ int FcitxUINewMessageToOldStyleMessage(FcitxInstance* instance, FcitxMessages* m
 
         if (i == 0
             && FcitxCandidateWordGetCurrentPage(input->candList) == 0
-            && type == MSG_OTHER)
+            && type == MSG_OTHER
+            && !FcitxInstanceGetContextBoolean(instance, CONTEXT_DISABLE_AUTO_FIRST_CANDIDATE_HIGHTLIGHT)
+        )
             type = MSG_FIRSTCAND;
 
         FcitxMessagesAddMessageAtLast(msgDown, type, "%s", candWord->strWord);
@@ -748,6 +758,27 @@ char* FcitxUIMessagesToCString(FcitxMessages* messages)
     return str;
 }
 
+char* FcitxUIMessagesToCStringForCommit(FcitxMessages* messages)
+{
+    int length = 0;
+    int i = 0;
+
+    for (i = 0; i < FcitxMessagesGetMessageCount(messages) ; i ++) {
+        if ((FcitxMessagesGetClientMessageType(messages, i) & MSG_DONOT_COMMIT_WHEN_UNFOCUS) == 0)
+            length += strlen(FcitxMessagesGetMessageString(messages, i));
+    }
+
+    char* str = fcitx_utils_malloc0(sizeof(char) * (length + 1));
+
+    for (i = 0; i < FcitxMessagesGetMessageCount(messages) ; i ++) {
+        if ((FcitxMessagesGetClientMessageType(messages, i) & MSG_DONOT_COMMIT_WHEN_UNFOCUS) == 0)
+            strcat(str, FcitxMessagesGetMessageString(messages, i));
+    }
+
+    return str;
+}
+
+
 FCITX_EXPORT_API
 char* FcitxUICandidateWordToCString(FcitxInstance* instance)
 {
@@ -844,9 +875,23 @@ void FcitxUIMoveInputWindowReal(FcitxInstance *instance)
 FCITX_EXPORT_API
 void FcitxUISwitchToFallback(struct _FcitxInstance* instance)
 {
-    if (!instance->uifallback || instance->ui != instance->uinormal)
+    if (!instance->fallbackuiName || instance->ui != instance->uinormal)
         return;
 
+    if (!instance->uifallback) {
+        // load fallback ui
+        FcitxAddon* fallbackAddon = FcitxAddonsGetAddonByName(&instance->addons, instance->fallbackuiName);
+        if (!fallbackAddon || !fallbackAddon->bEnabled || !FcitxUILoadInternal(instance, fallbackAddon)) {
+            // reset fallbackuiName, never load it again and again
+            free(instance->fallbackuiName);
+            instance->fallbackuiName = NULL;
+            return;
+        }
+        instance->uifallback = fallbackAddon;
+        if (instance->uifallback->ui->Suspend)
+            instance->uifallback->ui->Suspend(instance->uifallback->addonInstance);
+    }
+
     if (instance->uinormal->ui->Suspend)
         instance->uinormal->ui->Suspend(instance->uinormal->addonInstance);
 
@@ -889,6 +934,9 @@ void FcitxMenuItemFree(void* arg)
     FcitxMenuItem* item = arg;
     if (item->tipstr)
         free(item->tipstr);
+    if (item->data) {
+        free(item->data);
+    }
 }
 
 FCITX_EXPORT_API
@@ -900,7 +948,7 @@ FcitxUIMenu* FcitxUIGetMenuByStatusName(FcitxInstance* instance, const char* nam
         if (!compstatus)
             return NULL;
     }
-    
+
     UT_array* uimenus = FcitxInstanceGetUIMenus(instance);
     FcitxUIMenu** menupp, *menup = NULL;
     for (menupp = (FcitxUIMenu **) utarray_front(uimenus);
@@ -912,7 +960,7 @@ FcitxUIMenu* FcitxUIGetMenuByStatusName(FcitxInstance* instance, const char* nam
             break;
         }
     }
-    
+
     return menup;
 }
 
