@@ -23,6 +23,40 @@
  * @{
  */
 
+/**
+ * @file ui.h
+ *
+ * this file provides the UI component for Fcitx, fcitx doesn't provides any platform specific
+ * UI element, all UI are implemented via Fcitx's addon.
+ *
+ * For input method and module, they can register UI component and get native UI for free from
+ * Fcitx.
+ *
+ * There is two abstract UI element in Fcitx, one is Status, and the other is Menu. Both of them
+ * require a unique name in Fcitx. Usually this name is meaningful in order to increase
+ * readability.
+ *
+ * For status, due to backward compatible problem, there are two kinds of status, simple and complex
+ * one. Status have three property, short description, long description and icon name. The actual
+ * UI depends on specific UI implementation, but usually, status are displayed as a button, with
+ * an icon. Icon name can be absolute png file path. All description need to be I18N'ed by addon
+ * themselves.
+ *
+ * For menu, each menu item contains a string, like status, it need to I18N'ed before set. Though
+ * there is different type in menu item, DONOT use submenu or divline, they are not supported by
+ * specific implementation, and suppose to be only used internally.
+ *
+ * For display string, there are five field can be used to display string:
+ *  - Client Preedit
+ *  - Preedit
+ *  - AuxUp
+ *  - AuxDown
+ *  - Candidate words
+ *
+ * In spite of candidate words, other four fields are FcitxMessages pointer. FcitxMessages is
+ * basically a fixed length string array. Each string inside FcitxMessages can have a type.
+ */
+
 #ifndef _FCITX_UI_H_
 #define _FCITX_UI_H_
 
@@ -37,30 +71,37 @@
 extern "C" {
 #endif
 
-#define MESSAGE_TYPE_COUNT 7
+#define MESSAGE_TYPE_COUNT 7 /**< message color type count */
 
 #define MESSAGE_MAX_CHARNUM (150)   /**< Maximum length per-message */
 
 #define MESSAGE_MAX_LENGTH  (MESSAGE_MAX_CHARNUM*UTF8_MAX_LENGTH)   /**< maximum byte per message */
 
-#define MAX_MESSAGE_COUNT 64
+#define MAX_MESSAGE_COUNT 64 /**< maximum message string number */
 
+    struct _FcitxAddon;
+
+    /** fcitx menu item */
     typedef struct _FcitxMenuItem FcitxMenuItem;
     struct _FcitxUIMenu;
 
+    /** fcitx menu state */
     typedef enum _FcitxMenuState {
         MENU_ACTIVE = 0,
         MENU_INACTIVE = 1
     } FcitxMenuState;
 
+    /** fcitx menu item type */
     typedef enum _FcitxMenuItemType {
         MENUTYPE_SIMPLE,
         MENUTYPE_SUBMENU,
         MENUTYPE_DIVLINE
     } FcitxMenuItemType;
-    
-    
+
+
+    /** menu action prototype */
     typedef boolean (*FcitxMenuActionFunction)(struct _FcitxUIMenu *arg, int index);
+    /** menu update prototype */
     typedef void    (*FcitxUpdateMenuFunction)(struct _FcitxUIMenu *arg);
 
     /**
@@ -83,8 +124,13 @@ extern "C" {
          * the submenu to this entry
          **/
         struct _FcitxUIMenu *subMenu;
-        
-        int padding[16]; /**< padding */
+
+        union {
+            void* data;
+            int dummy[2];
+        };
+
+        int padding[14]; /**< padding */
     };
 
     /**
@@ -129,8 +175,10 @@ extern "C" {
          * mark of this menu
          **/
         int mark;
-        
-        int padding[16]; /**< padding */
+
+        boolean visible; /**< menu is visible or not */
+
+        int padding[15]; /**< padding */
     };
 
     /**
@@ -169,10 +217,10 @@ extern "C" {
          * visible
          */
         boolean visible;
-        
+
         int padding[16]; /**< padding */
     };
-    
+
 
     /**
      * Fcitx Status icon to be displayed on the UI
@@ -210,37 +258,37 @@ extern "C" {
          * visible
          */
         boolean visible;
-        
+
         int padding[16]; /**< padding */
     };
 
-    typedef enum _FcitxUIFlag {
-        UI_NONE = 0,
-        UI_MOVE = (1 << 1),
-        UI_UPDATE = (1 << 2),
-    } FcitxUIFlag;
-
     struct _FcitxInstance;
+
+    /** message type and flags */
     typedef enum _FcitxMessageType {
         MSG_TYPE_FIRST = 0,
         MSG_TYPE_LAST = 6,
         MSG_TIPS = 0,           /**< Hint Text */
         MSG_INPUT = 1,          /**< User Input */
         MSG_INDEX = 2,          /**< Index Number */
-        MSG_FIRSTCAND = 3,      /**< First candidate */
+        MSG_CANDIATE_CURSOR = 3,/**< candidate cursor */
+        MSG_FIRSTCAND = MSG_CANDIATE_CURSOR,      /**< deprecated */
         MSG_USERPHR = 4,        /**< User Phrase */
         MSG_CODE = 5,           /**< Typed character */
         MSG_OTHER = 6,          /**< Other Text */
         MSG_NOUNDERLINE = (1 << 3), /**< backward compatible, no underline is a flag */
-        MSG_HIGHLIGHT = (1 << 4),
-        MSG_REGULAR_MASK = 0x7
+        MSG_HIGHLIGHT = (1 << 4), /**< highlight the preedit */
+        MSG_DONOT_COMMIT_WHEN_UNFOCUS = (1 << 5), /**< backward compatible */
+        MSG_REGULAR_MASK = 0x7 /**< regular color type mask */
     } FcitxMessageType;
 
+    /** an array of message for display */
     typedef struct _FcitxMessages FcitxMessages;
-    struct _FcitxAddon;
-
+    /** Fcitx Menu */
     typedef struct _FcitxUIMenu FcitxUIMenu;
+    /** Fcitx Status, supports active/inactive */
     typedef struct _FcitxUIStatus FcitxUIStatus;
+    /** Fcitx Complex Status, ability to use custom icon */
     typedef struct _FcitxUIComplexStatus FcitxUIComplexStatus;
 
     /**
@@ -314,7 +362,7 @@ extern "C" {
 
         void (*Destroy)(void*); /**< destroy user interface addon */
         void (*RegisterComplexStatus)(void*, FcitxUIComplexStatus*); /**< register complex status */
-        void (*UpdateComplexStatus)(void *arg, FcitxUIComplexStatus*);; /**< register complext status */
+        void (*UpdateComplexStatus)(void *arg, FcitxUIComplexStatus*); /**< register complext status */
         void (*padding3)(void*); /**< padding */
     } FcitxUI;
 
@@ -354,7 +402,7 @@ extern "C" {
      * @param  ...
      * @return void
      **/
-    void FcitxMessagesSetMessage(FcitxMessages* message, int position, FcitxMessageType type, const char* fmt, ...);
+    void FcitxMessagesSetMessage(FcitxMessages* message, int position, int type, const char* fmt, ...);
     /**
      * set only message string
      *
@@ -392,7 +440,7 @@ extern "C" {
      * @param ap arguments
      * @return void
      **/
-    void FcitxMessagesSetMessageV(FcitxMessages* message, int position, FcitxMessageType type, const char* fmt, va_list ap);
+    void FcitxMessagesSetMessageV(FcitxMessages* message, int position, int type, const char* fmt, va_list ap);
     /**
      * set message count
      *
@@ -424,14 +472,14 @@ extern "C" {
      * @return FcitxMessageType
      **/
     FcitxMessageType FcitxMessagesGetMessageType(FcitxMessages* m, int index);
-    
+
     /**
      * get message type at index, will not filter non regular type
      *
      * @param m message
      * @param index index
      * @return FcitxMessageType
-     * 
+     *
      * @see FcitxMessagesGetMessageType
      * @since 4.2.1
      **/
@@ -461,6 +509,16 @@ extern "C" {
      * @return void
      **/
     void FcitxMenuAddMenuItem(FcitxUIMenu* menu, const char* string, FcitxMenuItemType type, FcitxUIMenu* subMenu);
+    /**
+     * add a new menu shell
+     *
+     * @param menu menu
+     * @param string menu text
+     * @param type menu type
+     * @param subMenu submenu pointer
+     * @return void
+     **/
+    void FcitxMenuAddMenuItemWithData(FcitxUIMenu* menu, const char* string, FcitxMenuItemType type, FcitxUIMenu* subMenu, void* arg);
 
     /**
      * clear all menu shell
@@ -521,7 +579,7 @@ extern "C" {
      * @param shortDesc short description
      * @param longDesc long description
      * @param toggleStatus callback for toggle status
-     * @param getStatus get current status
+     * @param getIconName get current icon name
      * @return void
      **/
     void FcitxUIRegisterComplexStatus(struct _FcitxInstance* instance,
@@ -598,7 +656,7 @@ extern "C" {
      * @param instance fcitx instance
      * @param name status name
      * @return FcitxUIMenu*
-     * 
+     *
      * @since 4.2.1
      **/
     FcitxUIMenu* FcitxUIGetMenuByStatusName(struct _FcitxInstance* instance, const char* name);
@@ -611,8 +669,8 @@ extern "C" {
      * @return FcitxUIStatus*
      **/
     FcitxUIComplexStatus *FcitxUIGetComplexStatusByName(struct _FcitxInstance* instance, const char* name);
-    
-    
+
+
     /**
      * set visibility for a status icon
      *
@@ -622,7 +680,7 @@ extern "C" {
      * @return void
      **/
     void FcitxUISetStatusVisable(struct _FcitxInstance* instance, const char* name, boolean visible);
-    
+
     /**
      * @brief set string for a status icon
      *
@@ -631,7 +689,7 @@ extern "C" {
      * @param shortDesc short description
      * @param longDesc long description
      * @return void
-     * 
+     *
      * @since 4.2.1
      **/
     void FcitxUISetStatusString(struct _FcitxInstance* instance, const char* name, const char* shortDesc, const char* longDesc);
@@ -704,6 +762,15 @@ extern "C" {
 
     char* FcitxUICandidateWordToCString(struct _FcitxInstance* instance);
 
+
+    /**
+     * @brief commit current preedit string if any
+     *
+     * @param instance fcitx instance
+     * @return void
+     **/
+    void FcitxUICommitPreedit(struct _FcitxInstance* instance);
+
     /**
      * mark input window should update
      *
