@@ -58,7 +58,7 @@ typedef struct _WidePunc {
 typedef struct _FcitxPunc {
     char* langCode;
     WidePunc* curPunc;
-    
+
     UT_hash_handle hh;
 } FcitxPunc;
 
@@ -67,8 +67,10 @@ static FcitxPunc* LoadPuncFile(const char* filename);
 static char *GetPunc(struct _FcitxPuncState* puncState, int iKey);
 static void FreePunc(struct _FcitxPuncState* puncState);
 static void* PuncCreate(FcitxInstance* instance);
+static boolean PuncPreFilter(void* arg, FcitxKeySym sym, unsigned int state, INPUT_RETURN_VALUE* retVal);
 static boolean ProcessPunc(void* arg, FcitxKeySym sym, unsigned int state, INPUT_RETURN_VALUE* retVal);
 static void* PuncGetPunc(void* a, FcitxModuleFunctionArg arg);
+static void* PuncGetPunc2(void* a, FcitxModuleFunctionArg arg);
 static void TogglePuncState(void *arg);
 static boolean GetPuncState(void *arg);
 static void ReloadPunc(void *arg);
@@ -110,6 +112,9 @@ void* PuncCreate(FcitxInstance* instance)
 
     FcitxInstanceRegisterPostInputFilter(instance, hk);
 
+    hk.func = PuncPreFilter;
+    FcitxInstanceRegisterPreInputFilter(instance, hk);
+
     puncState->cLastIsAutoConvert = '\0';
     puncState->bLastIsNumber = false;
 
@@ -124,16 +129,20 @@ void* PuncCreate(FcitxInstance* instance)
     hook.func = ResetPunc;
 
     FcitxInstanceRegisterResetInputHook(instance, hook);
-    
+
     hook.func = ResetPuncWhichStatus;
 
     FcitxInstanceRegisterInputUnFocusHook(instance, hook);
-    
+
     FcitxInstanceWatchContext(instance, CONTEXT_IM_LANGUAGE, PuncLanguageChanged, puncState);
 
-    FcitxUIRegisterStatus(instance, puncState, "punc", _("Full Width Punctuation"), _("Full Width Punctuation"), TogglePuncState, GetPuncState);
+    FcitxProfile* profile = FcitxInstanceGetProfile(instance);
+    FcitxUIRegisterStatus(instance, puncState, "punc",
+                          profile->bUseWidePunc ? _("Full width punct") :  _("Latin punct"),
+                          _("Toggle Full Width Punctuation"), TogglePuncState, GetPuncState);
 
     AddFunction(puncaddon, PuncGetPunc);
+    AddFunction(puncaddon, PuncGetPunc2);
     return puncState;
 }
 
@@ -148,9 +157,9 @@ void PuncLanguageChanged(void* arg, const void* value)
             puncState->curPunc = punc->curPunc;
         else
             puncState->curPunc = NULL;
-    } else 
+    } else
         puncState->curPunc = NULL;
-    
+
     FcitxUISetStatusVisable (puncState->owner, "punc",  puncState->curPunc != NULL) ;
 }
 
@@ -161,6 +170,33 @@ void* PuncGetPunc(void* a, FcitxModuleFunctionArg arg)
     return GetPunc(puncState, *key);
 }
 
+
+void* PuncGetPunc2(void* a, FcitxModuleFunctionArg arg)
+{
+    FcitxPuncState* puncState = (FcitxPuncState*) a;
+    int *key = arg.args[0];
+    char** p1 = arg.args[1];
+    char** p2 = arg.args[2];
+    int             iIndex = 0;
+    WidePunc       *curPunc = puncState->curPunc;
+
+    if (!curPunc)
+        return NULL;
+
+    while (curPunc[iIndex].ASCII) {
+        if (curPunc[iIndex].ASCII == *key) {
+            if (p1)
+                *p1 = curPunc[iIndex].strWidePunc[0];
+            if (curPunc[iIndex].iCount > 1 && p2)
+                *p2 = curPunc[iIndex].strWidePunc[1];
+            break;
+        }
+        iIndex++;
+    }
+
+    return NULL;
+}
+
 void ResetPunc(void* arg)
 {
     FcitxPuncState* puncState = (FcitxPuncState*) arg;
@@ -172,18 +208,30 @@ void ResetPuncWhichStatus(void* arg)
 {
     FcitxPuncState* puncState = (FcitxPuncState*) arg;
     WidePunc       *curPunc = puncState->curPunc;
-    
+
     if (!curPunc)
         return;
-    
+
     int iIndex = 0;
-    
+
     while (curPunc[iIndex].ASCII) {
         curPunc[iIndex].iWhich = 0;
         iIndex++;
     }
 }
 
+boolean PuncPreFilter(void* arg, FcitxKeySym sym, unsigned int state, INPUT_RETURN_VALUE* retVal)
+{
+    if (!FcitxHotkeyIsHotKeySimple(sym, state))
+        return false;
+
+    FcitxPuncState* puncState = (FcitxPuncState*) arg;
+    if (!FcitxHotkeyIsHotKeyDigit(sym, state) && !IsHotKeyPunc(sym, state))
+        puncState->bLastIsNumber = false;
+
+    return false;
+}
+
 boolean ProcessPunc(void* arg, FcitxKeySym sym, unsigned int state, INPUT_RETURN_VALUE* retVal)
 {
     FcitxPuncState* puncState = (FcitxPuncState*) arg;
@@ -291,7 +339,7 @@ boolean LoadPuncDict(FcitxPuncState* puncState)
             HASH_ADD_KEYPTR(hh, puncState->puncSet, punc->langCode, strlen(punc->langCode), punc);
         curpuncfile = curpuncfile->hh.next;
     }
-    
+
     fcitx_utils_free_string_hash_set(puncfiles);
     return true;
 }
@@ -304,7 +352,7 @@ FcitxPunc* LoadPuncFile(const char* filename)
     char           *pstr;               // 临时指针
     int             i;
     fpDict = FcitxXDGGetFileWithPrefix("data", filename, "r", NULL);
-    
+
     if (strlen(filename) < strlen(PUNC_DICT_FILENAME))
         return NULL;
 
@@ -371,18 +419,18 @@ FcitxPunc* LoadPuncFile(const char* filename)
 
     punc[iRecordNo].ASCII = '\0';
     fclose(fpDict);
-    
+
     FcitxPunc* p = fcitx_utils_malloc0(sizeof(FcitxPunc));
     p->langCode = "";
-    
+
     const char* langcode = filename + strlen(PUNC_DICT_FILENAME);
     if (*langcode == '\0')
         p->langCode = strdup("C");
     else
         p->langCode = strdup(langcode + 1);
-        
+
     p->curPunc = punc;
-    
+
     return p;
 }
 
@@ -432,8 +480,11 @@ void TogglePuncState(void* arg)
     FcitxInstance* instance = puncState->owner;
     FcitxProfile* profile = FcitxInstanceGetProfile(instance);
     profile->bUseWidePunc = !profile->bUseWidePunc;
+
+    FcitxUISetStatusString(puncState->owner, "punc",
+                           profile->bUseWidePunc ? _("Full width punct") :  _("Latin punct"),
+                          _("Toggle Full Width Punctuation"));
     FcitxProfileSave(profile);
-    FcitxInstanceResetInput(puncState->owner);
 }
 
 INPUT_RETURN_VALUE TogglePuncStateWithHotkey(void* arg)
@@ -462,7 +513,7 @@ void ReloadPunc(void* arg)
     FcitxPuncState* puncState = (FcitxPuncState*) arg;
     FreePunc(puncState);
     LoadPuncDict(puncState);
-    
+
     PuncLanguageChanged(puncState, FcitxInstanceGetContextString(puncState->owner, CONTEXT_IM_LANGUAGE));
 }
 
