@@ -38,6 +38,7 @@
 #include "fcitx/candidate.h"
 #include "fcitx/context.h"
 #include "module/punc/punc.h"
+#include "module/lua/luamod.h"
 
 
 typedef enum _QuickPhraseTriggerKey
@@ -55,6 +56,7 @@ typedef struct _QuickPhraseState {
     int iLastQuickPhrase;
     boolean enabled;
     FcitxInstance* owner;
+    char buffer[MAX_USER_INPUT * UTF8_MAX_LENGTH + 1];
 } QuickPhraseState;
 
 typedef struct _QuickPhraseCand {
@@ -70,6 +72,7 @@ static INPUT_RETURN_VALUE QuickPhraseGetCandWords(QuickPhraseState* qpstate);
 static UT_icd qp_icd = {sizeof(QUICK_PHRASE), NULL, NULL, NULL};
 static void ShowQuickPhraseMessage(QuickPhraseState *qpstate);
 static INPUT_RETURN_VALUE QuickPhraseGetCandWord(void* arg, FcitxCandidateWord* candWord);
+static INPUT_RETURN_VALUE QuickPhraseGetLuaCandWord(void* arg, FcitxCandidateWord* candWord);
 static boolean LoadQuickPhraseConfig(QuickPhraseState* qpstate);
 static void SaveQuickPhraseConfig(QuickPhraseState* qpstate);
 static boolean QuickPhrasePostFilter(void* arg, FcitxKeySym sym,
@@ -131,12 +134,12 @@ void * QuickPhraseCreate(FcitxInstance *instance)
     qpstate->iFirstQuickPhrase = -1;
     qpstate->owner = instance;
     qpstate->enabled = false;
-    
+
     if (!LoadQuickPhraseConfig(qpstate)) {
         free(qpstate);
         return NULL;
     }
-    
+
     LoadQuickPhrase(qpstate);
 
     FcitxKeyFilterHook hk;
@@ -151,7 +154,7 @@ void * QuickPhraseCreate(FcitxInstance *instance)
     resethk.arg = qpstate;
     resethk.func = QuickPhraseReset;
     FcitxInstanceRegisterResetInputHook(instance, resethk);
-    
+
     FcitxInstanceRegisterWatchableContext(instance, CONTEXT_DISABLE_QUICKPHRASE, FCT_Boolean, FCF_ResetOnInputMethodChange);
 
     return qpstate;
@@ -193,7 +196,7 @@ void LoadQuickPhrase(QuickPhraseState * qpstate)
         while (*p && !isspace(*p))
             p ++;
 
-        if (p == '\0')
+        if (*p == '\0')
             continue;
 
         while (isspace(*p)) {
@@ -204,7 +207,7 @@ void LoadQuickPhrase(QuickPhraseState * qpstate)
         if (strlen(buf1) >= QUICKPHRASE_CODE_LEN)
             continue;
 
-        if (strlen(p) >= QUICKPHRASE_PHRASE_LEN * UTF8_MAX_LENGTH + 1)
+        if (strlen(p) >= QUICKPHRASE_PHRASE_LEN * UTF8_MAX_LENGTH)
             continue;
 
         strcpy(tempQuickPhrase.strCode, buf1);
@@ -213,12 +216,6 @@ void LoadQuickPhrase(QuickPhraseState * qpstate)
         utarray_push_back(qpstate->quickPhrases, &tempQuickPhrase);
     }
 
-    strcpy(tempQuickPhrase.strCode, "\x47\x4e\x4f\x4d\x45");
-    strcpy(tempQuickPhrase.strPhrase, "\x48\x6f\x77\x20\x63\x61\x6e\x20\x69\x74\x20\x66\x75\x63\x6b\x20\x66"
-           "\x63\x69\x74\x78\x20\x73\x6f\x20\x6d\x61\x6e\x79\x20\x74\x69\x6d\x65\x73"
-           "\x3f\x20\x49\x74\x20\x73\x75\x63\x6b\x73\x2e");
-    utarray_push_back(qpstate->quickPhrases, &tempQuickPhrase);
-
     if (buf)
         free(buf);
     if (buf1)
@@ -243,11 +240,12 @@ void FreeQuickPhrase(void *arg)
 void ShowQuickPhraseMessage(QuickPhraseState *qpstate)
 {
     FcitxInputState *input = FcitxInstanceGetInputState(qpstate->owner);
-    FcitxInputStateSetCursorPos(input, strlen(FcitxInputStateGetRawInputBuffer(input)));
+    FcitxInputStateSetCursorPos(input, strlen(qpstate->buffer));
+    FcitxInputStateSetClientCursorPos(input, strlen(qpstate->buffer));
     FcitxInstanceCleanInputWindowUp(qpstate->owner);
     FcitxMessagesAddMessageAtLast(FcitxInputStateGetAuxUp(input), MSG_TIPS, "%s", _("Quick Phrase: "));
-    FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
-    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
+    FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", qpstate->buffer);
+    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", qpstate->buffer);
 }
 
 boolean QuickPhrasePreFilter(void* arg, FcitxKeySym sym,
@@ -262,7 +260,7 @@ boolean QuickPhrasePreFilter(void* arg, FcitxKeySym sym,
         if (FcitxHotkeyIsHotKeySimple(keymain, state)) {
             *retval = QuickPhraseDoInput(qpstate, keymain, state);
             if (*retval == IRV_TO_PROCESS) {
-                if (strlen(FcitxInputStateGetRawInputBuffer(input)) == 0
+                if (strlen(qpstate->buffer) == 0
                     && (FcitxHotkeyIsHotKey(keymain, state, QuickPhraseTriggerKeys[qpstate->triggerKey]) || FcitxHotkeyIsHotKey(keymain, state, FCITX_SPACE))) {
                     char c[2] = { (char) (QuickPhraseTriggerKeys[qpstate->triggerKey][0].sym & 0xff), '\0'};
                     FcitxModuleFunctionArg farg;
@@ -275,17 +273,17 @@ boolean QuickPhrasePreFilter(void* arg, FcitxKeySym sym,
                     char buf[2];
                     buf[0] = keymain;
                     buf[1] = '\0';
-                    if (strlen(FcitxInputStateGetRawInputBuffer(input)) < MAX_USER_INPUT)
-                        strcat(FcitxInputStateGetRawInputBuffer(input), buf);
+                    if (strlen(qpstate->buffer) < MAX_USER_INPUT)
+                        strcat(qpstate->buffer, buf);
                     ShowQuickPhraseMessage(qpstate);
                     *retval = QuickPhraseGetCandWords(qpstate);
                 }
             } else
                 return true;
         } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)) {
-            size_t len = strlen(FcitxInputStateGetRawInputBuffer(input));
+            size_t len = strlen(qpstate->buffer);
             if (len > 0)
-                FcitxInputStateGetRawInputBuffer(input)[--len] = '\0';
+                qpstate->buffer[--len] = '\0';
             if (len == 0) {
                 *retval = IRV_CLEAN;
             } else {
@@ -294,8 +292,8 @@ boolean QuickPhrasePreFilter(void* arg, FcitxKeySym sym,
             }
         } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_ENTER)) {
 
-            if (strlen(FcitxInputStateGetRawInputBuffer(input)) > 0) {
-                strcpy(FcitxInputStateGetOutputString(input), FcitxInputStateGetRawInputBuffer(input));
+            if (strlen(qpstate->buffer) > 0) {
+                strcpy(FcitxInputStateGetOutputString(input), qpstate->buffer);
                 QuickPhraseReset(qpstate);
                 *retval = IRV_COMMIT_STRING;
             } else {
@@ -328,15 +326,15 @@ boolean QuickPhrasePostFilter(void* arg, FcitxKeySym sym,
     if (*retval != IRV_TO_PROCESS)
         return false;
 
-    if (!disableQuickPhrase 
+    if (!disableQuickPhrase
         && !qpstate->enabled
-        && FcitxInputStateGetRawInputBufferSize(input) == 0
+        && qpstate->buffer[0] == '\0'
         && FcitxHotkeyIsHotKey(sym, state, QuickPhraseTriggerKeys[qpstate->triggerKey])) {
         FcitxInstanceCleanInputWindow(qpstate->owner);
         FcitxInputStateSetShowCursor(input, true);
         FcitxMessagesAddMessageAtLast(FcitxInputStateGetAuxUp(input), MSG_TIPS, "%s", _("Quick Phrase: "));
         FcitxInputStateSetCursorPos(input, 0);
-    
+
         char c[2] = { (char) (QuickPhraseTriggerKeys[qpstate->triggerKey][0].sym & 0xff), '\0'};
         FcitxModuleFunctionArg farg;
         FcitxKeySym s = QuickPhraseTriggerKeys[qpstate->triggerKey][0].sym;
@@ -357,6 +355,7 @@ void QuickPhraseReset(void* arg)
 {
     QuickPhraseState *qpstate = (QuickPhraseState*) arg;
     qpstate->enabled = false;
+    qpstate->buffer[0] = '\0';
 }
 INPUT_RETURN_VALUE QuickPhraseDoInput(void* arg, FcitxKeySym sym, int state)
 {
@@ -364,11 +363,11 @@ INPUT_RETURN_VALUE QuickPhraseDoInput(void* arg, FcitxKeySym sym, int state)
     FcitxInputState *input = FcitxInstanceGetInputState(qpstate->owner);
     FcitxGlobalConfig* fc = FcitxInstanceGetGlobalConfig(qpstate->owner);
     int retVal = IRV_TO_PROCESS;
-    
+
     const FcitxHotkey* hkPrevPage = FcitxInstanceGetContextHotkey(qpstate->owner, CONTEXT_ALTERNATIVE_PREVPAGE_KEY);
     if (hkPrevPage == NULL)
         hkPrevPage = fc->hkPrevPage;
-    
+
     const FcitxHotkey* hkNextPage = FcitxInstanceGetContextHotkey(qpstate->owner, CONTEXT_ALTERNATIVE_NEXTPAGE_KEY);
     if (hkNextPage == NULL)
         hkNextPage = fc->hkNextPage;
@@ -398,7 +397,6 @@ INPUT_RETURN_VALUE QuickPhraseGetCandWords(QuickPhraseState* qpstate)
     FcitxInputState *input = FcitxInstanceGetInputState(qpstate->owner);
     FcitxInstance *instance = qpstate->owner;
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(instance);
-    
     FcitxInstanceCleanInputWindowDown(qpstate->owner);
 
     FcitxCandidateWordSetPageSize(FcitxInputStateGetCandidateList(input), config->iMaxCandWord);
@@ -406,16 +404,23 @@ INPUT_RETURN_VALUE QuickPhraseGetCandWords(QuickPhraseState* qpstate)
 
     pKey = &searchKey;
 
+    {
+        FcitxModuleFunctionArg arg;
+        char *text = qpstate->buffer;
+        arg.args[0] = text;
+        arg.args[1] = QuickPhraseGetLuaCandWord;
+        arg.args[2] = qpstate;
+        InvokeFunction(qpstate->owner, FCITX_LUA, CALLCOMMAND, arg);
+    }
+
     if (!qpstate->quickPhrases)
         return IRV_DISPLAY_MESSAGE;
 
-    iInputLen = strlen(FcitxInputStateGetRawInputBuffer(input));
+    iInputLen = strlen(qpstate->buffer);
     if (iInputLen > QUICKPHRASE_CODE_LEN)
         return IRV_DISPLAY_MESSAGE;
 
-    strcpy(searchKey.strCode, FcitxInputStateGetRawInputBuffer(input));
-
-    FcitxCandidateWordReset(FcitxInputStateGetCandidateList(input));
+    strcpy(searchKey.strCode, qpstate->buffer);
 
     currentQuickPhrase = utarray_custom_bsearch(pKey, qpstate->quickPhrases, false, PhraseCmp);
     qpstate->iFirstQuickPhrase = utarray_eltidx(qpstate->quickPhrases, currentQuickPhrase);
@@ -423,8 +428,7 @@ INPUT_RETURN_VALUE QuickPhraseGetCandWords(QuickPhraseState* qpstate)
     qpstate->iLastQuickPhrase = utarray_eltidx(qpstate->quickPhrases, lastQuickPhrase);
     if (qpstate->iLastQuickPhrase < 0)
         qpstate->iLastQuickPhrase = utarray_len(qpstate->quickPhrases);
-    if (!currentQuickPhrase || strncmp(FcitxInputStateGetRawInputBuffer(input), currentQuickPhrase->strCode, iInputLen)) {
-        FcitxInstanceCleanInputWindowDown(instance);
+    if (!currentQuickPhrase || strncmp(qpstate->buffer, currentQuickPhrase->strCode, iInputLen)) {
         currentQuickPhrase = NULL;
         return IRV_DISPLAY_MESSAGE;
     }
@@ -433,7 +437,7 @@ INPUT_RETURN_VALUE QuickPhraseGetCandWords(QuickPhraseState* qpstate)
                               qpstate->iFirstQuickPhrase);
             currentQuickPhrase != NULL;
             currentQuickPhrase = (QUICK_PHRASE*) utarray_next(qpstate->quickPhrases, currentQuickPhrase)) {
-        if (!strncmp(FcitxInputStateGetRawInputBuffer(input), currentQuickPhrase->strCode, iInputLen)) {
+        if (!strncmp(qpstate->buffer, currentQuickPhrase->strCode, iInputLen)) {
             QuickPhraseCand* qpcand = fcitx_utils_malloc0(sizeof(QuickPhraseCand));
             qpcand->cand = currentQuickPhrase;
             FcitxCandidateWord candWord;
@@ -459,6 +463,21 @@ void ReloadQuickPhrase(void* arg)
     LoadQuickPhrase(qpstate);
 }
 
+INPUT_RETURN_VALUE QuickPhraseGetLuaCandWord(void* arg, FcitxCandidateWord* candWord)
+{
+    QuickPhraseState *qpstate = (QuickPhraseState*) arg;
+    FcitxInputState *input = FcitxInstanceGetInputState(qpstate->owner);
+    if (candWord->strExtra) {
+        strcat(qpstate->buffer, candWord->strExtra);
+        ShowQuickPhraseMessage(qpstate);
+        return QuickPhraseGetCandWords(qpstate);
+    }
+    else {
+        strcpy(FcitxInputStateGetOutputString(input), candWord->strWord);
+        return IRV_COMMIT_STRING;
+    }
+}
+
 INPUT_RETURN_VALUE QuickPhraseGetCandWord(void* arg, FcitxCandidateWord* candWord)
 {
     QuickPhraseState *qpstate = (QuickPhraseState*) arg;
