@@ -47,6 +47,20 @@
 #define VK_NUMBERS      47
 #define VK_MAX          50
 
+const char* dummyTranslate[] = {
+    N_("Latin"),
+    N_("Fullwidth"),
+    N_("Greek"),
+    N_("Russian"),
+    N_("Index"),
+    N_("Math"),
+    N_("Number"),
+    N_("Special"),
+    N_("Hiragana"),
+    N_("Katakana"),
+    N_("Table Symbol")
+};
+
 struct _FcitxVKState;
 
 typedef struct _VKS {
@@ -56,13 +70,11 @@ typedef struct _VKS {
 
 typedef struct _VKWindow {
     Window          window;
-    FcitxConfigColor* fontColor;
     int fontSize;
     cairo_surface_t* surface;
     cairo_surface_t* keyboard;
     Display*        dpy;
     struct _FcitxVKState* owner;
-    char **font;
     char *defaultFont;
     int iVKWindowX;
     int iVKWindowY;
@@ -78,7 +90,6 @@ typedef struct _FcitxVKState {
     boolean         bVK;
     FcitxUIMenu     vkmenu;
     FcitxInstance* owner;
-    FcitxAddon* classicui;
 } FcitxVKState;
 
 const char            vkTable[VK_NUMBERS + 1] = "`1234567890-=qwertyuiop[]\\asdfghjkl;'zxcvbnm,./";
@@ -137,7 +148,6 @@ void *VKCreate(FcitxInstance* instance)
     FcitxVKState *vkstate = fcitx_utils_malloc0(sizeof(FcitxVKState));
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(instance);
     vkstate->owner = instance;
-    vkstate->classicui = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(instance), FCITX_CLASSIC_UI_NAME);
 
     FcitxHotkeyHook hotkey;
     hotkey.hotkey = config->hkVK;
@@ -145,7 +155,7 @@ void *VKCreate(FcitxInstance* instance)
     hotkey.arg = vkstate;
     FcitxInstanceRegisterHotkeyFilter(instance, hotkey);
 
-    FcitxUIRegisterStatus(instance, vkstate, "vk", _("Virtual Keyboard"), _("Virtual Keyboard State"),  ToggleVKState, GetVKState);
+    FcitxUIRegisterStatus(instance, vkstate, "vk", _("Toggle Virtual Keyboard"), _("Virtual Keyboard State"),  ToggleVKState, GetVKState);
 
     LoadVKMapFile(vkstate);
 
@@ -173,10 +183,6 @@ void *VKCreate(FcitxInstance* instance)
     vkstate->vkmenu.priv = vkstate;
     vkstate->vkmenu.isSubMenu = false;
 
-    int i;
-    for (i = 0; i < vkstate->iVKCount; i ++)
-        FcitxMenuAddMenuItem(&vkstate->vkmenu, vkstate->vks[i].strName, MENUTYPE_SIMPLE, NULL);
-
     FcitxUIRegisterMenu(instance, &vkstate->vkmenu);
 
     return vkstate;
@@ -185,13 +191,26 @@ void *VKCreate(FcitxInstance* instance)
 boolean VKMenuAction(FcitxUIMenu *menu, int index)
 {
     FcitxVKState* vkstate = (FcitxVKState*) menu->priv;
-    SelectVK(vkstate, index);
+    if (index < vkstate->iVKCount)
+        SelectVK(vkstate, index);
+    else {
+        if (vkstate->bVK) {
+            FcitxUIUpdateStatus(vkstate->owner, "vk");
+        }
+    }
     return true;
 }
 
 void UpdateVKMenu(FcitxUIMenu *menu)
 {
     FcitxVKState* vkstate = (FcitxVKState*) menu->priv;
+    FcitxMenuClear(menu);
+    int i;
+    for (i = 0; i < vkstate->iVKCount; i ++)
+        FcitxMenuAddMenuItem(&vkstate->vkmenu, vkstate->vks[i].strName, MENUTYPE_SIMPLE, NULL);
+    if (vkstate->bVK) {
+        FcitxMenuAddMenuItem(&vkstate->vkmenu, _("Close virtual keyboard"), MENUTYPE_SIMPLE, NULL);
+    }
     menu->mark = vkstate->iCurrentVK;
 }
 
@@ -268,18 +287,10 @@ VKWindow* CreateVKWindow(FcitxVKState* vkstate)
     vkWindow->dpy = InvokeFunction(vkstate->owner, FCITX_X11, GETDISPLAY, arg);
 
     vkWindow->fontSize = 12;
-    if (vkstate->classicui) {
-        FcitxModuleFunctionArg arg;
-        vkWindow->fontColor = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, GETKEYBOARDFONTCOLOR, arg);
-        vkWindow->font = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, GETFONT, arg);
-    } else {
-        vkWindow->fontColor = &blackColor;
-        vkWindow->defaultFont = strdup("sans");
+    vkWindow->defaultFont = strdup("sans");
 #ifndef _ENABLE_PANGO
-        GetValidFont("zh", &vkWindow->defaultFont);
+    GetValidFont("zh", &vkWindow->defaultFont);
 #endif
-        vkWindow->font = &vkWindow->defaultFont;
-    }
 
     vkWindow->window = XCreateWindow(vkWindow->dpy,
                                      DefaultRootWindow(vkWindow->dpy),
@@ -334,22 +345,22 @@ boolean VKWindowEventHandler(void* arg, XEvent* event)
 cairo_surface_t* LoadVKImage(VKWindow* vkWindow)
 {
     FcitxVKState* vkstate = vkWindow->owner;
-    if (vkstate->classicui) {
-        FcitxModuleFunctionArg arg;
-        boolean fallback = true;
-        char vkimage[] = "keyboard.png";
-        arg.args[0] = vkimage;
-        arg.args[1] = &fallback;
-        return InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, LOADIMAGE, arg);
-    } else {
-        if (!vkWindow->keyboard) {
-            char* path = fcitx_utils_get_fcitx_path_with_filename("pkgdatadir", "skin/default/keyboard.png");
-            vkWindow->keyboard = cairo_image_surface_create_from_png(path);
-            free(path);
-        }
-        return vkWindow->keyboard;
+    FcitxModuleFunctionArg arg;
+    boolean fallback = true;
+    char vkimage[] = "keyboard.png";
+    arg.args[0] = vkimage;
+    arg.args[1] = &fallback;
+    cairo_surface_t* image = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, LOADIMAGE, arg);
+
+    if (image)
+        return image;
+
+    if (!vkWindow->keyboard) {
+        char* path = fcitx_utils_get_fcitx_path_with_filename("pkgdatadir", "skin/default/keyboard.png");
+        vkWindow->keyboard = cairo_image_surface_create_from_png(path);
+        free(path);
     }
-    return NULL;
+    return vkWindow->keyboard;
 }
 
 void DisplayVKWindow(VKWindow* vkWindow)
@@ -371,6 +382,14 @@ void DrawVKWindow(VKWindow* vkWindow)
     FcitxVKState *vkstate = vkWindow->owner;
     VKS *vks = vkstate->vks;
 
+    FcitxModuleFunctionArg arg;
+    FcitxConfigColor* fontColor = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, GETKEYBOARDFONTCOLOR, arg);
+    char** font = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, GETFONT, arg);
+
+    if (!fontColor || !font) {
+        fontColor = &blackColor;
+        font = &vkWindow->defaultFont;
+    }
 
     cr = cairo_create(vkWindow->surface);
     cairo_surface_t* vkimage = LoadVKImage(vkWindow);
@@ -378,35 +397,35 @@ void DrawVKWindow(VKWindow* vkWindow)
     cairo_paint(cr);
     /* 显示字符 */
     /* 名称 */
-    OutputString(cr, vks[vkstate->iCurrentVK].strName, *vkWindow->font, vkWindow->fontSize , (VK_WINDOW_WIDTH - StringWidth(vks[vkstate->iCurrentVK].strName, *vkWindow->font, vkWindow->fontSize)) / 2, 6, vkWindow->fontColor);
+    OutputString(cr, vks[vkstate->iCurrentVK].strName, *font, vkWindow->fontSize , (VK_WINDOW_WIDTH - StringWidth(vks[vkstate->iCurrentVK].strName, *font, vkWindow->fontSize)) / 2, 6, fontColor);
 
     /* 第一排 */
     iPos = 13;
     for (i = 0; i < 13; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *vkWindow->font, vkWindow->fontSize, iPos, 27, vkWindow->fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *vkWindow->font, vkWindow->fontSize, iPos - 5, 40, vkWindow->fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 27, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 40, fontColor);
         iPos += 24;
     }
     /* 第二排 */
     iPos = 48;
     for (i = 13; i < 26; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *vkWindow->font, vkWindow->fontSize, iPos, 55, vkWindow->fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *vkWindow->font, vkWindow->fontSize, iPos - 5, 68, vkWindow->fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 55, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 68, fontColor);
         iPos += 24;
     }
     /* 第三排 */
     iPos = 55;
     for (i = 26; i < 37; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *vkWindow->font, vkWindow->fontSize, iPos, 83, vkWindow->fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *vkWindow->font, vkWindow->fontSize, iPos - 5, 96, vkWindow->fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 83, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 96, fontColor);
         iPos += 24;
     }
 
     /* 第四排 */
     iPos = 72;
     for (i = 37; i < 47; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *vkWindow->font, vkWindow->fontSize, iPos, 111, vkWindow->fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *vkWindow->font, vkWindow->fontSize, iPos - 5, 124, vkWindow->fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 111, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 124, fontColor);
         iPos += 24;
     }
 
@@ -519,7 +538,6 @@ boolean VKMouseKey(FcitxVKState* vkstate, int x, int y)
 
         if (pstr) {
             FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), pstr);
-            FcitxInstanceIncreateInputCharacterCount(instance, fcitx_utf8_strlen(pstr));
         }
     }
 
@@ -572,7 +590,7 @@ void LoadVKMapFile(FcitxVKState *vkstate)
         if (!strcmp(pstr, "[VK]"))
             vkstate->iVKCount++;
         else if (!strncmp(pstr, "NAME=", 5))
-            vks[vkstate->iVKCount - 1].strName = strdup(pstr + 5);
+            vks[vkstate->iVKCount - 1].strName = strdup(gettext(pstr + 5));
         else {
             if (pstr[1] != '=' && !vkstate->iVKCount)
                 continue;
@@ -710,7 +728,7 @@ void SwitchVK(FcitxVKState *vkstate)
 
         if (!FcitxUISupportMainWindow(instance)) {
             x = dwidth / 2 - VK_WINDOW_WIDTH / 2;
-            y = 0;
+            y = 40;
         } else {
             int mx = 0, my = 0, mw = 0, mh = 0;
             FcitxUIGetMainWindowSize(instance, &mx, &my, &mw, &mh);
