@@ -43,16 +43,14 @@
 #include "module/x11/x11stuff.h"
 #include "classicui.h"
 #include "skin.h"
-#include "AboutWindow.h"
 #include "MenuWindow.h"
-#include <fcitx-utils/utils.h>
+#include "fcitx-utils/utils.h"
 
 #define FCITX_MAX(a,b) ((a) > (b)?(a) : (b))
 
 #define MAIN_BAR_MAX_WIDTH 2
 #define MAIN_BAR_MAX_HEIGHT 2
 
-static void ResizeSurface(cairo_surface_t** surface, int w, int h);
 static boolean MainWindowEventHandler(void *arg, XEvent* event);
 static void UpdateStatusGeometry(FcitxClassicUIStatus *privstat, SkinImage *image, int x, int y);
 static void ReloadMainWindow(void* arg, boolean enabled);
@@ -135,9 +133,7 @@ MainWindow* CreateMainWindow(FcitxClassicUI* classicui)
 void DisplayMainWindow(Display* dpy, MainWindow* mainWindow)
 {
     FcitxLog(DEBUG, _("DISPLAY MainWindow"));
-
-    if (!mainWindow->bMainWindowHidden)
-        XMapRaised(dpy, mainWindow->window);
+    XMapRaised(dpy, mainWindow->window);
 }
 
 void DrawMainWindow(MainWindow* mainWindow)
@@ -146,12 +142,39 @@ void DrawMainWindow(MainWindow* mainWindow)
     FcitxSkin *sc = &mainWindow->owner->skin;
     FcitxInstance *instance = mainWindow->owner->owner;
 
-    if (mainWindow->bMainWindowHidden)
-        return;
-
     FcitxLog(DEBUG, _("DRAW MainWindow"));
 
-    if (mainWindow->owner->hideMainWindow == HM_SHOW || (mainWindow->owner->hideMainWindow == HM_AUTO && (FcitxInstanceGetCurrentState(mainWindow->owner->owner) != IS_CLOSED))) {
+    FcitxUIStatus* status;
+    UT_array* uistats = FcitxInstanceGetUIStats(instance);
+    for (status = (FcitxUIStatus*) utarray_front(uistats);
+            status != NULL;
+            status = (FcitxUIStatus*) utarray_next(uistats, status)
+        ) {
+        FcitxClassicUIStatus* privstat =  GetPrivateStatus(status);
+        if (privstat == NULL)
+            continue;
+        /* reset status */
+        privstat->x = privstat->y = -1;
+        privstat->w = privstat->h = 0;
+        privstat->avail = false;
+    }
+
+    FcitxUIComplexStatus* compstatus;
+    UT_array* uicompstats = FcitxInstanceGetUIComplexStats(instance);
+    for (compstatus = (FcitxUIComplexStatus*) utarray_front(uicompstats);
+            compstatus != NULL;
+            compstatus = (FcitxUIComplexStatus*) utarray_next(uicompstats, compstatus)
+        ) {
+        FcitxClassicUIStatus* privstat =  GetPrivateStatus(compstatus);
+        if (privstat == NULL)
+            continue;
+        /* reset status */
+        privstat->x = privstat->y = -1;
+        privstat->w = privstat->h = 0;
+        privstat->avail = false;
+    }
+
+    if (mainWindow->owner->hideMainWindow == HM_SHOW || (mainWindow->owner->hideMainWindow == HM_AUTO && (FcitxInstanceGetCurrentState(mainWindow->owner->owner) == IS_ACTIVE))) {
         SkinImage* activeIcon = LoadImage(sc, sc->skinMainBar.active, false);
         cairo_t *c;
 
@@ -176,34 +199,7 @@ void DrawMainWindow(MainWindow* mainWindow)
             EnlargeCairoSurface(&mainWindow->cs_main_bar, width, height);
             XResizeWindow(mainWindow->dpy, mainWindow->window, width, height);
             DrawResizableBackground(c, back->image, height, width, 0, 0, 0, 0, F_RESIZE, F_COPY);
-            
-            FcitxUIStatus* status;
-            UT_array* uistats = FcitxInstanceGetUIStats(instance);
-            for (status = (FcitxUIStatus*) utarray_front(uistats);
-                    status != NULL;
-                    status = (FcitxUIStatus*) utarray_next(uistats, status)
-                ) {
-                FcitxClassicUIStatus* privstat =  GetPrivateStatus(status);
-                if (privstat == NULL)
-                    continue;
-                /* reset status */
-                privstat->x = privstat->y = -1;
-                privstat->w = privstat->h = 0;
-            }
-            
-            FcitxUIComplexStatus* compstatus;
-            UT_array* uicompstats = FcitxInstanceGetUIComplexStats(instance);
-            for (compstatus = (FcitxUIComplexStatus*) utarray_front(uicompstats);
-                 compstatus != NULL;
-                 compstatus = (FcitxUIComplexStatus*) utarray_next(uicompstats, compstatus)
-                ) {
-                FcitxClassicUIStatus* privstat =  GetPrivateStatus(compstatus);
-                if (privstat == NULL)
-                    continue;
-                /* reset status */
-                privstat->x = privstat->y = -1;
-                privstat->w = privstat->h = 0;
-            }
+
 
             SkinPlacement* sp;
             for (sp = (SkinPlacement*) utarray_front(&sc->skinMainBar.skinPlacement);
@@ -220,24 +216,7 @@ void DrawMainWindow(MainWindow* mainWindow)
                     if (FcitxInstanceGetCurrentStatev2(instance) != IS_ACTIVE || FcitxInstanceGetCurrentIM(instance) == NULL)
                         imicon = LoadImage(sc, sc->skinMainBar.eng, false);
                     else {
-                        FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
-                        char* path;
-                        if (im->strIconName[0] == '/')
-                            path = strdup(im->strIconName);
-                        else
-                            asprintf(&path, "%s.png", im->strIconName);
-                        imicon = LoadImage(sc, path, false);
-                        if (imicon == NULL)
-                            imicon = LoadImage(sc, sc->skinMainBar.active, false);
-                        else {
-                            SkinImage* activeIcon = LoadImage(sc, sc->skinMainBar.active, false);
-                            if (activeIcon) {
-                                ResizeSurface(&imicon->image,
-                                              cairo_image_surface_get_width(activeIcon->image),
-                                              cairo_image_surface_get_height(activeIcon->image));
-                            }
-                        }
-                        free(path);
+                        imicon = GetIMIcon(instance, sc, sc->skinMainBar.active, 3, false);
                     }
 
                     if (imicon) {
@@ -261,10 +240,11 @@ void DrawMainWindow(MainWindow* mainWindow)
                         free(path);
                         if (statusicon == NULL)
                             continue;
+                        privstat->avail = true;
                         DrawImage(c, statusicon->image, sp->x, sp->y, privstat->mouse);
                         UpdateStatusGeometry(privstat, statusicon, sp->x, sp->y);
                     }
-                    
+
                     compstatus = FcitxUIGetComplexStatusByName(instance, sp->name);
                     if (compstatus && compstatus->visible) {
                         FcitxClassicUIStatus* privstat = GetPrivateStatus(compstatus);
@@ -281,6 +261,7 @@ void DrawMainWindow(MainWindow* mainWindow)
                         free(path);
                         if (statusicon == NULL)
                             continue;
+                        privstat->avail = true;
                         DrawImage(c, statusicon->image, sp->x, sp->y, privstat->mouse);
                         UpdateStatusGeometry(privstat, statusicon, sp->x, sp->y);
                     }
@@ -304,23 +285,7 @@ void DrawMainWindow(MainWindow* mainWindow)
             if (FcitxInstanceGetCurrentStatev2(instance) != IS_ACTIVE || FcitxInstanceGetCurrentIM(instance) == NULL)
                 imicon = LoadImage(sc, sc->skinMainBar.eng, false);
             else {
-                FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
-                char *path;
-                if (im->strIconName[0] == '/')
-                    path = strdup(im->strIconName);
-                else
-                    asprintf(&path, "%s.png", im->strIconName);
-                imicon = LoadImage(sc, path, false);
-                if (imicon == NULL)
-                    imicon = LoadImage(sc, sc->skinMainBar.active, false);
-                else {
-                    if (activeIcon) {
-                        ResizeSurface(&imicon->image,
-                                      cairo_image_surface_get_width(activeIcon->image),
-                                      cairo_image_surface_get_height(activeIcon->image));
-                    }
-                }
-                free(path);
+                imicon = GetIMIcon(instance, sc, sc->skinMainBar.active, 3, false);
             }
 
             if (imicon) {
@@ -330,7 +295,7 @@ void DrawMainWindow(MainWindow* mainWindow)
                     height = imageheight;
             }
 
-            
+
             FcitxUIComplexStatus* compstatus;
             UT_array* uicompstats = FcitxInstanceGetUIComplexStats(instance);
             for (compstatus = (FcitxUIComplexStatus*) utarray_front(uicompstats);
@@ -346,8 +311,9 @@ void DrawMainWindow(MainWindow* mainWindow)
                 else
                     path = strdup(icon);
                 SkinImage* statusicon = LoadImage(sc, path, false);
-                if (statusicon == NULL)
+                if (statusicon == NULL) {
                     continue;
+                }
                 else {
                     if (icon[0] == '/' && activeIcon) {
                         ResizeSurface(&statusicon->image,
@@ -376,6 +342,15 @@ void DrawMainWindow(MainWindow* mainWindow)
                 else
                     asprintf(&path, "%s_inactive.png", status->name);
                 SkinImage* statusicon = LoadImage(sc, path, false);
+                if (statusicon == NULL || statusicon->textIcon) {
+                    if (activeIcon) {
+                        statusicon = LoadImageWithText(classicui, sc, path, status->shortDescription,
+                                                       cairo_image_surface_get_width(activeIcon->image),
+                                                       cairo_image_surface_get_height(activeIcon->image),
+                                                       active
+                                                      );
+                    }
+                }
                 free(path);
                 if (statusicon == NULL)
                     continue;
@@ -413,20 +388,16 @@ void DrawMainWindow(MainWindow* mainWindow)
                 UpdateStatusGeometry(&mainWindow->imiconstat, imicon, currentX, sc->skinMainBar.marginTop);
                 currentX += cairo_image_surface_get_width(imicon->image);
             }
-            
+
             for (compstatus = (FcitxUIComplexStatus*) utarray_front(uicompstats);
                  compstatus != NULL;
                  compstatus = (FcitxUIComplexStatus*) utarray_next(uicompstats, compstatus)
                 ) {
-                
-                if (!compstatus->visible)
-                    continue;
                 FcitxClassicUIStatus* privstat = GetPrivateStatus(compstatus);
                 if (privstat == NULL)
                     continue;
-                /* reset status */
-                privstat->x = privstat->y = -1;
-                privstat->w = privstat->h = 0;
+                if (!compstatus->visible)
+                    continue;
                 const char* icon = compstatus->getIconName(compstatus->arg);
                 char* path;
                 if (icon[0] != '/')
@@ -437,6 +408,7 @@ void DrawMainWindow(MainWindow* mainWindow)
                 free(path);
                 if (statusicon == NULL)
                     continue;
+                privstat->avail = true;
                 DrawImage(c, statusicon->image, currentX, sc->skinMainBar.marginTop, privstat->mouse);
                 UpdateStatusGeometry(privstat, statusicon, currentX, sc->skinMainBar.marginTop);
                 currentX += cairo_image_surface_get_width(statusicon->image);
@@ -446,14 +418,12 @@ void DrawMainWindow(MainWindow* mainWindow)
                     status != NULL;
                     status = (FcitxUIStatus*) utarray_next(uistats, status)
                 ) {
-                if (!status->visible)
-                    continue;
                 FcitxClassicUIStatus* privstat = GetPrivateStatus(status);
                 if (privstat == NULL)
                     continue;
+                if (!status->visible)
+                    continue;
                 /* reset status */
-                privstat->x = privstat->y = -1;
-                privstat->w = privstat->h = 0;
                 boolean active =  status->getCurrentStatus(status->arg);
                 char *path;
                 if (active)
@@ -464,6 +434,7 @@ void DrawMainWindow(MainWindow* mainWindow)
                 free(path);
                 if (statusicon == NULL)
                     continue;
+                privstat->avail = true;
                 DrawImage(c, statusicon->image, currentX, sc->skinMainBar.marginTop, privstat->mouse);
                 UpdateStatusGeometry(privstat, statusicon, currentX, sc->skinMainBar.marginTop);
                 currentX += cairo_image_surface_get_width(statusicon->image);
@@ -516,7 +487,8 @@ void UpdateStatusGeometry(FcitxClassicUIStatus *privstat, SkinImage *image, int
 
 void CloseMainWindow(MainWindow *mainWindow)
 {
-    XUnmapWindow(mainWindow->dpy, mainWindow->window);
+    if (mainWindow->owner->hideMainWindow != HM_SHOW || mainWindow->owner->isSuspend)
+        XUnmapWindow(mainWindow->dpy, mainWindow->window);
 }
 
 boolean SetMouseStatus(MainWindow *mainWindow, MouseE* mouseE, MouseE value, MouseE other)
@@ -630,16 +602,16 @@ boolean MainWindowEventHandler(void *arg, XEvent* event)
                     classicui->iMainWindowOffsetY = event->xbutton.y;
 
                     if (!ClassicUIMouseClick(mainWindow->owner, mainWindow->window, &classicui->iMainWindowOffsetX, &classicui->iMainWindowOffsetY)) {
-                        if (FcitxInstanceGetCurrentState(instance) == IS_CLOSED) {
-                            FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
-                        } else {
-                            FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
-                        }
+                        FcitxInstanceChangeIMState(instance, FcitxInstanceGetCurrentIC(instance));
                     }
                     SaveClassicUIConfig(classicui);
                 } else if (IsInRspArea(event->xbutton.x, event->xbutton.y, &mainWindow->imiconstat)) {
                     mouse = &mainWindow->imiconstat.mouse;
-                    FcitxInstanceSwitchIM(instance, -1);
+                    /* if close, we enable it, if enabled, we scroll it */
+                    if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE)
+                        FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
+                    else
+                        FcitxInstanceSwitchIM(instance, -1);
                 } else {
                     FcitxUIComplexStatus *compstatus;
                     UT_array* uicompstats = FcitxInstanceGetUIComplexStats(instance);
@@ -703,9 +675,6 @@ boolean MainWindowEventHandler(void *arg, XEvent* event)
                 if (SetMouseStatus(mainWindow, NULL, RELEASE, RELEASE))
                     DrawMainWindow(mainWindow);
                 break;
-            case Button2:
-                DisplayAboutWindow(mainWindow->owner->aboutWindow);
-                break;
             }
             break;
         }
@@ -714,46 +683,14 @@ boolean MainWindowEventHandler(void *arg, XEvent* event)
     return false;
 }
 
-static void ResizeSurface(cairo_surface_t** surface, int w, int h)
-{
-    int ow = cairo_image_surface_get_width(*surface);
-    int oh = cairo_image_surface_get_height(*surface);
-
-    if ((ow == w && oh == h) || w == 0 || h == 0 || ow == 0 || oh == 0)
-        return;
-    
-    double scalex = (double)w / ow;
-    double scaley = (double)h / oh;
-    double scale = (scalex > scaley) ? scaley : scalex;
-
-    int nw = ow * scale;
-    int nh = oh * scale;
-
-    cairo_surface_t* newsurface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, w, h);
-    cairo_t* c = cairo_create(newsurface);
-    cairo_set_operator(c, CAIRO_OPERATOR_SOURCE);
-    cairo_set_source_rgba(c ,1, 1, 1, 0.0);
-    cairo_paint(c);
-    cairo_translate(c, (w - nw) / 2.0 , (h - nh) / 2.0);
-    cairo_scale(c, scale, scale);
-    cairo_set_source_surface(c, *surface, 0, 0);
-    cairo_rectangle(c, 0, 0, ow, oh);
-    cairo_clip(c);
-    cairo_paint(c);
-    cairo_destroy(c);
-
-    cairo_surface_destroy(*surface);
-
-    *surface = newsurface;
-}
-
 static void UpdateMenuGeometry(MainWindow* mainWindow, XlibMenu* menuWindow)
 {
     FcitxClassicUI* classicui = mainWindow->owner;
-    
+
     unsigned int height;
     int sheight;
     XWindowAttributes attr;
+    FcitxMenuUpdate(menuWindow->menushell);
     GetMenuSize(menuWindow);
     GetScreenSize(classicui, NULL, &sheight);
     XGetWindowAttributes(classicui->dpy, mainWindow->window, &attr);
