@@ -43,13 +43,13 @@
 #include "MainWindow.h"
 #include "TrayWindow.h"
 #include "MenuWindow.h"
-#include "AboutWindow.h"
 #include "MessageWindow.h"
 #include "fcitx/hook.h"
-#include <fcitx-utils/utils.h>
+#include "fcitx-utils/utils.h"
 
 struct _FcitxSkin;
-boolean MainMenuAction(FcitxUIMenu* menu, int index);
+static boolean MainMenuAction(FcitxUIMenu* menu, int index);
+static void UpdateMainMenu(FcitxUIMenu* menu);
 
 static void* ClassicUICreate(FcitxInstance* instance);
 static void ClassicUICloseInputWindow(void* arg);
@@ -139,26 +139,9 @@ void* ClassicUICreate(FcitxInstance* instance)
 
     InitSkinMenu(classicui);
     FcitxUIRegisterMenu(instance, &classicui->skinMenu);
-
     /* Main Menu Initial */
     FcitxMenuInit(&classicui->mainMenu);
-    FcitxMenuAddMenuItem(&classicui->mainMenu, _("About Fcitx"), MENUTYPE_SIMPLE, NULL);
-    FcitxMenuAddMenuItem(&classicui->mainMenu, _("Online Help"), MENUTYPE_SIMPLE, NULL);
-    FcitxMenuAddMenuItem(&classicui->mainMenu, NULL, MENUTYPE_DIVLINE, NULL);
-
-    FcitxUIMenu **menupp;
-    UT_array* uimenus = FcitxInstanceGetUIMenus(instance);
-    for (menupp = (FcitxUIMenu **) utarray_front(uimenus);
-            menupp != NULL;
-            menupp = (FcitxUIMenu **) utarray_next(uimenus, menupp)
-        ) {
-        FcitxUIMenu * menup = *menupp;
-        if (!menup->isSubMenu)
-            FcitxMenuAddMenuItem(&classicui->mainMenu, menup->name, MENUTYPE_SUBMENU, menup);
-    }
-    FcitxMenuAddMenuItem(&classicui->mainMenu, NULL, MENUTYPE_DIVLINE, NULL);
-    FcitxMenuAddMenuItem(&classicui->mainMenu, _("Configure"), MENUTYPE_SIMPLE, NULL);
-    FcitxMenuAddMenuItem(&classicui->mainMenu, _("Exit"), MENUTYPE_SIMPLE, NULL);
+    classicui->mainMenu.UpdateMenu = UpdateMainMenu;
     classicui->mainMenu.MenuAction = MainMenuAction;
     classicui->mainMenu.priv = classicui;
     classicui->mainMenu.mark = -1;
@@ -167,7 +150,6 @@ void* ClassicUICreate(FcitxInstance* instance)
     classicui->inputWindow = CreateInputWindow(classicui);
     classicui->mainWindow = CreateMainWindow(classicui);
     classicui->trayWindow = CreateTrayWindow(classicui);
-    classicui->aboutWindow = CreateAboutWindow(classicui);
     classicui->messageWindow = CreateMessageWindow(classicui);
     classicui->mainMenuWindow = CreateMainMenuWindow(classicui);
 
@@ -279,7 +261,7 @@ static void ClassicUIOnInputUnFocus(void *arg)
     FcitxClassicUI* classicui = (FcitxClassicUI*) arg;
     if (classicui->isSuspend)
         return;
-    CloseMainWindow(classicui->mainWindow);
+    DrawMainWindow(classicui->mainWindow);
     DrawTrayWindow(classicui->trayWindow);
 }
 
@@ -421,14 +403,88 @@ void ClassicUIDisplayMessage(void* arg, char* title, char** msg, int length)
     DrawMessageWindow(classicui->messageWindow, title, msg, length);
 }
 
+
+static void UpdateMainMenu(FcitxUIMenu* menu)
+{
+    FcitxClassicUI* classicui = (FcitxClassicUI*) menu->priv;
+    FcitxInstance* instance = classicui->owner;
+    FcitxMenuClear(menu);
+
+    FcitxMenuAddMenuItem(menu, _("Online Help"), MENUTYPE_SIMPLE, NULL);
+    FcitxMenuAddMenuItem(menu, NULL, MENUTYPE_DIVLINE, NULL);
+    boolean flag = false;
+
+    FcitxUIStatus* status;
+    UT_array* uistats = FcitxInstanceGetUIStats(instance);
+    for (status = (FcitxUIStatus*) utarray_front(uistats);
+            status != NULL;
+            status = (FcitxUIStatus*) utarray_next(uistats, status)
+        ) {
+        FcitxClassicUIStatus* privstat =  GetPrivateStatus(status);
+        if (privstat == NULL || !status->visible || privstat->avail )
+            continue;
+
+        flag = true;
+        FcitxMenuAddMenuItemWithData(menu, status->shortDescription, MENUTYPE_SIMPLE, NULL, strdup(status->name));
+    }
+
+    FcitxUIComplexStatus* compstatus;
+    UT_array* uicompstats = FcitxInstanceGetUIComplexStats(instance);
+    for (compstatus = (FcitxUIComplexStatus*) utarray_front(uicompstats);
+            compstatus != NULL;
+            compstatus = (FcitxUIComplexStatus*) utarray_next(uicompstats, compstatus)
+        ) {
+        FcitxClassicUIStatus* privstat =  GetPrivateStatus(compstatus);
+        if (privstat == NULL || !compstatus->visible || privstat->avail)
+            continue;
+        if (FcitxUIGetMenuByStatusName(instance, compstatus->name))
+            continue;
+
+        flag = true;
+        FcitxMenuAddMenuItemWithData(menu, compstatus->shortDescription, MENUTYPE_SIMPLE, NULL, strdup(compstatus->name));
+    }
+
+    if (flag)
+        FcitxMenuAddMenuItem(menu, NULL, MENUTYPE_DIVLINE, NULL);
+
+    FcitxUIMenu **menupp;
+    UT_array* uimenus = FcitxInstanceGetUIMenus(instance);
+    for (menupp = (FcitxUIMenu **) utarray_front(uimenus);
+            menupp != NULL;
+            menupp = (FcitxUIMenu **) utarray_next(uimenus, menupp)
+        ) {
+        FcitxUIMenu * menup = *menupp;
+        if (menup->isSubMenu)
+            continue;
+
+        if (!menup->visible)
+            continue;
+
+        if (menup->candStatusBind) {
+            FcitxUIComplexStatus* compStatus = FcitxUIGetComplexStatusByName(instance, menup->candStatusBind);
+            if (compStatus) {
+                if (!compStatus->visible)
+                    continue;
+                if (GetPrivateStatus(compStatus)->avail)
+                    continue;
+            }
+        }
+
+        FcitxMenuAddMenuItem(menu, menup->name, MENUTYPE_SUBMENU, menup);
+    }
+    FcitxMenuAddMenuItem(menu, NULL, MENUTYPE_DIVLINE, NULL);
+    FcitxMenuAddMenuItem(menu, _("Configure Current Input Method"), MENUTYPE_SIMPLE, NULL);
+    FcitxMenuAddMenuItem(menu, _("Configure"), MENUTYPE_SIMPLE, NULL);
+    FcitxMenuAddMenuItem(menu, _("Exit"), MENUTYPE_SIMPLE, NULL);
+}
+
 boolean MainMenuAction(FcitxUIMenu* menu, int index)
 {
     FcitxClassicUI* classicui = (FcitxClassicUI*) menu->priv;
+    FcitxInstance* instance = classicui->owner;
     int length = utarray_len(&menu->shell);
     if (index == 0) {
-        DisplayAboutWindow(classicui->mainWindow->owner->aboutWindow);
-    } else if (index == 1) {
-        FILE* p = popen("xdg-open http://fcitx.github.com/handbook/ &", "r");
+        FILE* p = popen("xdg-open http://fcitx-im.org/ &", "r");
         if (p)
             pclose(p);
         else
@@ -436,13 +492,18 @@ boolean MainMenuAction(FcitxUIMenu* menu, int index)
     } else if (index == length - 1) { /* Exit */
         FcitxInstanceEnd(classicui->owner);
     } else if (index == length - 2) { /* Configuration */
-        char* command = fcitx_utils_get_fcitx_path_with_filename("bindir", "/fcitx-configtool &");
-        FILE* p = popen(command, "r");
-        free(command);
-        if (p)
-            pclose(p);
-        else
-            FcitxLog(ERROR, _("Unable to create process"));
+        fcitx_utils_launch_configure_tool();
+    } else if (index == length - 3) { /* Configuration */
+        FcitxIM* im = FcitxInstanceGetCurrentIM(classicui->owner);
+        if (im && im->owner) {
+            fcitx_utils_launch_configure_tool_for_addon(im->owner->name);
+        }
+    } else {
+        FcitxMenuItem* item = (FcitxMenuItem*) utarray_eltptr(&menu->shell, index);
+        if (item && item->type == MENUTYPE_SIMPLE && item->data) {
+            const char* name = item->data;
+            FcitxUIUpdateStatus(instance, name);
+        }
     }
     return true;
 }
@@ -530,21 +591,55 @@ boolean EnlargeCairoSurface(cairo_surface_t** sur, int w, int h)
 {
     int ow = cairo_image_surface_get_width(*sur);
     int oh = cairo_image_surface_get_height(*sur);
-    
+
     if (ow >= w && oh >= h)
         return false;
-    
+
     while (ow < w) {
         ow *= 2;
     }
-    
+
     while (oh < h) {
         oh *= 2;
     }
-    
+
     cairo_surface_destroy(*sur);
     *sur = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, ow, oh);
     return true;
 }
 
+void ResizeSurface(cairo_surface_t** surface, int w, int h)
+{
+    int ow = cairo_image_surface_get_width(*surface);
+    int oh = cairo_image_surface_get_height(*surface);
+
+    if ((ow == w && oh == h) || w == 0 || h == 0 || ow == 0 || oh == 0)
+        return;
+
+    double scalex = (double)w / ow;
+    double scaley = (double)h / oh;
+    double scale = (scalex > scaley) ? scaley : scalex;
+
+    int nw = ow * scale;
+    int nh = oh * scale;
+
+    cairo_surface_t* newsurface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, w, h);
+    cairo_t* c = cairo_create(newsurface);
+    cairo_set_operator(c, CAIRO_OPERATOR_SOURCE);
+    cairo_set_source_rgba(c ,1, 1, 1, 0.0);
+    cairo_paint(c);
+    cairo_translate(c, (w - nw) / 2.0 , (h - nh) / 2.0);
+    cairo_scale(c, scale, scale);
+    cairo_set_source_surface(c, *surface, 0, 0);
+    cairo_rectangle(c, 0, 0, ow, oh);
+    cairo_clip(c);
+    cairo_paint(c);
+    cairo_destroy(c);
+
+    cairo_surface_destroy(*surface);
+
+    *surface = newsurface;
+}
+
+
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
