@@ -59,7 +59,6 @@ typedef struct _FcitxClassicUI {
     struct _MainWindow* mainWindow;
     struct _MessageWindow* messageWindow;
     struct _TrayWindow* trayWindow;
-    struct _AboutWindow* aboutWindow;
     FcitxUIMenu skinMenu;
 
     FcitxSkin skin;
@@ -99,6 +98,7 @@ boolean LoadClassicUIConfig(FcitxClassicUI* classicui);
 void SaveClassicUIConfig(FcitxClassicUI* classicui);
 boolean WindowIsVisable(Display* dpy, Window window);
 boolean EnlargeCairoSurface(cairo_surface_t** sur, int w, int h);
+void ResizeSurface(cairo_surface_t** surface, int w, int h);
 
 #define GetPrivateStatus(status) ((FcitxClassicUIStatus*)(status)->uipriv[classicui->isfallback])
 
