@@ -64,6 +64,7 @@ static const UT_icd place_icd = {sizeof(SkinPlacement), NULL, NULL, NULL };
 static boolean SkinMenuAction(FcitxUIMenu* menu, int index);
 static void UpdateSkinMenu(FcitxUIMenu* menu);
 static void UnloadImage(FcitxSkin* skin);
+static void UnloadSingleImage(FcitxSkin* sc, const char* name);
 
 CONFIG_DESC_DEFINE(GetSkinDesc, "skin.desc")
 
@@ -157,54 +158,134 @@ reload:
 
 }
 
-SkinImage* LoadImage(FcitxSkin* sc, const char* name, boolean fallback)
+SkinImage* LoadImageWithText(FcitxClassicUI* classicui, FcitxSkin* sc, const char* name, const char* text, int w, int h, boolean active)
+{
+    if (!text || strlen(text) == 0)
+        return NULL;
+
+    UnloadSingleImage(sc, name);
+
+
+    int len = fcitx_utf8_char_len(text);
+
+    char* iconText = strndup(text, len);
+
+    FcitxLog(DEBUG, "%s", iconText);
+
+    cairo_surface_t* newsurface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, w, h);
+    cairo_t* c = cairo_create(newsurface);
+
+    int min = w > h? h: w;
+    min = min * 0.8;
+
+    cairo_set_operator(c, CAIRO_OPERATOR_SOURCE);
+    cairo_set_source_rgba(c ,1, 1, 1, 0.0);
+    cairo_paint(c);
+
+    FcitxConfigColor color;
+    if (sc->skinMainBar.bUseCustomTextIconColor) {
+        if (active)
+            color = sc->skinMainBar.textIconColor[0];
+        else
+            color = sc->skinMainBar.textIconColor[1];
+    }
+    else
+        color = sc->skinFont.menuFontColor[1];
+
+    int textw, texth;
+    StringSizeStrict(iconText, classicui->font, min, &textw, &texth);
+
+    int fh = FontHeight(classicui->font, min);
+    OutputString(c, iconText, classicui->font, min, (w - textw) * 0.5, (h - texth) * 0.5 - (fh - texth), &color);
+
+    cairo_destroy(c);
+    SkinImage* image = fcitx_utils_malloc0(sizeof(SkinImage));
+    image->name = strdup(name);
+    image->image = newsurface;
+    image->textIcon = true;
+    HASH_ADD_KEYPTR(hh, sc->imageTable, image->name, strlen(image->name), image);
+    return image;
+}
+
+SkinImage* LoadImageFromTable(SkinImage** imageTable, const char* skinType, const char* name, int flag)
 {
     cairo_surface_t *png = NULL;
     SkinImage *image = NULL;
+    char* buf = NULL;
+    asprintf(&buf, "skin/%s", skinType);
+    const char* fallbackChainNoFallback[] = { buf };
+    const char* fallbackChainPanel[] = { buf, "skin/default" };
+    const char* fallbackChainTray[] = { "imicon" };
+    const char* fallbackChainPanelIMIcon[] = { buf, "imicon" };
 
-    HASH_FIND_STR(sc->imageTable, name, image);
-    if (image != NULL)
+    HASH_FIND_STR(*imageTable, name, image);
+    if (image != NULL) {
         return image;
+    }
+
+    const char** fallbackChain;
+    int fallbackSize;
+    switch(flag) {
+        case 1:
+            fallbackChain = fallbackChainPanel;
+            fallbackSize = 2;
+            break;
+        case 2:
+            fallbackChain = fallbackChainTray;
+            fallbackSize = 1;
+            break;
+        case 3:
+            fallbackChain = fallbackChainPanelIMIcon;
+            fallbackSize = 3;
+            break;
+        case 0:
+        default:
+            /* fall through */
+            fallbackChain = fallbackChainNoFallback;
+            fallbackSize = 1;
+            break;
+    }
+
     if (strlen(name) > 0 && strcmp(name , "NONE") != 0) {
-        char *skintype = strdup(*sc->skinType);
-        char *filename;
-        while (true) {
-            char* buf = NULL;
-            asprintf(&buf, "skin/%s", skintype);
+        int i = 0;
+        for (i = 0; i < fallbackSize; i ++) {
+            char* filename;
+            const char* skintype = fallbackChain[i];
 
-            FILE* fp = FcitxXDGGetFileWithPrefix(buf, name, "r", &filename);
-            free(buf);
-
-            Bool flagNoFile = (fp == NULL);
+            FILE* fp = FcitxXDGGetFileWithPrefix(skintype, name, "r", &filename);
             if (fp) {
-                fclose(fp);
-
                 png = cairo_image_surface_create_from_png(filename);
-                break;
-            }
-            if (flagNoFile && (!fallback || strcmp(skintype, "default") == 0)) {
-                png = NULL;
-                break;
+                if (cairo_surface_status (png)) {
+                    png = NULL;
+                }
             }
 
             free(filename);
-            free(skintype);
-            skintype = strdup("default");
+
+            if (png)
+                break;
         }
-        free(filename);
-        free(skintype);
     }
+    free(buf);
 
     if (png != NULL) {
         image = fcitx_utils_malloc0(sizeof(SkinImage));
         image->name = strdup(name);
         image->image = png;
-        HASH_ADD_KEYPTR(hh, sc->imageTable, image->name, strlen(image->name), image);
+        HASH_ADD_KEYPTR(hh, *imageTable, image->name, strlen(image->name), image);
         return image;
     }
     return NULL;
 }
 
+SkinImage* LoadImage(FcitxSkin* sc, const char* name, int flag)
+{
+    if (flag == 2)
+        return LoadImageFromTable(&sc->trayImageTable, *sc->skinType, name, flag);
+    else
+        return LoadImageFromTable(&sc->imageTable, *sc->skinType, name, flag);
+}
+
 void DrawResizableBackground(cairo_t *c,
                              cairo_surface_t *background,
                              int height,
@@ -642,7 +723,7 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
     } else {
         newWidth = (newWidth < INPUT_BAR_HMIN_WIDTH) ? INPUT_BAR_HMIN_WIDTH : newWidth;
     }
-    
+
     *iwidth = newWidth;
     *iheight = newHeight;
 
@@ -748,9 +829,9 @@ void DisplaySkin(FcitxClassicUI* classicui, char * skinname)
     SaveClassicUIConfig(classicui);
 }
 
-void UnloadImage(FcitxSkin* skin)
+void FreeImageTable(SkinImage* table)
 {
-    SkinImage *images = skin->imageTable;
+    SkinImage *images = table;
     while (images) {
         SkinImage* curimage = images;
         HASH_DEL(images, curimage);
@@ -758,7 +839,28 @@ void UnloadImage(FcitxSkin* skin)
         cairo_surface_destroy(curimage->image);
         free(curimage);
     }
+}
+
+void UnloadImage(FcitxSkin* skin)
+{
+    FreeImageTable(skin->imageTable);
     skin->imageTable = NULL;
+
+    FreeImageTable(skin->trayImageTable);
+    skin->trayImageTable = NULL;
+}
+
+void UnloadSingleImage(FcitxSkin* sc, const char* name)
+{
+    SkinImage *image;
+    HASH_FIND_STR(sc->imageTable, name, image);
+    if (image != NULL) {
+        SkinImage* curimage = image;
+        HASH_DEL(sc->imageTable, image);
+        free(curimage->name);
+        cairo_surface_destroy(curimage->image);
+        free(curimage);
+    }
 }
 
 //图片文件加载函数完成
@@ -818,7 +920,7 @@ void InitSkinMenu(FcitxClassicUI* classicui)
 {
     utarray_init(&classicui->skinBuf, &ut_str_icd);
     FcitxMenuInit(&classicui->skinMenu);
-    classicui->skinMenu.candStatusBind = strdup("skin");
+    classicui->skinMenu.candStatusBind = NULL;
     classicui->skinMenu.name =  strdup(_("Skin"));
 
     classicui->skinMenu.UpdateMenu = UpdateSkinMenu;
@@ -880,4 +982,28 @@ void ParsePlacement(UT_array* sps, char* placment)
 
     utarray_free(array);
 }
+
+SkinImage* GetIMIcon(FcitxInstance* instance, FcitxSkin *sc, const char* fallbackIcon, int flag, boolean fallbackToDefault)
+{
+    FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
+    char* path;
+    if (im->strIconName[0] == '/')
+        path = strdup(im->strIconName);
+    else
+        asprintf(&path, "%s.png", im->strIconName);
+    SkinImage* imicon = LoadImage(sc, path, flag);
+    if (imicon == NULL)
+        imicon = LoadImage(sc, fallbackIcon, fallbackToDefault);
+    else {
+        SkinImage* activeIcon = LoadImage(sc, fallbackIcon, fallbackToDefault);
+        if (activeIcon) {
+            ResizeSurface(&imicon->image,
+                          cairo_image_surface_get_width(activeIcon->image),
+                          cairo_image_surface_get_height(activeIcon->image));
+        }
+    }
+    free(path);
+    return imicon;
+}
+
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
