@@ -1,6 +1,6 @@
 /***************************************************************************
  *   Copyright (C) 2008~2010 by Zealot.Hoi                                 *
- *   Copyright (C) 2010~2011 by CSSlayer                                   *
+ *   Copyright (C) 2010~2012 by CSSlayer                                   *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -115,6 +115,10 @@ typedef struct _FcitxKimpanelUI {
     int iCursorPos;
     int lastUpdateY;
     int lastUpdateX;
+    int version;
+    int lastUpdateW;
+    int lastUpdateH;
+    int lastCursor;
 } FcitxKimpanelUI;
 
 static void* KimpanelCreate(FcitxInstance* instance);
@@ -130,11 +134,14 @@ static void KimpanelOnInputFocus(void *arg);
 static void KimpanelOnInputUnFocus(void *arg);
 static void KimpanelOnTriggerOn(void *arg);
 static void KimpanelOnTriggerOff(void *arg);
+static void KimpanelSuspend(void* arg);
 static void KimpanelDestroy(void* arg);
 
+static void KimUpdateLookupTableCursor(FcitxKimpanelUI* kimpanel, int cursor);
 static void KimShowAux(FcitxKimpanelUI* kimpanel, boolean toShow);
 static void KimShowPreedit(FcitxKimpanelUI* kimpanel, boolean toShow);
 static void KimUpdateSpotLocation(FcitxKimpanelUI* kimpanel, int x, int y);
+static void KimSetSpotRect(FcitxKimpanelUI* kimpanel, int x, int y, int w, int h);
 static void KimShowLookupTable(FcitxKimpanelUI* kimpanel, boolean toShow);
 static void KimUpdateLookupTable(FcitxKimpanelUI* kimpanel, char *labels[], int nLabel, char *texts[], int nText, boolean has_prev, boolean has_next);
 static void KimUpdatePreeditText(FcitxKimpanelUI* kimpanel, char *text);
@@ -153,6 +160,8 @@ static void KimpanelRegisterAllStatus(FcitxKimpanelUI* kimpanel);
 static void KimpanelSetIMStatus(FcitxKimpanelUI* kimpanel);
 static void KimExecMenu(FcitxKimpanelUI* kimpanel, char *props[], int n);
 static void KimpanelServiceExistCallback(DBusPendingCall *call, void *data);
+static void KimpanelIntrospect(FcitxKimpanelUI* kimpanel);
+static void KimpanelIntrospectCallback(DBusPendingCall *call, void *data);
 
 #define KIMPANEL_BUFFER_SIZE 4096
 
@@ -172,7 +181,7 @@ FcitxUI ui = {
     NULL,
     NULL,
     NULL,
-    NULL,
+    KimpanelSuspend,
     NULL,
     KimpanelDestroy,
     KimpanelRegisterComplexStatus,
@@ -188,6 +197,8 @@ void* KimpanelCreate(FcitxInstance* instance)
     FcitxKimpanelUI *kimpanel = fcitx_utils_malloc0(sizeof(FcitxKimpanelUI));
     FcitxModuleFunctionArg arg;
 
+    kimpanel->lastCursor = -2;
+    kimpanel->version = 1;
     kimpanel->iCursorPos = 0;
     kimpanel->owner = instance;
     kimpanel->conn = InvokeFunction(instance, FCITX_DBUS, GETCONNECTION, arg);
@@ -209,6 +220,14 @@ void* KimpanelCreate(FcitxInstance* instance)
             FcitxLog(ERROR, "Match Error (%s)", err.message);
             break;
         }
+        dbus_bus_add_match(kimpanel->conn,
+                           "type='signal',interface='org.kde.impanel2'",
+                           &err);
+        dbus_connection_flush(kimpanel->conn);
+        if (dbus_error_is_set(&err)) {
+            FcitxLog(ERROR, "Match Error (%s)", err.message);
+            break;
+        }
 
         dbus_bus_add_match(kimpanel->conn,
                            "type='signal',"
@@ -243,7 +262,6 @@ void* KimpanelCreate(FcitxInstance* instance)
         const char* kimpanelServiceName = "org.kde.impanel";
         DBusMessage* message = dbus_message_new_method_call(DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS, "NameHasOwner");
         dbus_message_append_args(message, DBUS_TYPE_STRING, &kimpanelServiceName, DBUS_TYPE_INVALID);
-        dbus_connection_send(kimpanel->conn, message, NULL);
 
         DBusPendingCall* call = NULL;
         dbus_bool_t reply = dbus_connection_send_with_reply(kimpanel->conn, message,
@@ -256,6 +274,7 @@ void* KimpanelCreate(FcitxInstance* instance)
                                          NULL);
         }
         dbus_connection_flush(kimpanel->conn);
+        dbus_message_unref(message);
 
         KimpanelRegisterAllStatus(kimpanel);
         dbus_error_free(&err);
@@ -274,7 +293,7 @@ void KimpanelRegisterAllStatus(FcitxKimpanelUI* kimpanel)
     UT_array* uicompstats = FcitxInstanceGetUIComplexStats(instance);
     char **prop = fcitx_utils_malloc0(sizeof(char*) * (2 + utarray_len(uistats) + utarray_len(uicompstats)));
     asprintf(&prop[0], "/Fcitx/logo:%s:%s:%s", _("Fcitx"), "fcitx", _("Fcitx"));
-    
+
     char* icon;
     char* imname;
     char* description;
@@ -294,7 +313,7 @@ void KimpanelRegisterAllStatus(FcitxKimpanelUI* kimpanel)
         imname = _("Disabled");
         description = _("Input Method Disabled");
     }
-    /* add fcitx- prefix */
+    /* add fcitx- prefix, unless icon name is an absolute path */
     if (icon[0] == '/')
         asprintf(&prop[1], "/Fcitx/im:%s:%s:%s", imname, icon, description);
     else
@@ -337,7 +356,13 @@ void KimpanelSetIMStatus(FcitxKimpanelUI* kimpanel)
     char* icon;
     char* imname;
     char* description;
-    if (FcitxInstanceGetCurrentStatev2(instance) == IS_ACTIVE) {
+    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
+    if (ic == NULL) {
+        icon = "kbd";
+        imname = _("No input window");
+        description = _("No input window");
+    }
+    else if (FcitxInstanceGetCurrentStatev2(instance) == IS_ACTIVE) {
         FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
         if (im) {
             icon = im->strIconName;
@@ -414,12 +439,15 @@ void KimpanelMoveInputWindow(void* arg)
     kimpanel->iOffsetX = 12;
     kimpanel->iOffsetY = 0;
 
-    int x = 0, y = 0;
+    int x = 0, y = 0, w = 0, h = 0;
 
     FcitxInputContext* ic = FcitxInstanceGetCurrentIC(kimpanel->owner);
-    FcitxInstanceGetWindowPosition(kimpanel->owner, ic, &x, &y);
+    FcitxInstanceGetWindowRect(kimpanel->owner, ic, &x, &y, &w, &h);
 
-    KimUpdateSpotLocation(kimpanel, x, y);
+    if (kimpanel->version == 1)
+        KimUpdateSpotLocation(kimpanel, x, y + h);
+    else
+        KimSetSpotRect(kimpanel, x, y, w, h);
 }
 
 void KimpanelRegisterMenu(void* arg, FcitxUIMenu* menu)
@@ -459,9 +487,22 @@ char* Status2String(FcitxUIStatus* status)
 char* ComplexStatus2String(FcitxUIComplexStatus* status)
 {
     const char* iconName = status->getIconName(status->arg);
-    
+
+    const char* templ;
+    switch(iconName[0]) {
+        case '/':
+            templ = "/Fcitx/%s:%s:%s:%s";
+            break;
+        case '\0':
+            templ = "/Fcitx/%s:%s:%s:%s";
+            break;
+        default:
+            templ = "/Fcitx/%s:%s:fcitx-%s:%s";
+            break;
+    }
+
     char *result = NULL;
-    asprintf(&result, iconName[0] == '/' ? "/Fcitx/%s:%s:%s:%s" :  "/Fcitx/%s:%s:fcitx-%s:%s",
+    asprintf(&result, templ,
              status->name,
              status->shortDescription,
              iconName,
@@ -480,6 +521,7 @@ void KimpanelShowInputWindow(void* arg)
     FcitxMessages* messageDown = kimpanel->messageDown;
     FcitxMessages* messageUp = kimpanel->messageUp;
     FcitxLog(DEBUG, "KimpanelShowInputWindow");
+    KimpanelMoveInputWindow(kimpanel);
 
     int n = FcitxMessagesGetMessageCount(messageDown);
     int nLabels = 0;
@@ -488,6 +530,7 @@ void KimpanelShowInputWindow(void* arg)
     char *text[33];
     char cmb[KIMPANEL_BUFFER_SIZE] = "";
     int i;
+    int pos = -1;
 
     if (n) {
         for (i = 0; i < n; i++) {
@@ -513,11 +556,13 @@ void KimpanelShowInputWindow(void* arg)
                     msgstr = needfree;
                 else
                     msgstr = FcitxMessagesGetMessageString(messageDown, i);
-                
+
                 if (strlen(cmb) + strlen(msgstr) + 1 < KIMPANEL_BUFFER_SIZE)
                     strcat(cmb, msgstr);
                 if (needfree)
                     free(needfree);
+                if (FcitxMessagesGetMessageType(messageDown, i) == MSG_FIRSTCAND)
+                    pos = nTexts;
             }
         }
         text[nTexts++] = strdup(cmb);
@@ -557,6 +602,7 @@ void KimpanelShowInputWindow(void* arg)
                              FcitxCandidateWordHasNext(FcitxInputStateGetCandidateList(input)));
         KimShowLookupTable(kimpanel, false);
     }
+    KimUpdateLookupTableCursor(kimpanel, pos);
 
     n = FcitxMessagesGetMessageCount(messageUp);
     char aux[MESSAGE_MAX_LENGTH] = "";
@@ -624,6 +670,15 @@ static DBusHandlerResult KimpanelDBusEventHandler(DBusConnection *connection, DB
     return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
+void KimpanelReset(FcitxKimpanelUI* kimpanel)
+{
+    kimpanel->lastCursor = -2;
+    kimpanel->lastUpdateH = -2;
+    kimpanel->lastUpdateW = -2;
+    kimpanel->lastUpdateX = -2;
+    kimpanel->lastUpdateY = -2;
+}
+
 DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* msg, void* user_data)
 {
     FCITX_UNUSED(connection);
@@ -671,31 +726,22 @@ DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* ms
             if (strlen(s0) > len) {
                 s0 += len;
                 if (strcmp("logo", s0) == 0) {
-                    if (FcitxInstanceGetCurrentState(instance) == IS_CLOSED) {
-                        FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
-                    } else {
-                        FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
-                    }
+                    FcitxInstanceChangeIMState(instance, FcitxInstanceGetCurrentIC(instance));
                 } else if (strcmp("keyboard", s0) == 0) {
                     FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
                 } else if (strncmp("im/", s0, strlen("im/")) == 0) {
                     s0 += strlen("im/");
-                    int index = FcitxInstanceGetIMIndexByName(instance, s0);
-                    FcitxInstanceSwitchIM(instance, index);
-                    if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
-                        FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
-                    }
+                    FcitxInstanceSwitchIMByName(instance, s0);
                 } else if (strncmp("im", s0, strlen("im")) == 0) {
                     UT_array* imes = FcitxInstanceGetIMEs(instance);
                     FcitxIM* pim;
-                    int index = 1;
-                    size_t len = utarray_len(imes) + 1;
+                    int index = 0;
+                    size_t len = utarray_len(imes);
                     char **prop = fcitx_utils_malloc0(len * sizeof(char*));
-                    asprintf(&prop[0], "/Fcitx/keyboard:%s:fcitx-%s:%s", _("Disabled"), "kbd", _("Input Method Disabled"));
                     for (pim = (FcitxIM *) utarray_front(imes);
                             pim != NULL;
                             pim = (FcitxIM *) utarray_next(imes, pim)) {
-                        
+
                         if (pim->strIconName[0] == '/')
                             asprintf(&prop[index], "/Fcitx/im/%s:%s:%s:%s", pim->uniqueName, pim->strName, pim->strIconName, pim->strName);
                         else
@@ -723,7 +769,7 @@ DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* ms
                         FcitxUIMenu* menu = FcitxUIGetMenuByStatusName(instance, s0);
                         if (menu) {
                             menu->UpdateMenu(menu);
-                            
+
                             int i = 0, index = 0;
                             char **prop = fcitx_utils_malloc0(utarray_len(&menu->shell) * sizeof(char*));
                             for (i = 0; i < utarray_len(&menu->shell); i++) {
@@ -748,6 +794,14 @@ DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* ms
     } else if (dbus_message_is_signal(msg, "org.kde.impanel", "PanelCreated")) {
         FcitxLog(DEBUG, "PanelCreated");
         FcitxUIResumeFromFallback(instance);
+        KimpanelReset(kimpanel);
+        KimpanelRegisterAllStatus(kimpanel);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_signal(msg, "org.kde.impanel2", "PanelCreated2")) {
+        FcitxLog(DEBUG, "PanelCreated2");
+        FcitxUIResumeFromFallback(instance);
+        kimpanel->version = 2;
+        KimpanelReset(kimpanel);
         KimpanelRegisterAllStatus(kimpanel);
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_signal(msg, "org.kde.impanel", "Exit")) {
@@ -760,14 +814,7 @@ DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* ms
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_signal(msg, "org.kde.impanel", "Configure")) {
         FcitxLog(DEBUG, "Configure");
-
-        char* command = fcitx_utils_get_fcitx_path_with_filename("bindir", "/fcitx-configtool &");
-        FILE* p = popen(command, "r");
-        free(command);
-        if (p)
-            pclose(p);
-        else
-            FcitxLog(ERROR, _("Unable to create process"));
+        fcitx_utils_launch_configure_tool();
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_signal(msg, DBUS_INTERFACE_DBUS, "NameOwnerChanged")) {
         const char* service, *oldowner, *newowner;
@@ -906,7 +953,6 @@ void KimRegisterProperties(FcitxKimpanelUI* kimpanel, char *props[], int n)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -939,7 +985,6 @@ void KimUpdateProperty(FcitxKimpanelUI* kimpanel, char *prop)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -972,7 +1017,6 @@ void KimRemoveProperty(FcitxKimpanelUI* kimpanel, char *prop)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1004,7 +1048,6 @@ void KimEnable(FcitxKimpanelUI* kimpanel, boolean toEnable)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1037,7 +1080,6 @@ void KimShowAux(FcitxKimpanelUI* kimpanel, boolean toShow)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1070,13 +1112,46 @@ void KimShowPreedit(FcitxKimpanelUI* kimpanel, boolean toShow)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
 
 }
 
+void KimUpdateLookupTableCursor(FcitxKimpanelUI* kimpanel, int cursor)
+{
+    if (kimpanel->lastCursor != cursor)
+        kimpanel->lastCursor = cursor;
+    else
+        return;
+    dbus_uint32_t serial = 0; // unique number to associate replies with requests
+    DBusMessage* msg;
+    DBusMessageIter args;
+
+    // create a signal and check for errors
+    msg = dbus_message_new_signal(FCITX_KIMPANEL_PATH, // object name of the signal
+                                  FCITX_KIMPANEL_INTERFACE, // interface name of the signal
+                                  "UpdateLookupTableCursor"); // name of the signal
+    if (NULL == msg) {
+        FcitxLog(DEBUG, "Message Null");
+        return;
+    }
+
+    // append arguments onto signal
+    dbus_message_iter_init_append(msg, &args);
+    if (!dbus_message_iter_append_basic(&args, DBUS_TYPE_INT32, &cursor)) {
+        FcitxLog(DEBUG, "Out Of Memory!");
+    }
+
+    // send the message and flush the connection
+    if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
+        FcitxLog(DEBUG, "Out Of Memory!");
+    }
+
+    // free the message
+    dbus_message_unref(msg);
+}
+
 void KimShowLookupTable(FcitxKimpanelUI* kimpanel, boolean toShow)
 {
 
@@ -1103,7 +1178,6 @@ void KimShowLookupTable(FcitxKimpanelUI* kimpanel, boolean toShow)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1164,7 +1238,6 @@ void KimUpdateLookupTable(FcitxKimpanelUI* kimpanel, char *labels[], int nLabel,
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1197,7 +1270,6 @@ void KimUpdatePreeditCaret(FcitxKimpanelUI* kimpanel, int position)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1234,7 +1306,6 @@ void KimUpdatePreeditText(FcitxKimpanelUI* kimpanel, char *text)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1272,7 +1343,6 @@ void KimUpdateAux(FcitxKimpanelUI* kimpanel, char *text)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1311,13 +1381,54 @@ void KimUpdateSpotLocation(FcitxKimpanelUI* kimpanel, int x, int y)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
 
 }
 
+void KimSetSpotRect(FcitxKimpanelUI* kimpanel, int x, int y, int w, int h)
+{
+    if (kimpanel->lastUpdateX == x && kimpanel->lastUpdateY == y && kimpanel->lastUpdateW == w && kimpanel->lastUpdateH == h)
+        return;
+    kimpanel->lastUpdateX = x;
+    kimpanel->lastUpdateY = y;
+    kimpanel->lastUpdateW = w;
+    kimpanel->lastUpdateH = h;
+    dbus_uint32_t serial = 0; // unique number to associate replies with requests
+    DBusMessage* msg;
+
+    // create a signal and check for errors
+    msg = dbus_message_new_method_call("org.kde.impanel",
+                                  "/org/kde/impanel",
+                                  "org.kde.impanel2",
+                                  "SetSpotRect"); // name of the signal
+    if (NULL == msg) {
+        FcitxLog(DEBUG, "Message Null");
+        return;
+    }
+
+    if (!dbus_message_append_args(
+        msg,
+        DBUS_TYPE_INT32, &x,
+        DBUS_TYPE_INT32, &y,
+        DBUS_TYPE_INT32, &w,
+        DBUS_TYPE_INT32, &h,
+        DBUS_TYPE_INVALID
+    )) {
+        FcitxLog(DEBUG, "Out Of Memory!");
+    }
+
+    // send the message and flush the connection
+    if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
+        FcitxLog(DEBUG, "Out Of Memory!");
+    }
+
+    // free the message
+    dbus_message_unref(msg);
+}
+
+
 void KimUpdateScreen(FcitxKimpanelUI* kimpanel, int id)
 {
 
@@ -1344,7 +1455,6 @@ void KimUpdateScreen(FcitxKimpanelUI* kimpanel, int id)
     if (!dbus_connection_send(kimpanel->conn, msg, &serial)) {
         FcitxLog(DEBUG, "Out Of Memory!");
     }
-    dbus_connection_flush(kimpanel->conn);
 
     // free the message
     dbus_message_unref(msg);
@@ -1393,8 +1503,12 @@ void KimpanelServiceExistCallback(DBusPendingCall *call, void *data)
         DBusError error;
         dbus_error_init(&error);
         dbus_message_get_args(msg, &error, DBUS_TYPE_BOOLEAN, &has , DBUS_TYPE_INVALID);
-        if (!has)
+        if (!has) {
             FcitxUISwitchToFallback(kimpanel->owner);
+        }
+        else {
+            KimpanelIntrospect(kimpanel);
+        }
         dbus_message_unref(msg);
         dbus_error_free(&error);
     }
@@ -1403,6 +1517,55 @@ void KimpanelServiceExistCallback(DBusPendingCall *call, void *data)
     dbus_pending_call_unref(call);
 }
 
+void KimpanelIntrospect(FcitxKimpanelUI* kimpanel)
+{
+    const char* kimpanelServiceName = "org.kde.impanel";
+    DBusMessage* message = dbus_message_new_method_call(kimpanelServiceName, "/org/kde/impanel", DBUS_INTERFACE_INTROSPECTABLE, "Introspect");
+
+    DBusPendingCall* call = NULL;
+    dbus_bool_t reply = dbus_connection_send_with_reply(kimpanel->conn, message,
+                        &call,
+                        0);
+    if (reply == TRUE) {
+        dbus_pending_call_set_notify(call,
+                                    KimpanelIntrospectCallback,
+                                    kimpanel,
+                                    NULL);
+    }
+    dbus_connection_flush(kimpanel->conn);
+    dbus_message_unref(message);
+}
+
+void KimpanelIntrospectCallback(DBusPendingCall* call, void* data)
+{
+    FcitxKimpanelUI* kimpanel = (FcitxKimpanelUI*) data;
+
+    DBusMessage* msg = dbus_pending_call_steal_reply(call);
+
+    if (msg) {
+        const char* s;
+        DBusError error;
+        dbus_error_init(&error);
+        if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &s , DBUS_TYPE_INVALID)) {
+            if (strstr(s, "org.kde.impanel2"))
+                kimpanel->version = 2;
+        }
+        dbus_message_unref(msg);
+        dbus_error_free(&error);
+    }
+
+    dbus_pending_call_cancel(call);
+    dbus_pending_call_unref(call);
+}
+
+
+void KimpanelSuspend(void* arg)
+{
+    FcitxKimpanelUI* kimpanel = (FcitxKimpanelUI*) arg;
+    kimpanel->version = 1;
+}
+
+
 void KimpanelDestroy(void* arg)
 {
     FcitxKimpanelUI* kimpanel = (FcitxKimpanelUI*) arg;
@@ -1427,7 +1590,7 @@ void KimpanelDestroy(void* arg)
 
     dbus_connection_flush(kimpanel->conn);
     dbus_error_free(&err);
-    
+
     free(kimpanel->messageUp);
     free(kimpanel->messageDown);
     free(kimpanel);
